/*
 *  TOPPERS/FMP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Flexible MultiProcessor Kernel
 * 
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2004-2015 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2016-2017 by TOPPERS PROJECT Educational Working Group.
 * 
 *  上記著作権者は，以下の(1)～(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 * 
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 * 
 *  $Id: lcdshield.c 2416 2016-02-12 12:05:32Z roi $
 */

/* 
 *  LCDシールド制御プログラムの本体
 */

#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include <stdio.h>
#include <string.h>
#include <target_syssvc.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include "kernel_cfg.h"
#include "device.h"
#include "pinmode.h"
#include "lcdshield.h"
#include "topamelogo.h"

/*
 *  サービスコールのエラーのログ出力
 */
Inline void
svc_perror(const char *file, int_t line, const char *expr, ER ercd)
{
	if (ercd < 0) {
		t_perror(LOG_ERROR, file, line, expr, ercd);
	}
}

#define	SVC_PERROR(expr)	svc_perror(__FILE__, __LINE__, #expr, (expr))


LCD_Handler_t   LcdHandle;
LCD_DrawProp_t DrawProp;

static Point testPolygon[7] = {
	{0, 0},
	{127, 159},
	{64, 0},
	{64, 159},
	{127, 0},
	{0, 159},
	{0, 0}
};

static const uint32_t pos_tbl[NUM_JSPOSITION-1][2] = {
	{  395, JS_DOWN  },
	{  790, JS_RIGHT },
	{ 1185, JS_ON    },
	{ 1775, JS_UP    },
	{ 2990, JS_LEFT  }
};

/*
 *  SW1割込み
 */
void sw_int(void)
{
	syslog_0(LOG_NOTICE, "## sw_int() ##");
}

/*
 *  文字描画
 *  param1  x0: 文字描画 X座標始点
 *  param2  y0: 文字描画 Y座標始点
 *  param3  c:  文字イメージへのポインタ
 */
static void
DrawChar(uint16_t x0, uint16_t y0, const font_t *pf)
{
	LCD_Handler_t *hlcd = DrawProp.hlcd;
	uint32_t h, w;
	const uint8_t  *pchar;

	if(pf == NULL)
		return;
	lcd_setAddrWindow(hlcd, x0, y0, x0 + pf->width - 1, y0 + pf->height - 1);

	for(h = 0 ; h < pf->height ; h++){
		pchar = (pf->image + pf->stride * h);
		for(w = 0; w < pf->width ; w++){
			if(pf->stride != 0 && (pchar[w/8] & (1 << (7 - (w % 8)))) != 0){
				lcd_writedata2(hlcd, DrawProp.TextColor, 1);
			}
			else{
				lcd_writedata2(hlcd, DrawProp.BackColor, 1);
			}
		}
	}
}

/*
 *  一文字描画
 *  param1  x0: 文字描画 X座標始点
 *  param2  y0: 文字描画 Y座標始点
 *  param3  描画アスキーコード(範囲は0x20から0x7Eまで)
 */
void
lcd_DisplayChar(uint16_t x0, uint16_t y0, uint8_t Ascii)
{
	font_t  **phead = (font_t **)DrawProp.pFont;
	DrawChar(x0, y0, phead[Ascii]);
}

/*
 *  文字列描画
 *  param1  x0: X position (in pixel)
 *  param2  y0: Y position (in pixel)   
 *  param3  Text: Pointer to string to display on LCD
 *  param4  Mode: Display mode
 */
void
lcd_DisplayStringAt(uint16_t x0, uint16_t y0, char *Text, Text_AlignModeTypdef Mode)
{
	font_t   **phead = (font_t **)DrawProp.pFont;
	uint16_t ref_column = 1;
	uint32_t size = 0;
	uint8_t  *ptr = (uint8_t *)Text, c;

	/* Get the text size */
	while (*ptr){
		size += phead[*ptr]->width;
		ptr++;
	}

	switch (Mode){
	case CENTER_MODE:
		ref_column = x0 + (DrawProp.hlcd->_width - size) / 2;
		break;
	case LEFT_MODE:
		ref_column = x0;
		break;
	case RIGHT_MODE:
		ref_column = - x0 + (DrawProp.hlcd->_width - size);
		break;
	default:
		ref_column = x0;
		break;
	}

	/* Check that the Start column is located in the screen */
	if ((ref_column < 1) || (ref_column >= 0x8000)){
		ref_column = 1;
	}

	/* Send the string character by character on LCD */
	while (((c = *Text) != 0) && phead[c] != 0 && (ref_column + phead[c]->width) <= DrawProp.hlcd->_width){
		/* Display one character on LCD */
		lcd_DisplayChar(ref_column, y0, c);
		/* Decrement the column position by 16 */
		ref_column += phead[c]->width;
		/* Point on the next character */
		Text++;
	}
}

/*
 *  グラフィック表示テスト
 */
void
grapics_test(LCD_Handler_t *hlcd)
{
	uint16_t x, y;
	int      i;

	DrawProp.BackColor = ST7735_BLACK;
	lcd_fillScreen(&DrawProp);
	dly_tsk(1000);
	lcd_drawPixel(hlcd, hlcd->_width/2, hlcd->_height/2, ST7735_GREEN);
	dly_tsk(1000);

	lcd_fillScreen(&DrawProp);
	for(y=0; y < hlcd->_height; y+=5) {
		lcd_drawFastHLine(hlcd, 0, y, hlcd->_width, ST7735_RED);
	}
	for (x=0; x < hlcd->_width; x+=5) {
    	lcd_drawFastVLine(hlcd, x, 0, hlcd->_height, ST7735_BLUE);
	}
	dly_tsk(1000);

	lcd_fillScreen(&DrawProp);
	DrawProp.TextColor = ST7735_GREEN;
	lcd_drawPolygon(&DrawProp, testPolygon, sizeof(testPolygon)/sizeof(Point));
	dly_tsk(1000);

	lcd_fillScreen(&DrawProp);
	DrawProp.TextColor = ST7735_GREEN;
	for (x=0; x < hlcd->_width; x+=6) {
		lcd_drawRect(&DrawProp, hlcd->_width/2 -x/2, hlcd->_height/2 -x/2 , x, x);
	}
	for(i = 0 ; i < 3 ; i++){
		lcd_invertDisplay(hlcd, true);
		dly_tsk(500);
		lcd_invertDisplay(hlcd, false);
		dly_tsk(500);
	}
	lcd_fillScreen(&DrawProp);
	lcd_drawBitmap(hlcd, 33, 48, (uint8_t *)topamelogo);
	dly_tsk(1000);
}


/*
 *  メインタスク
 */
void
main_task(intptr_t exinf)
{
	LCD_Handler_t   *hlcd;
	FPGASPI_Handle_t  *hspi;
	ER_UINT	ercd;
	uint32_t value;
	int  event = JS_OFF, oevent = JS_OFF;
	int  no, i;

	SVC_PERROR(syslog_msk_log(LOG_UPTO(LOG_INFO), LOG_UPTO(LOG_EMERG)));
	syslog(LOG_NOTICE, "Sample program starts (exinf = %d).", (int_t) exinf);

	/*
	 *  シリアルポートの初期化
	 *
	 *  システムログタスクと同じシリアルポートを使う場合など，シリアル
	 *  ポートがオープン済みの場合にはここでE_OBJエラーになるが，支障は
	 *  ない．
	 */
	ercd = serial_opn_por(TASK_PORTID);
	if (ercd < 0 && MERCD(ercd) != E_OBJ) {
		syslog(LOG_ERROR, "%s (%d) reported by `serial_opn_por'.",
									itron_strerror(ercd), SERCD(ercd));
	}
	SVC_PERROR(serial_ctl_por(TASK_PORTID,
							(IOCTL_CRLF | IOCTL_FCSND | IOCTL_FCRCV)));

	syslog_0(LOG_NOTICE, "## START ##");
	pinMode(RST_PORT, OUTPUT);
	pinMode(RS_PORT, OUTPUT);
	hspi = fpgaspi_init(FPGASPI1_PORTID);
	if(hspi == NULL){
		syslog_0(LOG_ERROR, "hspi initial error !");
		slp_tsk();
	}
	hspi->semid   = SPITRS_SEM1;
	hspi->semlock = 0;

	hlcd = &LcdHandle;
	hlcd->hspi     = hspi;
	hlcd->spi_lock = 0;
	hlcd->cs_pin   = CS_PORT;
	hlcd->rs_pin   = RS_PORT;
	hlcd->rst_pin  = RST_PORT;
	hlcd->cs2_pin  = CS2_PORT;
	DrawProp.hlcd  = hlcd;
	lcd_initR(hlcd, INITR_BLACKTAB);

	DrawProp.BackColor = ST7735_BLACK;
	DrawProp.TextColor = ST7735_WHITE;
	grapics_test(hlcd);

	DrawProp.TextColor = ST7735_BLACK;
	DrawProp.BackColor = ST7735_WHITE;
	lcd_fillScreen(&DrawProp);
	lcd_fillRect(hlcd, 0, 0, hlcd->_width, hlcd->_height/3, ST7735_BLUE);

	DrawProp.TextColor = ST7735_BLACK;
	DrawProp.BackColor = ST7735_WHITE;

	DrawProp.TextColor = ST7735_WHITE;
	DrawProp.BackColor = ST7735_BLUE;
	DrawProp.pFont = &Font9;
	lcd_DisplayStringAt(0, 12, "TOPPERS", CENTER_MODE);
	lcd_DisplayStringAt(0, 24, "BASE PROTFORM(CV)", CENTER_MODE);

	DrawProp.TextColor = ST7735_BLACK;
	DrawProp.BackColor = ST7735_WHITE;
	DrawProp.pFont = &Font9;
	lcd_DisplayStringAt(0, 100, "Helle World !", CENTER_MODE);

	DrawProp.TextColor = ST7735_RED;
	for(i = 20 ; i < 40 ; i++){
		lcd_DrawCircle(&DrawProp, hlcd->_width/2, hlcd->_height/2+30, i);
		if(i == 25)
			i += 10;
	}
	for(no = 0 ; no < 10000 ; no++){
		sil_wrw_mem((uint32_t *)(TADR_FPGAADC_BASE+TOFF_ADC_UPDATE), 0);
		value = sil_rew_mem((uint32_t *)(TADR_FPGAADC_BASE+TOFF_ADC_CH3_REG));
		for(i = 0 ; i < (NUM_JSPOSITION-1) ; i++){
			if(value < pos_tbl[i][0]){
				event = pos_tbl[i][1];
				break;
			}
		}
		if(i == (NUM_JSPOSITION-1))
			event = JS_OFF;
		if(event != oevent){
			syslog_2(LOG_NOTICE, "ADC JOY(%d) value[%08x]", event, value);
			oevent = event;
		}
		dly_tsk(100);
	}

	syslog_0(LOG_NOTICE, "LCD TEST END");
	syslog_0(LOG_NOTICE, "## STOP ##");
	slp_tsk();
	fpgaspi_deinit(hspi);

	syslog(LOG_NOTICE, "Sample program ends.");
//	SVC_PERROR(ext_ker());
}

