/******************************************************************************
*
* Copyright 2014 Altera Corporation. All Rights Reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
* 1. Redistributions of source code must retain the above copyright notice,
* this list of conditions and the following disclaimer.
*
* 2. Redistributions in binary form must reproduce the above copyright notice,
* this list of conditions and the following disclaimer in the documentation
* and/or other materials provided with the distribution.
*
* 3. The name of the author may not be used to endorse or promote products
* derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER "AS IS" AND ANY EXPRESS OR
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, ARE DISCLAIMED. IN NO
* EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
* OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
* IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
* OF SUCH DAMAGE.
*
******************************************************************************/

#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include <sil.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include "kernel_cfg.h"
#include <stdio.h>
#include <string.h>

#include "device.h"
#include "i2c.h"
#include "i2c_demo.h"

#define I2C_MASTER_DEVICE     ALT_I2C_I2C0                // Device to be used as master
#define I2C_SLAVE_DEVICE      ALT_I2C_I2C1                // Device to be used as slave
#define I2C_SLAVE_ADDRESS     0x86                        // I2C address of the slave device
#define I2C_INTERRUPT_SLAVE   ALT_INT_INTERRUPT_I2C1_IRQ  // Slave IRQ
#define I2C_MASTER_SPEED      40000                       // I2C bus speed for master
#define I2C_BUFFER_SIZE       64                          // Size of data to be exchanged, up to 64bytes


char buffer_send[I2C_BUFFER_SIZE];
char buffer_recv[I2C_BUFFER_SIZE];
char buffer_srecv[I2C_BUFFER_SIZE];

extern void HAL_I2C_MemTxCpltCallback(I2C_Handle_t *hi2c);
extern void HAL_I2C_MemRxCpltCallback(I2C_Handle_t *hi2c);
extern void HAL_I2C_ErrorCallback(I2C_Handle_t *hi2c);

void slave_task(intptr_t exinf)
{
    ER status = E_OK;
    I2C_Init_t cfg;
    I2C_Handle_t *device_slave = NULL;
	int i;

    cfg.AddressingMode = I2C_ADDRESSINGMODE_7BIT;
    cfg.OwnAddress1    = I2C_SLAVE_ADDRESS;
	cfg.SdaHoldClock   = 1;
    cfg.nack_enable    = I2C_NACK_E_NORMAL;
	cfg.semid     = I2CTRS2_SEM;
	cfg.semlock   = I2CLOC2_SEM;
    device_slave = i2c_init(I2C2_PORTID, &cfg);
	if(device_slave == NULL)
		status = E_PAR;
	else{
		device_slave->writecallback = HAL_I2C_MemTxCpltCallback;
		device_slave->readcallback  = HAL_I2C_MemRxCpltCallback;
		device_slave->errorcallback = HAL_I2C_ErrorCallback;
	}
	syslog_1(LOG_NOTICE, "## i2c_initS(%d) ##", status);
#if 1	/* ROI DEBUG */
	i2c_int_enable(device_slave, I2C_INTR_STAT_TX_EMPTY | I2C_INTR_STAT_RX_FULL);
#endif	/* ROI DEBUG */
    // Perform the transfers
    for (i=0; i<2; i++)
    {
        // first run with half size, 2nd run with full size
        int size = i ? (I2C_BUFFER_SIZE/2) : I2C_BUFFER_SIZE;
		if(status == E_OK){
			do{
				status = i2c_slaveread(device_slave, buffer_srecv, size);
    	        syslog_2(LOG_NOTICE, "INFO: Slave receiving no(%d) status(%d).", i, status);
				if(status == E_TMOUT)
					dly_tsk(2000);
			}while(status == E_TMOUT);
		}
		if(status == E_OK){
			status = i2c_slavewrite(device_slave, buffer_srecv, size);
   	        syslog_2(LOG_NOTICE, "INFO: Slave transmit no(%d) status(%d).", i, status);
		}
	}
	syslog_1(LOG_NOTICE, "slave EXIT status(%d) !", status);
	dly_tsk(500);
}

/******************************************************************************/
/*!
 * Master slave demo.
 *
 * \return      result of the function
 */
ER i2c_demo_master_slave(void)
{
    ER status = E_OK;
    I2C_Handle_t *device_master, *device_slave;
    I2C_Init_t cfg;
	int i;

    syslog_0(LOG_NOTICE, "INFO: I2C demo master slave started.");

    // Setup master
    cfg.AddressingMode = I2C_ADDRESSINGMODE_7BIT;
	cfg.ClockSpeed     = I2C_MASTER_SPEED;
	cfg.OwnAddress1     = 0;
	cfg.GeneralCallMode = 0;
	cfg.SdaHoldClock   = 3;
	cfg.restart_enable = I2C_RESATRT_ENABLE;
	cfg.fs_spklen      = 11;
	cfg.semid     = I2CTRS1_SEM;
	cfg.semlock   = I2CLOC1_SEM;
    device_master = i2c_init(I2C1_PORTID, &cfg);
	if(device_master == NULL)
		status = E_PAR;
	else{
		device_master->writecallback = HAL_I2C_MemTxCpltCallback;
		device_master->readcallback  = HAL_I2C_MemRxCpltCallback;
		device_master->errorcallback = HAL_I2C_ErrorCallback;
	}
	syslog_1(LOG_NOTICE, "## i2c_initM(%d) ##", status);
	dly_tsk(100);
	act_tsk(SLAVE_TASK1);

    // Perform the transfers
    for (i=0; i<2; i++)
    {
        // first run with half size, 2nd run with full size
        int size = i ? (I2C_BUFFER_SIZE/2) : I2C_BUFFER_SIZE;

        syslog_1(LOG_NOTICE, "TEST: Buffer size = %d.", size);

        // Transmit
        if (status == E_OK)
        {
            for (i = 0; i < size; ++i)
            {
                buffer_send[i] = (char)rand();
            }
            status = i2c_memwrite(device_master, I2C_SLAVE_ADDRESS, 0, 0, buffer_send, size);
            syslog_2(LOG_NOTICE, "INFO: Master transmitting no(%d) status(%d).", i, status);
        }

        // Receive
        if(status == E_OK)
        {
            syslog_0(LOG_NOTICE, "INFO: Master receiving.");

            // Zero out receive buffer
            memset(buffer_recv, 0, sizeof(buffer_send));

            status = i2c_memread(device_master, I2C_SLAVE_ADDRESS, 0, 0, buffer_recv, size);
            syslog_2(LOG_NOTICE, "INFO: Comparing received data with transmitted data no(%d) status(%d).", i, status);
        }

        // Compare data
        if (status == E_OK)
        {
            if (memcmp(buffer_send, buffer_recv, size) != 0)
            {
                status = E_OBJ;
            }
        }

        // Check for ISR error
        if (status == E_OK)
        {
//            if (slave_context.status != E_OK)
//            {
//                status = slave_context.status;
                syslog_0(LOG_ERROR, "ERROR: ISR Reported error.");
//            }
        }

        // Decode error
        if (status != E_OK)
        {
            syslog_2(LOG_NOTICE, "DEBUG: TX Abort Cause [Master]: Status[%08x] Error[%08x].", device_master->Status, device_master->ErrorCode);

			if(device_slave != NULL)
	            syslog_2(LOG_NOTICE, "DEBUG: TX Abort Cause [Slave]:  Status[%08x] Error[%08x]", device_slave->Status, device_slave->ErrorCode);
        }
    }

    // Cleanup I2C Modules
    if(status == E_OK)
    {
        syslog_0(LOG_NOTICE, "INFO: Cleaning up master.\n");
    	if(status == E_OK)
    	{
        	status = i2c_deinit(device_master);
    	}
    }

    if(status == E_OK)
    {
        syslog_0(LOG_NOTICE, "INFO: Cleaning up slave.\n");
       	status = i2c_deinit(device_slave);
    }

    // Cleanup Interrupt
    if(status == E_OK)
    {
		dis_int(INTNO_I2C2);
    }

    // Display result
    if(status == E_OK)
    {
        syslog_0(LOG_NOTICE, "INFO: I2C demo master slave succeeded.\n");
    }
    else
    {
        syslog_0(LOG_NOTICE, "INFO: I2C demo master slave failed.\n");
    }
	return status;
}
