/*
 *  TOPPERS/FMP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Flexible MultiProcessor Kernel
 * 
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2004-2015 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2017 by TOPPERS PROJECT Educational Working Group.
 * 
 *  上記著作権者は，以下の(1)～(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 * 
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 * 
 *  $Id: media.c 2416 2017-05-27 18:50:36Z roi $
 */

/* 
 *  MEDIAテストの本体
 */

#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include <sil.h>
#include <stdio.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include "kernel_cfg.h"
#include <target_syssvc.h>
#include "storagedevice.h"
#include "fcntl.h"
#include "device.h"
#include "media.h"

/*
 *  サービスコールのエラーのログ出力
 */
Inline void
svc_perror(const char *file, int_t line, const char *expr, ER ercd)
{
	if (ercd < 0) {
		t_perror(LOG_ERROR, file, line, expr, ercd);
	}
}

#define	SVC_PERROR(expr)	svc_perror(__FILE__, __LINE__, #expr, (expr))

/*
 *  ユーザー領域設定
 */
volatile int         ukey = 0;
static volatile int  vol_status = 0;
static uint32_t      heap_area[HEAP_AREA_SIZE/sizeof(uint32_t)];

uint32_t heap_param[2] = {
	(uint32_t)heap_area,
	HEAP_AREA_SIZE
};


/*
 *  CPU例外ハンドラ
 */
#ifdef CPUEXC1

void
cpuexc_handler(void *p_excinf)
{
	ID		tskid;

	syslog(LOG_NOTICE, "CPU exception handler1 (p_excinf = %08p).", p_excinf);
	if (sns_ctx() != true) {
		syslog(LOG_WARNING,
					"sns_ctx() is not true in CPU exception handler.");
	}
	if (sns_dpn() != true) {
		syslog(LOG_WARNING,
					"sns_dpn() is not true in CPU exception handler.");
	}
	syslog(LOG_INFO, "sns_loc = %d sns_dsp = %d sns_tex = %d",
									sns_loc(), sns_dsp(), sns_tex());

	syslog(LOG_INFO, "xsns_dpn = %d xsns_xpn = %d",
									xsns_dpn(p_excinf), xsns_xpn(p_excinf));

    
	if (xsns_xpn(p_excinf)) {
		syslog(LOG_NOTICE, "Sample program ends with exception.");
		SVC_PERROR(ext_ker());
		assert(0);
	}

	SVC_PERROR(iget_tid(&tskid));
	SVC_PERROR(iras_tex(tskid, 0x8000U));
}

#endif /* CPUEXC1 */

/*
 *  SW1割込み
 */
void sw_int(void)
{
	ukey = 1;
	syslog_0(LOG_NOTICE, "## sw_int() ##");
}

/*
 *  ストレージマネージャからのコールバック
 */
static void
storagedev_notice(void *psdev, bool_t sw)
{
	if(sw)
		vol_status |= 1<<((StorageDevice_t*)psdev)->_sdev_devno;
	else
		vol_status &= ~(1<<((StorageDevice_t*)psdev)->_sdev_devno);
	syslog_3(LOG_NOTICE, "MEDIA[%08x][%04x] on/off(%d) !", psdev, vol_status, sw);
}

/*
 *  FAT時間の取得
 */
static uint32_t
get_fat_time(void)
{
	int    reqtime = 0;
#ifndef SAMPLE_PROGRAM
	struct tm2 timedate;

	rtc_get_time(&timedate);
	reqtime  = ((timedate.tm_year-10)<<25)+(timedate.tm_mon<<21)+(timedate.tm_mday<<16);
	reqtime += (timedate.tm_hour<<11)+(timedate.tm_min<<5)+(timedate.tm_sec/2);
#else
	reqtime = 0x00210000;
#endif
	return reqtime;
}

/*
 *  USB HOST INFORMATION CALLBACK
 *  parameter1:  Host Handler
 *  parameter2:  Device Handler
 *  parameter3:  Host state ID.
 */
static void USBH_UserProcess(USBH_HandleTypeDef *phost, TUSBH_Device_t *pdevice, uint8_t id)
{
	switch(id){
	case HOST_USER_SELECT_CONFIGURATION:
		syslog_0(LOG_NOTICE, "USB HOST USER SELECT CONFIGURATION !");
		break;
 
	case HOST_USER_DISCONNECTION:
		syslog_0(LOG_NOTICE, "USB HOST USER DISCONNECT !");
		break;

	case HOST_USER_CLASS_ACTIVE:
		syslog_0(LOG_NOTICE, "USB HOST USER CLASS ACTIVATE !");
		break;

	case HOST_USER_CONNECTION:
		syslog_0(LOG_NOTICE, "USB HOST USER CONNECTION !");
		break;

	default:
		break;
	}
	if(pdevice->pClass != NULL){
		syslog_3(LOG_NOTICE, "## CLASS[%s](%d)(%d) ##", pdevice->pClass->Name, pdevice->type, id);
	}
}

/*
 *  メインタスク
 */
void main_task(intptr_t exinf)
{
	StorageDevice_t *psdev;
	USB_OTG_Init_t USB_Data_Init;
	ER_UINT	ercd;
	TUSBH_ERCODE result;
	USBH_HandleTypeDef *phusbh;
	USB_OTG_Handle_t *husb;

	SVC_PERROR(syslog_msk_log(LOG_UPTO(LOG_INFO), LOG_UPTO(LOG_EMERG)));
	syslog(LOG_NOTICE, "Sample program starts (exinf = %d).", (int_t) exinf);

	/*
	 *  シリアルポートの初期化
	 *
	 *  システムログタスクと同じシリアルポートを使う場合など，シリアル
	 *  ポートがオープン済みの場合にはここでE_OBJエラーになるが，支障は
	 *  ない．
	 */
	ercd = serial_opn_por(TASK_PORTID);
	if (ercd < 0 && MERCD(ercd) != E_OBJ) {
		syslog(LOG_ERROR, "%s (%d) reported by `serial_opn_por'.",
									itron_strerror(ercd), SERCD(ercd));
	}
	SVC_PERROR(serial_ctl_por(TASK_PORTID,
							(IOCTL_CRLF | IOCTL_FCSND | IOCTL_FCRCV)));

	/*
	 *  SDMMCデバイスの設定
	 */
	psdev = SDMGetStorageDevice(0);
	psdev->_sdev_notice = storagedev_notice;
	SDeviceHead._get_datetime = get_fat_time;

	/*
	 *  USB OTGハードの初期化
	 */
    USB_Data_Init.usb_otg_mode = USB_OTG_MODE_HOST;
	USB_Data_Init.host_channels = 11;	/* HOST */
	USB_Data_Init.dma_enable = 1;
	USB_Data_Init.phy_itface = USB_PHY_ULPI; 
	USB_Data_Init.speed = USB_SPEED_HIGH_IN_FULL;
	USB_Data_Init.spnlockid = USB2_SPN;
	husb = usbo_init(USB2_PORTID, &USB_Data_Init);
	if(husb == NULL){
		syslog_0(LOG_NOTICE, "## usbo_init error STOP ##");
		slp_tsk();
	}

	/*
	 *  USB HOST MIDDLEWARE初期化
	 */
	phusbh = (USBH_HandleTypeDef *)tusbmalloc(sizeof(USBH_HandleTypeDef));
	phusbh->pSysData = husb;
	tusbhInit(phusbh, USBH_UserProcess, 0);

	/*
	 *  MSCデバイスの設定
	 */
	psdev = SDMGetStorageDevice(MSC_DEVNO);
	psdev->_sdev_local[1] = phusbh;
	psdev->_sdev_notice = storagedev_notice;

	/*
	 *  USB HOST MIDDLEWAREスタート
	 */
	result = tusbhStart(phusbh);
	syslog_1(LOG_NOTICE, "## USBH_Start result(%d) ##", result);

	syslog_0(LOG_NOTICE, "MAIN TASK STOP !");
	slp_tsk();

	syslog(LOG_NOTICE, "Sample program ends.");
//	SVC_PERROR(ext_ker());
}

