/*
 *  TOPPERS/FMP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Flexible MultiProcessor Kernel
 * 
 *  Copyright (C) 2003-2017 by TOPPERS PROJECT Educational Working Group.
 * 
 *  上記著作権者は，Free Software Foundation によって公表されている 
 *  GNU General Public License の Version 2 に記述されている条件か，以
 *  下の(1)～(4)の条件を満たす場合に限り，本ソフトウェア（本ソフトウェ
 *  アを改変したものを含む．以下同じ）を使用・複製・改変・再配布（以下，
 *  利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを再利用可能なバイナリコード（リロケータブルオブ
 *      ジェクトファイルやライブラリなど）の形で利用する場合には，利用
 *      に伴うドキュメント（利用者マニュアルなど）に，上記の著作権表示，
 *      この利用条件および下記の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを再利用不可能なバイナリコードの形または機器に組
 *      み込んだ形で利用する場合には，次のいずれかの条件を満たすこと．
 *    (a) 利用に伴うドキュメント（利用者マニュアルなど）に，上記の著作
 *        権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 利用の形態を，別に定める方法によって，上記著作権者に報告する
 *        こと．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者を免責すること．
 * 
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者は，
 *  本ソフトウェアに関して，その適用可能性も含めて，いかなる保証も行わ
 *  ない．また，本ソフトウェアの利用により直接的または間接的に生じたい
 *  かなる損害に関しても，その責任を負わない．
 * 
 *  @(#) $Id: task_expansion.c,v 1.1 2017/05/27 11:47:27 roi Exp $
 */

/*
 *	TOPPERS/FMP用タスク管理拡張機能
 */

#include "kernel_impl.h"
#include "check.h"
#include "task.h"
#include "task_expansion.h"
#include "kernel_cfg.h"
#include "target_timer.h"
#include "mp.h"

/*
 *  コンテクスト中からの強制待ち状態への移行
 */
ER
isus_tsk(ID tskid)
{
	TCB  *p_tcb;
	PCB  *p_pcb;
	UINT tstat;
	ER   ercd;

	CHECK_INTCTX_UNL();
	CHECK_TSKID(tskid);
	p_tcb = get_tcb_self(tskid, get_my_p_pcb());

	i_lock_cpu();
	p_pcb = t_acquire_tsk_lock(p_tcb);
	if (TSTAT_DORMANT(tstat = p_tcb->tstat)) {
		ercd = E_OBJ;
	}
	else if (TSTAT_RUNNABLE(tstat)) {
		/*
		 *  実行できる状態から強制待ち状態への遷移
		 */
		p_tcb->tstat = TS_SUSPENDED;
		LOG_TSKSTAT(p_tcb);
		if (make_non_runnable(p_tcb)) {
			if (dispatch_request(p_pcb)){
				(get_my_p_pcb())->reqflg = true;
			}
		}
		ercd = E_OK;
	}
	else if (TSTAT_SUSPENDED(tstat)) {
		ercd = E_QOVR;
	}
	else {
		/*
		 *  待ち状態から二重待ち状態への遷移
		 */
		p_tcb->tstat |= TS_SUSPENDED;
		ercd = E_OK;
	}
	release_tsk_lock(p_pcb);
	i_unlock_cpu();

    error_exit:
	return(ercd);
}

/*
 *  タスク状態参照（簡易版）
 */
ER
ref_tst(ID tskid, T_RTST *pk_rtst)
{
	TCB	*p_tcb;
	UB	tstat;
	ER	ercd;
	PCB *p_pcb;

	CHECK_TSKCTX_UNL();
	CHECK_TSKID_SELF(tskid);

	t_lock_cpu();
	p_tcb = get_tcb_self(tskid, get_my_p_pcb());
	p_pcb = t_acquire_tsk_lock(p_tcb);
	pk_rtst->tskwait  = 0;
	pk_rtst->tskpri   = EXT_TSKPRI(p_tcb->priority);
#ifdef USE_TSKINICTXB
	pk_rtst->inistk   = (VP)p_tcb->p_tinib->tskinictxb.stk_bottom;
	pk_rtst->inistksz = p_tcb->p_tinib->tskinictxb.stksz;
#else
	pk_rtst->inistk   = (VP)p_tcb->p_tinib->stk;
	pk_rtst->inistksz = p_tcb->p_tinib->stksz;
#endif


	tstat = p_tcb->tstat;
	if (TSTAT_DORMANT(tstat)) {
		/*
 		 *  対象タスクが休止状態の場合
		 */
		pk_rtst->tskstat = TTS_DMT;
	}
	else{
		/*
 		 *  タスク状態の取出し
		 */
		if (TSTAT_SUSPENDED(tstat)) {
			if (TSTAT_WAITING(tstat)) {
				pk_rtst->tskstat = TTS_WAS;
			}
			else {
				pk_rtst->tskstat = TTS_SUS;
			}
		}
		else if (TSTAT_WAITING(tstat)) {
			pk_rtst->tskstat = TTS_WAI;
		}
		else if (p_tcb == p_pcb->p_runtsk) {
			pk_rtst->tskstat = TTS_RUN;
		}
		else {
			pk_rtst->tskstat = TTS_RDY;
		}
		if (TSTAT_WAITING(tstat)) {
			/*
	 		 *  待ち要因と待ち対象のオブジェクトのIDの取出し
			 */
			switch (tstat & TS_WAIT_MASK) {
			case TS_WAIT_SLP:
				pk_rtst->tskwait = TTW_SLP;
				break;
			case TS_WAIT_DLY:
				pk_rtst->tskwait = TTW_DLY;
				break;
			case TS_WAIT_SEM:
				pk_rtst->tskwait = TTW_SEM;
				break;
			case TS_WAIT_FLG:
				pk_rtst->tskwait = TTW_FLG;
				break;
			case TS_WAIT_SDTQ:
				pk_rtst->tskwait = TTW_SDTQ;
				break;
			case TS_WAIT_RDTQ:
				pk_rtst->tskwait = TTW_RDTQ;
				break;
			case TS_WAIT_SPDQ:
				pk_rtst->tskwait = TTW_SPDQ;
				break;
			case TS_WAIT_RPDQ:
				pk_rtst->tskwait = TTW_RPDQ;
				break;
			case TS_WAIT_MBX:
				pk_rtst->tskwait = TTW_MBX;
				break;
			case TS_WAIT_MPF:
				pk_rtst->tskwait = TTW_MPF;
				break;
			}
		}
	}
	if (pk_rtst->tskstat == TTS_DMT)
		pk_rtst->tskpc = (FP)p_tcb->p_tinib->task;
	else
		pk_rtst->tskpc = (FP)p_tcb->tskctxb.pc;
	pk_rtst->tsksp   = (VP)p_tcb->tskctxb.sp;
	pk_rtst->prcid   = p_pcb->prcid;
	ercd = E_OK;
	release_tsk_lock(p_pcb);
	t_unlock_cpu();

    error_exit:
	return(ercd);
}

/*
 *  タスク状態のロギング格納領域
 */
static T_TLOG   tsk_log[MAX_TASK_LOG+1];
static uint32_t check_tskid[TNUM_PRCID];
static SYSTIM   check_time[TNUM_PRCID];
static SYSTIM   pervious_time;


/*
 *  システム時間取り出し関数
 *  割込み中でも使用でき、共通化できるように
 *  インライン関数とする．
 */
Inline SYSTIM get_systime(PCB *p_pcb)
{
	return p_pcb->p_tevtcb->next_time;
}

/*
 *  タスクの実測時間測定用の時間
 *  取り出し関数、CPUロック状態で呼び出す．
 */
Inline SYSTIM get_anatime(PCB *p_pcb, bool_t up)
{
	SYSTIM  time;
#if TIC_DENO != 1
	uint_t  subtime;
#endif /* TIC_DENO != 1 */
	CLOCK   clock1, clock2;
	bool_t  ireq;

	/*
	 * 時間の取り出しはコンテキスト中でも行われるので
	 * get_utmと同等の記述をここにおく。
	 */
#if TIC_DENO != 1
	subtime = (INT) next_subtime;
#endif /* TIC_DENO != 1 */
	clock1 = target_timer_get_current();
	ireq = target_timer_probe_int();
	clock2 = target_timer_get_current();

	time = get_systime(p_pcb) * ANA_STIC;
#if TIC_DENO != 1
	time += subtime * ANA_STIC / TIC_DENO;
#endif /* TIC_DENO != 1 */
	if (!(ireq) || clock1 > clock2) {
		time -= TIC_NUME * ANA_STIC / TIC_DENO;
	}
	time += clock1 * ANA_STIC / TIMER_CLOCK;
	if(up && (clock1 % (TIMER_CLOCK / ANA_STIC)) != 0)
		time++;
	return time;
}

/*
 *  タスク実行領域設定関数
 *  この関数はdispatchの開始時、CPUロック状態で呼び出す．
 */
void log_dsp_enter(TCB *p_tcb)
{
	PCB *p_pcb = p_tcb->p_pcb;
	SYSTIM time = get_anatime(p_pcb, true);
	uint32_t index = INDEX_PRC(p_pcb->prcid);
	T_TLOG  *t;

	if(check_tskid[index] < MAX_TASK_LOG)
		t = &tsk_log[check_tskid[index]];
	else
		t = &tsk_log[MAX_TASK_LOG];
	t->runtimes += time - check_time[index];
	check_time[index] = time;
	if(p_pcb->p_schedtsk == NULL)
		check_tskid[index] = 0;
}

/*
 *  タスク実行領域設定関数
 *  この関数はdispatchの終了時時、CPUロック状態で呼び出す．
 */
void log_dsp_leave(TCB *p_tcb)
{
	PCB *p_pcb = p_tcb->p_pcb;
	SYSTIM time = get_anatime(p_pcb, false);
	uint32_t index = INDEX_PRC(p_pcb->prcid);
	T_TLOG  *t;

	if(check_tskid[index] < MAX_TASK_LOG)
		t = &tsk_log[check_tskid[index]];
	else
		t = &tsk_log[MAX_TASK_LOG];
	t->runtimes += time - check_time[index];

	if(p_pcb->p_runtsk){
		check_tskid[index] = TSKID(p_pcb->p_runtsk);
		t = &tsk_log[check_tskid[index]];
		t->runcount++;
	}
	check_time[index] = time;
}

/*
 *  タスクログ状態取り出し関数
 */
int_t get_tsklog(T_TPRM * pprm)
{
	INT no;
	INT num_item=0;

	get_tim(&pprm->currtime);
	t_lock_cpu();
	pprm->pervtime = pervious_time;
	pervious_time  = pprm->currtime;
	for(no = 0 ; no <= tmax_tskid && no <= MAX_TASK_LOG ; no++){
		if(no < NUM_LDSP){
			pprm->tlog[no] = tsk_log[no];
			num_item = no;
		}
		else{
			pprm->tlog[1].runcount += tsk_log[no].runcount;
			pprm->tlog[1].runtimes += tsk_log[no].runtimes;
			num_item = NUM_LDSP-1;
		}
		tsk_log[no].runcount = 0;
		tsk_log[no].runtimes = 0;
	}
	t_unlock_cpu();
	return num_item;
}


/*
 *  デバイス・ポートのログアウト要求判定用データ領域
 */

static T_PCHK port_log[NUM_PCHK];

/*
 *  デバイスログを行うかどうかの判定
 */
static uint_t check_device_log(ulong_t address, ulong_t size)
{
	int    i;
	T_PCHK *p;

	for(i = 0 ; i < NUM_PCHK ; i++){
		p = &port_log[i];
		if(p->portaddress >= address && p->portaddress < (address+size) && p->logtype)
			return (UINT)p->logtype;
	}
	return 0;
}

/*
 *  デバイス・ポートのログ判定データの取り出し
 */
T_PCHK *get_device_log(ulong_t no){
	if(no < NUM_PCHK)
		return &port_log[no];
	else
		return 0;
}

/*
 *  デバイス読み出し表示判定とログアウト関数
 */
void
ana_rdv(ulong_t address, ulong_t data, int_t size)
{
	SYSTIM time;
	PCB    *my_p_pcb;
	ID     id;
	uint_t logtype = check_device_log(address, size);

	if(logtype == 0)
		return;
	my_p_pcb = get_my_p_pcb();
	time = get_systime(my_p_pcb);
	if(sense_context() || my_p_pcb->p_runtsk == 0)
		syslog_3(logtype, "Device Read  time=%09d interrupt port=%08x data=0x%x", time, address, data);
	else{
		id = TSKID(my_p_pcb->p_runtsk);
		switch(size){
		case 4:
			syslog_4(logtype, "Device Read  time=%09d task=%04d port=%08x data=%08x", time, id, address, data);
			break;
		case 2:
			syslog_4(logtype, "Device Read  time=%09d task=%04d port=%08x data=%04x", time, id, address, data);
			break;
		default:
			syslog_4(logtype, "Device Read  time=%09d task=%04d port=%08x data=%02x", time, id, address, data);
			break;
		}
	}
}

/*
 *  デバイス書き込み表示判定とログアウト関数
 */
void
ana_wdv(ulong_t address, ulong_t data, int_t size)
{
	SYSTIM time;
	PCB    *my_p_pcb;
	ID     id;
	uint_t logtype = check_device_log(address, size);

	if(logtype == 0)
		return;
	my_p_pcb = get_my_p_pcb();
	time = get_systime(my_p_pcb);
	if(sense_context() || my_p_pcb->p_runtsk == 0)
		syslog_3(logtype, "Device Write  time=%09d interrupt port=%08x data=0x%x", time, address, data);
	else{
		id = TSKID(my_p_pcb->p_runtsk);
		switch(size){
		case 4:
			syslog_4(logtype, "Device Write time=%09d task=%04d port=%08x data=%08x", time, id, address, data);
			break;
		case 2:
			syslog_4(logtype, "Device Write time=%09d task=%04d port=%08x data=%04x", time, id, address, data);
			break;
		default:
			syslog_4(logtype, "Device Write time=%09d task=%04d port=%08x data=%02x", time, id, address, data);
			break;
		}
	}
}

