/*
 *  TOPPERS/FMP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Flexible MultiProcessor Kernel
 *
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2004-2015 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2015-2017 by TOPPERS PROJECT Educational Working Group.
 * 
 *  上記著作権者は，以下の(1)～(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 * 
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 * 
 *  $Id: pinmode.c 2416 2017-04-18 14:30:18Z roi $
 */

/*
 *  ARDUNO-PINドライバ
 */

#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include <string.h>
#include <target_syssvc.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include "kernel_cfg.h"
#include "pinmode.h"

#define sil_orw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) | (b))
#define sil_andw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) & ~(b))
#define sil_modw_mem(a, b, c)	sil_wrw_mem((a), (sil_rew_mem(a) & (~b)) | (c))


static const Arduino_PortControlBlock Digital_Port[] = {
  { 0,               0},	/* D0 */
  { 0,               0},	/* D1 */
  { TADR_PIO0_BASE,  0},	/* D2 */
  { TADR_PIO0_BASE,  1},	/* D3 */
  { TADR_PIO0_BASE,  2},	/* D4 */
  { TADR_PIO0_BASE,  3},	/* D5 */
  { TADR_PIO0_BASE,  4},	/* D6 */
  { TADR_PIO0_BASE,  5},	/* D7 */
  { TADR_PIO0_BASE,  6},	/* D8 */
  { TADR_PIO0_BASE,  7},	/* D9 */
  { 0,               0},	/* D10 */
  { 0,               0},	/* D11 */
  { 0,               0},	/* D12 */
  { 0,               0},	/* D13 */
  { 0,               0},	/* D13 */
  { 0,               0}		/* D13 */
};

#define NUM_DIGITAL_PORT (sizeof(Digital_Port)/sizeof(Arduino_PortControlBlock))


void
pinMode(uint8_t no, uint8_t mode)
{
	const Arduino_PortControlBlock *ppcb = &Digital_Port[no];

	if(no >= NUM_DIGITAL_PORT)
		return;
	if(ppcb->giobase == 0)
		return;
	if(mode == OUTPUT)
		sil_orw_mem((uint32_t *)(ppcb->giobase+TOFF_PIO_DIRECTION), (1<<ppcb->giopin));
	else
		sil_andw_mem((uint32_t *)(ppcb->giobase+TOFF_PIO_DIRECTION), (1<<ppcb->giopin));
}

void
digitalWrite(uint8_t no, int sw)
{
	const Arduino_PortControlBlock *ppcb = &Digital_Port[no];
	uint32_t tmp;

	if(no >= NUM_DIGITAL_PORT)
		return;
	if(ppcb->giobase == 0)
		return;
	tmp = sil_rew_mem((uint32_t *)(ppcb->giobase+TOFF_PIO_DATA));
	if(sw == 1){
		if((tmp & (1<<(ppcb->giopin))) == 0){
			tmp |= 1<<(ppcb->giopin);
			sil_wrw_mem((uint32_t *)(ppcb->giobase+TOFF_PIO_DATA), tmp);
		}
	}
	else{
		if((tmp & (1<<(ppcb->giopin))) != 0){
			tmp &= ~(1<<(ppcb->giopin));
			sil_wrw_mem((uint32_t *)(ppcb->giobase+TOFF_PIO_DATA), tmp);
		}
	}
}

int
digitalRead(uint8_t no)
{
	const Arduino_PortControlBlock *ppcb = &Digital_Port[no];
 	int sw;

	if(no >= NUM_DIGITAL_PORT)
		return 0;
	if(ppcb->giobase == 0)
		return 0;
	sw = sil_rew_mem((uint32_t *)(ppcb->giobase+TOFF_PIO_DATA));
	sw = (sw>>(ppcb->giopin)) & 1;
	return sw;
}

