/*
 *  TOPPERS/FMP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Flexible MultiProcessor Kernel
 * 
 *  Copyright (C) 2008-2011 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2015-2016 by TOPPERS PROJECT Educational Working Group.
 * 
 *  上記著作権者は，以下の(1)～(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 * 
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 * 
 *  @(#) $Id: device.c 698 2017-01-13 18:40:16Z roi $
 */
/*
 * 
 * 標準デバイス操作関数群の定義
 *
 */
#include "kernel_impl.h"
#include <t_syslog.h>
#include <t_stdlib.h>
#include <sil.h>
#include <target_syssvc.h>
#include "device.h"

/*
 *  SIL関数のマクロ定義
 */
#define sil_orw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) | (b))
#define sil_andw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) & ~(b))
#define sil_modw_mem(a, b, c)	sil_wrw_mem((a), (sil_rew_mem(a) & (~b)) | (c))

#define MIN(a, b) ((a) > (b) ? (b) : (a))

/*
 *  GPIOモードの内部定義
 */
#define GPIO_MODE               0x00000001
#define EXTI_MODE               0x10000000
#define GPIO_MODE_IT            0x00010000
#define EDGE_INT                0x00010000
#define ACTIVE_INT              0x00100000
#define GPIO_OUTPUT_TYPE        0x00000010

static const uint32_t gpio_index[] = {
	TADR_GPIO0_BASE,			/* index 0 */
	TADR_GPIO1_BASE,			/* index 1 */
	TADR_GPIO2_BASE				/* index 2 */
};

#define NUM_OF_GPIOPORT (sizeof(gpio_index)/sizeof(uint32_t))

uint_t dipsw_value;
void (*gpioi_func[NUM_OF_GPIOPORT][NUM_GPIOPIN])(void);
static uint32_t gpio_edge_int[NUM_OF_GPIOPORT];

/*
 *  DEVICEの初期化関数
 */
void
device_init(intptr_t exinf)
{
	int i, j;
	/*
	 *  GPIO管理データの初期化
	 */
	for(i = 0 ; i < NUM_OF_GPIOPORT ; i++){
		for(j = 0 ; j < NUM_GPIOPIN ; j++)
			gpioi_func[i][j] = NULL;
		gpio_edge_int[i] = 0;
	}

	/*
	 *  GPIOのリセット
	 */
	sil_orw_mem((uint32_t *)(TADR_RSTMGR_BASE+TOFF_RSTMGR_PERMODRST),
			 (RSTMGR_PERMODRST_GPIO0 | RSTMGR_PERMODRST_GPIO1 | RSTMGR_PERMODRST_GPIO2));

	/*
	 *  GPIOのリセット解除
	 */
	sil_andw_mem((uint32_t *)(TADR_RSTMGR_BASE+TOFF_RSTMGR_PERMODRST),
			 (RSTMGR_PERMODRST_GPIO0 | RSTMGR_PERMODRST_GPIO1 | RSTMGR_PERMODRST_GPIO2));
}

/*
 *  GPIOの初期設定関数
 */
void
gpio_setup(uint32_t base, GPIO_Init_t *init, uint32_t pin)
{
	uint32_t temp = 0x00;
	uint32_t index;

	for(index = 0 ; index < NUM_OF_GPIOPORT ; index++){
		if(gpio_index[index] == base)
			break;
	}
	if(index == NUM_OF_GPIOPORT)
		return;
	if(pin >= NUM_GPIOPIN)
		return;

	/*  入力モード設定 */
	sil_andw_mem((uint32_t *)(base+TOFF_GPIO_SWPORTA_DDR), (1<<pin));
	/* 割込みモード設定 */
	temp = (init->mode & EDGE_INT) >> 16;
	sil_modw_mem((uint32_t *)(base+TOFF_GPIO_INTTYPE_LEVEL), (1<<pin), (temp<<pin));
	/* 割込みアクティブレベル設定 */
	temp = (init->mode & ACTIVE_INT) >> 20;
	sil_modw_mem((uint32_t *)(base+TOFF_GPIO_INT_POL), (1<<pin), (temp<<pin));
	/* デバンス設定 */
	sil_modw_mem((uint32_t *)(base+TOFF_GPIO_DEBOUNCE), (1<<pin), ((init->debounce & 1)<<pin));
	/* 初期データ設定 */
	sil_modw_mem((uint32_t *)(base+TOFF_GPIO_SWPORTA_DR), (1<<pin), ((init->data & 1)<<pin));
	/*  入出力モード設定 */
	sil_modw_mem((uint32_t *)(base+TOFF_GPIO_SWPORTA_DDR), (GPIO_MODE<<pin), ((init->mode & GPIO_MODE)<<pin));

	/*
	 *  割込み設定
	 */
	if((init->mode & GPIO_MODE) == 0){
		/* マスク */
		if((init->mode & EXTI_MODE) != 0)
			sil_andw_mem((uint32_t *)(base+TOFF_GPIO_INTMSK), (1<<pin));
		else
			sil_orw_mem((uint32_t *)(base+TOFF_GPIO_INTMSK), (1<<pin));
		/* イネーブル設定 */
		sil_orw_mem((uint32_t *)(base+TOFF_GPIO_INTEN), (1<<pin));
		if((init->mode & EDGE_INT) != 0)
			gpio_edge_int[index] |= 1<<pin;
	}
}

/*
 *  GPIO割込みハンドラ
 */
void
gpio_dev_isr(intptr_t exinf)
{
	uint32_t index = (uint32_t)exinf;
	uint32_t status, base;
	volatile uint32_t   tmp;
	int      i;

	base = gpio_index[index];
	status = sil_rew_mem((uint32_t *)(base+TOFF_GPIO_INTSTAT));
	for(i = 0 ; i < NUM_GPIOPIN ; i++){
		if((status & (1<<i)) != 0 && gpioi_func[index][i] != NULL)
			gpioi_func[index][i]();
	}

	/*
	 *  割込み要因クリア
	 */
	if((gpio_edge_int[1] & status) != 0){
		sil_wrw_mem((uint32_t *)(base+TOFF_GPIO_PORTA_EOI), status);
	}
	sil_wrw_mem((uint32_t *)(base+TOFF_GPIO_INTSTAT), 0);
	tmp = sil_rew_mem((uint32_t *)(base+TOFF_GPIO_INTSTAT));
	((void)(tmp));
}


#define CACHE_LINE_MASK         (CACHE_LINE_SIZE-1)

#define ALT_MPUL2_COMMON_PA_C_SET_MSK     (0x1     << 0)

#define ALT_CACHE_L2_SYNC_TIMEOUT             128

#define ALT_CACHE_L2_INVALIDATE_ADDR_TIMEOUT  128
#define ALT_CACHE_L2_CLEAN_ADDR_TIMEOUT       128
#define ALT_CACHE_L2_PURGE_ADDR_TIMEOUT       128

/*
 *  アドレス指定Dキャッシュインバリデート関数
 *  parameter1  addr    address (aligned to 32-byte boundary)
 *  parameter2  dsize   size of memory block (in number of bytes)
 */
void
invalidatedcache_by_addr(uint8_t *addr, int32_t dsize)
{
	uint32_t op_end = (((uint32_t)addr) + dsize + CACHE_LINE_MASK) & ~CACHE_LINE_MASK;
	uint32_t op_addr = (((uint32_t)addr) & ~CACHE_LINE_MASK);
	int i = 0;
	uintptr_t pa;

	if(op_addr == op_end)
		return;

	/*
	 *  L2 CACHE インバリデート
	 */
	for (pa = (uintptr_t)op_addr ; pa < (uintptr_t)op_end ; pa += CACHE_LINE_SIZE){
		sil_wrw_mem((uint32_t *)(TADR_MPUL2_BASE+TOFF_MPUL2_INV_PA), pa);
	}

	while(sil_rew_mem((uint32_t *)(TADR_MPUL2_BASE+TOFF_MPUL2_INV_PA)) != 0){
		if (i == ALT_CACHE_L2_INVALIDATE_ADDR_TIMEOUT){
			syslog_0(LOG_ERROR, "invalidatedcache_by_addr purge timeout !");
			return;
		}
		++i;
    }
	/*
	 *  L2 CACHE 同期
	 */
	sil_wrw_mem((uint32_t *)(TADR_MPUL2_BASE+TOFF_MPUL2_CACHE_SYNC), 0);
	i = 0;
	while(sil_rew_mem((uint32_t *)(TADR_MPUL2_BASE+TOFF_MPUL2_CACHE_SYNC)) != 0){
		if (i == ALT_CACHE_L2_SYNC_TIMEOUT){
			syslog_0(LOG_ERROR, "invalidatedcache_by_addr l2 sync timeout !");
			return;
		}
		++i;
    }
    /*
	 *  L1 CACHE インバリデート
	 */
	for(pa = (uintptr_t)op_addr ; pa < (uintptr_t)op_end ; pa += CACHE_LINE_SIZE){
		__asm("MCR p15, 0, %0, c7, c6, 1" : : "r" (pa));
	}

	/* Ensure all cache maintenance operations complete before returning. */
	__asm("dsb");
}

/*
 *  アドレス指定Dキャッシュフラッシュ関数
 *  parameter1  addr    address (aligned to 32-byte boundary)
 *  parameter2  dsize   size of memory block (in number of bytes)
 */
void flushdcache_by_addr(uint8_t *addr, int32_t dsize)
{
	uint32_t op_end = (((uint32_t)addr) + dsize + CACHE_LINE_MASK) & ~CACHE_LINE_MASK;
	uint32_t op_addr = (((uint32_t)addr) & ~CACHE_LINE_MASK);
	int i = 0;
	uintptr_t pa;

	if(op_addr == op_end)
		return;

	/*
	 *  L1 CACHE フラッシュ
	 */
	for(pa = (uintptr_t)op_addr ; pa < (uintptr_t)op_end ; pa += CACHE_LINE_SIZE){
		__asm("MCR p15, 0, %0, c7, c10, 1" : : "r" (pa));
	}

	/* Ensure all cache maintenance operations complete before returning. */
	__asm("dsb");

	/*
	 *  L2 CACHE パージ
	 */
    for(pa = (uintptr_t)op_addr ; pa < (uintptr_t)op_end ; pa += CACHE_LINE_SIZE){
		sil_wrw_mem((uint32_t *)(TADR_MPUL2_BASE+TOFF_MPUL2_CLEAN_PA), pa);
    }
    while((sil_rew_mem((uint32_t *)(TADR_MPUL2_BASE+TOFF_MPUL2_CLEAN_PA)) & ALT_MPUL2_COMMON_PA_C_SET_MSK) != 0){
        if(i == ALT_CACHE_L2_CLEAN_ADDR_TIMEOUT){
			syslog_0(LOG_ERROR, "flushdcache_by_addr purge timeout !");
            return;
        }
        ++i;
    }

	/*
	 *  L2 CACHE 同期
	 */
	sil_wrw_mem((uint32_t *)(TADR_MPUL2_BASE+TOFF_MPUL2_CACHE_SYNC), 0);
	i = 0;
	while(sil_rew_mem((uint32_t *)(TADR_MPUL2_BASE+TOFF_MPUL2_CACHE_SYNC)) != 0){
		if(i == ALT_CACHE_L2_SYNC_TIMEOUT){
			syslog_0(LOG_ERROR, "flushdcache_by_addr l2 sync timeout !");
			return;
		}
		++i;
	}
}


/*
 *  アドレス指定Dキャッシュフラッシュ&インバリデート関数
 *  parameter1  addr    address (aligned to 32-byte boundary)
 *  parameter2  dsize   size of memory block (in number of bytes)
 */
void flushinvalidatedcache_by_addr(uint8_t *addr, int32_t dsize)
{
	uint32_t op_end = (((uint32_t)addr) + dsize + CACHE_LINE_MASK) & ~CACHE_LINE_MASK;
	uint32_t op_addr = (((uint32_t)addr) & ~CACHE_LINE_MASK);
	int i = 0;
	uintptr_t pa;

	if(op_addr == op_end)
		return;

	/*
	 *  L1 CACHE フラッシュ
	 */
	for (pa = (uintptr_t)op_addr ; pa < (uintptr_t)op_end ; pa += CACHE_LINE_SIZE){
		__asm("MCR p15, 0, %0, c7, c10, 1" : : "r" (pa));
	}

	/* Ensure all cache maintenance operations complete before returning. */
	__asm("dsb");

	/*
	 *  L2 CACHE パージ
	 */
	for (pa = op_addr ; pa < op_end ; pa += CACHE_LINE_SIZE){
		sil_wrw_mem((uint32_t *)(TADR_MPUL2_BASE+TOFF_MPUL2_CLEAN_INV_PA), pa);
	}
	while((sil_rew_mem((uint32_t *)(TADR_MPUL2_BASE+TOFF_MPUL2_CLEAN_INV_PA)) & ALT_MPUL2_COMMON_PA_C_SET_MSK) != 0){
		if(i == ALT_CACHE_L2_PURGE_ADDR_TIMEOUT){
			syslog_0(LOG_ERROR, "flushinvalidatedcache_by_addr purge timeout !");
			return;
		}
		++i;
	}

	/*
	 *  L2 CACHE 同期
	 */
	sil_wrw_mem((uint32_t *)(TADR_MPUL2_BASE+TOFF_MPUL2_CACHE_SYNC), 0);
	i = 0;
	while(sil_rew_mem((uint32_t *)(TADR_MPUL2_BASE+TOFF_MPUL2_CACHE_SYNC)) != 0){
		if (i == ALT_CACHE_L2_SYNC_TIMEOUT){
			syslog_0(LOG_ERROR, "flushinvalidatedcache_by_addr l2 sync timeout !");
			return;
		}
		++i;
    }
	/*
	 *  L1 CACHE インバリデート
	 */
	for (pa = (uintptr_t)op_addr ; pa < (uintptr_t)op_end ; pa += CACHE_LINE_SIZE){
		__asm("MCR p15, 0, %0, c7, c6, 1" : : "r" (pa));
	}

    /* Ensure all cache maintenance operations complete before returning. */
    __asm("dsb");
}


/*
 *  L4 WATCH-DOGドライバ
 */
/*
 *  WDOGポートIDから管理ブロックを取り出すためのマクロ
 */
#define INDEX_WDOG(wdogid)  ((uint_t)((wdogid) - 1))

/*
 *  WDOGハードウェア設定構造体
 */

typedef struct _WDOG_PortControlBlock{
	uint32_t              base;				/* WDOGポートベースアドレス */
	uint32_t              resetbit;			/* WDOGモジュールリセットビット */
} WDOG_PortControlBlock;

static const WDOG_PortControlBlock wdog_pcb[NUM_WDOGPORT] = {
	{ TADR_L4WD0_BASE, RSTMGR_PERMODRST_L4WD0},
	{ TADR_L4WD1_BASE, RSTMGR_PERMODRST_L4WD1}
};

/*
 *  L4 WATCH-DOG初期化
 *  parameter1  port: WDOGポート番号
 *  return 正常終了時 E_OK
 */
ER
wdog_init(ID portid)
{
	const WDOG_PortControlBlock *pcb;

	if(portid < WDOG1_PORTID || portid >= NUM_WDOGPORT)
		return E_PAR;
	pcb = &wdog_pcb[INDEX_WDOG(portid)];
	sil_andw_mem((uint32_t *)(TADR_RSTMGR_BASE+TOFF_RSTMGR_PERMODRST), pcb->resetbit);
    return E_OK;
}

/*
 *  L4 WATCH-DOG終了設定
 *  parameter1  port: WDOGポート番号
 *  return ERコード
 */
ER
wdog_deinit(ID portid)
{
	const WDOG_PortControlBlock *pcb;

	if(portid < WDOG1_PORTID || portid >= NUM_WDOGPORT)
		return E_PAR;
	pcb = &wdog_pcb[INDEX_WDOG(portid)];
	sil_orw_mem((uint32_t *)(TADR_RSTMGR_BASE+TOFF_RSTMGR_PERMODRST), pcb->resetbit);
	return E_OK;
}

/*
 *  L4 WATCH-DOGスタート
 *  parameter1  port: WDOGポート番号
 *  return ERコード
 */
ER
wdog_start(ID portid, uint8_t regtim, uint8_t initim)
{
	const WDOG_PortControlBlock *pcb;

	if(portid < WDOG1_PORTID || portid >= NUM_WDOGPORT)
		return E_PAR;
	if(regtim > WDOG_TIMEOUT2G || initim > WDOG_TIMEOUT2G)
		return E_PAR;

	pcb = &wdog_pcb[INDEX_WDOG(portid)];
	sil_wrw_mem((uint32_t *)(pcb->base+TOFF_L4WD_TORR), (initim<<4) | regtim);
	sil_orw_mem((uint32_t *)(pcb->base+TOFF_L4WD_CR), L4WD_CR_WDT_EN);
    return E_OK;
}

/*
 *  L4 WATCH-DOG停止
 *  parameter1  port: WDOGポート番号
 *  return ERコード
 */
ER
wdog_stop(ID portid)
{
	const WDOG_PortControlBlock *pcb;
    uint32_t config, loadreg, tmp;
	int        i;

	if(portid < WDOG1_PORTID || portid >= NUM_WDOGPORT)
		return E_PAR;
	pcb = &wdog_pcb[INDEX_WDOG(portid)];
	config = sil_rew_mem((uint32_t *)(pcb->base+TOFF_L4WD_CR));
	loadreg = sil_rew_mem((uint32_t *)(pcb->base+TOFF_L4WD_TORR));
	sil_orw_mem((uint32_t *)(TADR_RSTMGR_BASE+TOFF_RSTMGR_PERMODRST), pcb->resetbit);
	for(i = 0 ; i < 8 ; i++){
		tmp = sil_rew_mem((uint32_t *)(TADR_RSTMGR_BASE+TOFF_RSTMGR_PERMODRST));
	}
	(void)(tmp);
	sil_andw_mem((uint32_t *)(TADR_RSTMGR_BASE+TOFF_RSTMGR_PERMODRST), pcb->resetbit);
	sil_wrw_mem((uint32_t *)(pcb->base+TOFF_L4WD_TORR), loadreg);
	config &= ~L4WD_CR_WDT_EN;
	sil_wrw_mem((uint32_t *)(pcb->base+TOFF_L4WD_CR), config);
    return  E_OK;
}

/*
 *  L4 WATCH-DOGリセット
 *  parameter1  port: WDOGポート番号
 *  return ERコード
 */
ER
wdog_reset(ID portid)
{
	const WDOG_PortControlBlock *pcb;

	if(portid < WDOG1_PORTID || portid >= NUM_WDOGPORT)
		return E_PAR;
	pcb = &wdog_pcb[INDEX_WDOG(portid)];
	sil_wrw_mem((uint32_t *)(pcb->base+TOFF_L4WD_CRR), 0x76);
	return E_OK;
}


/*
 *  DMAチャネルIDから管理ブロックを取り出すためのマクロ
 */
#define INDEX_DMACH(dmachid)          ((uint_t)((dmachid) - 1))

/*
 *  DMA-FLAG定義
 */
#define DMA_PROGRAM_FLAG_LOOP0 (1 << 0)		/* LOOP0設定中 */
#define DMA_PROGRAM_FLAG_LOOP1 (1 << 1)		/* LOOP1設定中 */
#define DMA_PROGRAM_FLAG_LOOP_ALL (DMA_PROGRAM_FLAG_LOOP0 | DMA_PROGRAM_FLAG_LOOP1)
#define DMA_PROGRAM_FLAG_LOOP0_IS_FE (1 << 2)	/* LOOP0永久ループ */
#define DMA_PROGRAM_FLAG_LOOP1_IS_FE (1 << 3)	/* LOOP1永久ループ */
#define DMA_PROGRAM_FLAG_SAR   (1 << 4)		/* SAR設定中 */
#define DMA_PROGRAM_FLAG_DAR   (1 << 5)		/* DAR設定中 */
#define DMA_PROGRAM_FLAG_ENDED (1 << 7)		/* DMA PROGRAM設定終了 */

#define DMA_CCR_MEMMEM         (DMA_CCR_OPT_SP(0) | DMA_CCR_OPT_SC(7) | DMA_CCR_OPT_DP(0) | DMA_CCR_OPT_DC(7) | DMA_CCR_OPT_ES8)
#define DMA_CCR_ZEROMEM        (DMA_CCR_OPT_SP(0) | DMA_CCR_OPT_SC(0) | DMA_CCR_OPT_DP(0) | DMA_CCR_OPT_DC(7) | DMA_CCR_OPT_ES8)

/*
 *  DMAハンドラ管理
 */
DMACH_Handle_t *DMACh_Handle[NUM_DMACHANNEL];

/*
 *  DMAモジュールの有効化
 */
void
dma_init(intptr_t exinf)
{
	uint8_t  manager_sec = (uint8_t)exinf;
	uint32_t dmactrl = 0;
	int      i;

	if(manager_sec == DMA_SECURITY_NONSECURE)
		dmactrl |= SYSMGR_DMA_CTL_MGRNONSECURE;
	sil_wrw_mem((uint32_t *)(TADR_SYSMGR_BASE+TOFF_SYSMGR_DMA_CTL), dmactrl);
	sil_wrw_mem((uint32_t *)(TADR_SYSMGR_BASE+TOFF_SYSMGR_DMA_PERSECURITY), 0);
	sil_andw_mem((uint32_t *)(TADR_RSTMGR_BASE+TOFF_RSTMGR_PERMODRST), RSTMGR_PERMODRST_DMA);
	for(i = 0 ; i < NUM_DMACHANNEL ; i++)
		DMACh_Handle[i] = NULL;
}

/*
 *  DMAモジュールの無効化
 */
void
dma_deinit(void)
{
	sil_orw_mem((uint32_t *)(TADR_RSTMGR_BASE+TOFF_RSTMGR_PERMODRST), RSTMGR_PERMODRST_DMA);
}

/*
 *  DMA CHANNELオープン
 *  parameter1  DMAチャネルID
 *  parameter2  DMAチャネルハンドラへのポインタ
 *  return      正常終了時、E_OK
 */
ER
dma_channel_open(ID channelid, DMACH_Handle_t *hdmach)
{
	uint8_t  no, periph;
	uint32_t boot_info, temp;

	if(channelid < 1 || channelid > NUM_DMACHANNEL)
		return E_PAR;
	if(hdmach == NULL)
		return E_PAR;
	if(hdmach->periph_info > DMA_PERIPH_UART1_RX)
		return E_PAR;
	if(hdmach->irq_sec > DMA_SECURITY_NONSECURE)
		return E_PAR;
	if(hdmach->periph_sec > DMA_SECURITY_NONSECURE)
		return E_PAR;
	no = INDEX_DMACH(hdmach->dmach_id);
	if(DMACh_Handle[no] != NULL)
		return E_OBJ;

	hdmach->event = no;

	/*
	 *  DMA IREQのセキュア設定
	 */
	temp = sil_rew_mem((uint32_t *)(TADR_SYSMGR_BASE+TOFF_SYSMGR_DMA_CTL));
	if(hdmach->irq_sec == DMA_SECURITY_NONSECURE)
		temp |= 1 << (no+5);
	else
		temp &= ~(1 << (no+5));
	sil_wrw_mem((uint32_t *)(TADR_SYSMGR_BASE+TOFF_SYSMGR_DMA_CTL), temp);

	/*
	 *  DMA ペリフェラルのセキュア設定
	 */
	temp = sil_rew_mem((uint32_t *)(TADR_SYSMGR_BASE+TOFF_SYSMGR_DMA_PERSECURITY));
	if(hdmach->periph_sec == DMA_SECURITY_NONSECURE)
		temp |= 1 << hdmach->periph_info;
	else
		temp &= ~(1 << hdmach->periph_info);
	sil_wrw_mem((uint32_t *)(TADR_SYSMGR_BASE+TOFF_SYSMGR_DMA_PERSECURITY), temp);

	/*
	 *  DMA CAN/FPGA設定
	 */
	boot_info = sil_rew_mem((uint32_t *)(TADR_SYSMGR_BASE+TOFF_SYSMGR_HPSINFO));
	boot_info &= SYSMGR_HPSINFO_CAN;
	if(boot_info != 0 && hdmach->periph_info >= DMA_PERIPH_FPGA_4_OR_CAN0_IF1
			&& hdmach->periph_info <= DMA_PERIPH_FPGA_7_OR_CAN1_IF2){
		periph = hdmach->periph_info - DMA_PERIPH_FPGA_4_OR_CAN0_IF1;
		temp = sil_rew_mem((uint32_t *)(TADR_SYSMGR_BASE+TOFF_SYSMGR_DMA_CTL));
		if(hdmach->periph_can)
			temp |= 1 << periph;
		else
			temp &= ~(1 << periph);
		sil_wrw_mem((uint32_t *)(TADR_SYSMGR_BASE+TOFF_SYSMGR_DMA_CTL), temp);
	}

	/*
	 *  DMA割込み、イベント選択
	 */
	temp = sil_rew_mem((uint32_t *)(TADR_DMA_BASE+TOFF_DMA_INTEN));
	if(hdmach->evt_req == DMA_EVENT_INTREQ)
		temp |= 1 << hdmach->event;
	else
		temp &= ~(1 << hdmach->event);
	sil_wrw_mem((uint32_t *)(TADR_DMA_BASE+TOFF_DMA_INTEN), temp);

	dma_program_start(hdmach);
	DMACh_Handle[no] = hdmach;
	return E_OK;
}

/*
 *  DMA CHANNELクローズ
 *  parameter1  DMAチャネルハンドラへのポインタ
 *  return      正常終了時、E_OK
 */
ER
dma_channel_close(DMACH_Handle_t *hdmach)
{
	ER_UINT raw_state = dma_channel_state(hdmach);

	if(raw_state < 0)
		return E_PAR;
	if(raw_state != DMACH_STATE_STOPPED)
		return E_OBJ;
	DMACh_Handle[INDEX_DMACH(hdmach->dmach_id)] = NULL;
	return E_OK;
}

/*
 *  DMAペリフェラル転送
 *  parameter1  DMAチャネルハンドラへのポインタ
 *  parameter2  転送メモリアドレス
 *  parameter3  転送メモリフレームサイズ
 *  return      正常終了時、E_OK
 */
ER
dma_periph_transfer(DMACH_Handle_t *hdmach, void * addr, uint32_t size)
{
    ER ercd = E_OK;

	if(hdmach == NULL)
		return E_PAR;
	if(size == 0 && hdmach->evt_req == DMA_EVENT_NONE)
		return E_OK;

	hdmach->pbuff     = addr;
	hdmach->xfersize  = size;
	hdmach->xfercount = 0;
	ercd  = dma_program_start(hdmach);
	if(ercd != E_OK)
		return ercd;

	ercd = hdmach->dma_start(hdmach);
	if(ercd != E_OK)
		return ercd;

    /*
	 *  PROGRAM終了処理
	 */
	ercd = dma_program_end(hdmach, false);
	if(ercd != E_OK)
		return ercd;

    /*
	 *  PROGRAM実行
	 */
    return dma_channel_exec(hdmach);
}

/*
 *  DMAメモリセグメント転送
 */
static ER
dma_memory_to_memory_segment(DMACH_Handle_t *hdmach)
{
	uint32_t  burstcount, burstlength;
	bool_t    correction;
	uint32_t  size = hdmach->xfersize - hdmach->xfercount;
	uintptr_t srcpa = (uintptr_t)hdmach->phandle;
	uintptr_t dstpa = (uintptr_t)hdmach->pbuff;
	uint32_t  sizeleft;
	ER ercd = E_OK;

	sizeleft = size;
	syslog_4(LOG_INFO, "dma_memory_to_memory_segment:src[%08x] dst[%08x] size(%d) align(%d)",
				srcpa, dstpa, size, (srcpa & 0x7));
	ercd = dma_program_move(hdmach, DMA_PROGRAM_REG_SAR, (uint32_t)hdmach->phandle);
	if(ercd != E_OK)
		return ercd;
	ercd = dma_program_move(hdmach, DMA_PROGRAM_REG_DAR, (uint32_t)hdmach->pbuff);
	if(ercd != E_OK)
		return ercd;

    /*
	 *  ソースアドレスの８バイトアライン外データの転送
	 */
	if((srcpa & 0x7) != 0){
		uint32_t aligncount = MIN(8 - (srcpa & 0x7), sizeleft);
		sizeleft -= aligncount;

		/*
		 *  DMAコード作成
		 *  - SS8   : Source      burst size of 1-byte
		 *  - DS8   : Destination burst size of 1-byte
		 *  - SBx   : Source      burst length of [aligncount] transfer(s)
		 *  - DBx   : Destination burst length of [aligncount] transfer(s)
		 *  - 他はメモリtoメモリ設定値
		 */
		ercd = dma_program_move(hdmach, DMA_PROGRAM_REG_CCR,
							(  ((aligncount - 1) << 4) | DMA_CCR_OPT_SS8 | DMA_CCR_OPT_SAI	/* SB */
							 | ((aligncount - 1) << 18) | DMA_CCR_OPT_DS8 | DMA_CCR_OPT_DAI	/* DB */
							 | DMA_CCR_MEMMEM));

		if(ercd != E_OK)
			return ercd;
		ercd = dma_program_mode(hdmach, DMA_PROGRAM_INST_MOD_NONE, DMA_PROGRAM_INST_LD);
		if(ercd != E_OK)
			return ercd;
		ercd = dma_program_mode(hdmach, DMA_PROGRAM_INST_MOD_NONE, DMA_PROGRAM_INST_ST);
		if(ercd != E_OK)
			return ercd;
	}

    /*
	 *  １２８、１６バイト転送値
	 */
	burstcount   = sizeleft >> 3;
	burstlength  = burstcount >> 4;

    /*
	 *  モッド８のコレクション判定
	 */
	correction = (burstcount != 0) && ((srcpa & 0x7) != (dstpa & 0x7));
	burstcount  &= 0xf;

	/*
	 *  残り８バイトアラインあまり転送値を計算
	 */
	sizeleft &= 0x7;
	syslog_4(LOG_INFO, "dma_memory_to_memory_segment:burstlength(%d) burstcount(%d) sizeleft(%d) correction(%d)",
		burstlength, burstcount, sizeleft, correction);

    /*
	 *  １２８バイト転送
	 */
	if(burstlength > 0){
		/*
		 *  DMAコード作成
		 *  - SS64  : Source      burst size of 8-byte
		 *  - DS64  : Destination burst size of 8-byte
		 *  - SB16  : Source      burst length of 16 transfers
		 *  - DB16  : Destination burst length of 16 transfers
		 *  - 他はメモリtoメモリ設定値
		 */
		ercd = dma_program_move(hdmach, DMA_PROGRAM_REG_CCR,
								(  DMA_CCR_OPT_SB16 | DMA_CCR_OPT_SS64 | DMA_CCR_OPT_SAI
								 | DMA_CCR_OPT_DB16 | DMA_CCR_OPT_DS64 | DMA_CCR_OPT_DAI
								 | DMA_CCR_MEMMEM ));

		while(burstlength > 0){
			uint32_t loopcount = MIN(burstlength, 256);
			if(ercd != E_OK)
				return ercd;
			burstlength -= loopcount;

			if(ercd == E_OK)
				ercd = dma_program_lpstart(hdmach, loopcount);
			if(ercd == E_OK)
				ercd = dma_program_mode(hdmach, DMA_PROGRAM_INST_MOD_NONE, DMA_PROGRAM_INST_LD);
			if(ercd == E_OK)
				ercd = dma_program_mode(hdmach, DMA_PROGRAM_INST_MOD_NONE, DMA_PROGRAM_INST_ST);
			if(ercd == E_OK)
				ercd = dma_program_lpend(hdmach, DMA_PROGRAM_INST_MOD_NONE, loopcount);
		}
		if(ercd != E_OK)
			return ercd;
	}

	/*
	 *  ８バイト転送
	 */
	if(burstcount > 0){
		/*
		 *  DMAコード作成
		 *  - SS64  : Source      burst size of 8-byte
		 *  - DS64  : Destination burst size of 8-byte
		 *  - SBx   : Source      burst length of [burstlength] transfer(s)
		 *  - DBx   : Destination burst length of [burstlength] transfer(s)
		 *  - 他はメモリtoメモリ設定値
		 */
		ercd = dma_program_move(hdmach, DMA_PROGRAM_REG_CCR,
								(  ((burstcount - 1) << 4) | DMA_CCR_OPT_SS64 | DMA_CCR_OPT_SAI	/* SB */
								 | ((burstcount - 1) << 18) | DMA_CCR_OPT_DS64 | DMA_CCR_OPT_DAI	/* DB */
								 | DMA_CCR_MEMMEM ));
		if(ercd != E_OK)
			return ercd;
		ercd = dma_program_mode(hdmach, DMA_PROGRAM_INST_MOD_NONE, DMA_PROGRAM_INST_LD);
		if(ercd != E_OK)
			return ercd;
		ercd = dma_program_mode(hdmach, DMA_PROGRAM_INST_MOD_NONE, DMA_PROGRAM_INST_ST);
		if(ercd != E_OK)
			return ercd;
	}

    /*
	 *  コレクション転送
	 */
	if(correction){
		/*
		 * 送信、受信アドレスのアラインずれ転送
		 * ずれ分の１バイト転送が必要
		 */
		int correctcount = (dstpa + (8 - (srcpa & 0x7))) & 0x7;

		/*
		 *  DMAコード作成
		 *  - SS8   : Source      burst size of 1-byte
		 *  - DS8   : Destination burst size of 1-byte
		 *  - SBx   : Source      burst length of [correctcount] transfer(s)
		 *  - DBx   : Destination burst length of [correctcount] transfer(s)
		 *  - 他はメモリtoメモリ設定値
		 */
		ercd = dma_program_move(hdmach, DMA_PROGRAM_REG_CCR,
								(  ((correctcount - 1) << 4) | DMA_CCR_OPT_SS8 | DMA_CCR_OPT_SAI	/* SB */
								 | ((correctcount - 1) << 18) | DMA_CCR_OPT_DS8 | DMA_CCR_OPT_DAI /* DB */
								 | DMA_CCR_MEMMEM ));
		if(ercd != E_OK)
			return ercd;
		ercd = dma_program_mode(hdmach, DMA_PROGRAM_INST_MOD_NONE, DMA_PROGRAM_INST_ST);
		if(ercd != E_OK)
			return ercd;
	}

	/*
	 *  残り８バイト未満の転送
	 */
	if(sizeleft > 0){
		/*
		 *  DMAコード作成
		 *  - SS8   : Source      burst size of 1-byte)
		 *  - DS8   : Destination burst size of 1-byte)
		 *  - SBx   : Source      burst length of [sizeleft] transfer(s)
		 *  - DBx   : Destination burst length of [sizeleft] transfer(s)
		 *  - 他はメモリtoメモリ設定値
		 */
		ercd = dma_program_move(hdmach, DMA_PROGRAM_REG_CCR,
								(  ((sizeleft - 1) << 4) | DMA_CCR_OPT_SS8 | DMA_CCR_OPT_SAI	/* SB */
								 | ((sizeleft - 1) << 18) | DMA_CCR_OPT_DS8 | DMA_CCR_OPT_DAI	/* DB */
								 | DMA_CCR_MEMMEM ));
		if(ercd != E_OK)
			return ercd;
		ercd = dma_program_mode(hdmach, DMA_PROGRAM_INST_MOD_NONE, DMA_PROGRAM_INST_LD);
		if(ercd != E_OK)
			return ercd;
		ercd = dma_program_mode(hdmach, DMA_PROGRAM_INST_MOD_NONE, DMA_PROGRAM_INST_ST);
		if(ercd != E_OK)
			return ercd;
	}
	hdmach->phandle += size;
	hdmach->pbuff   += size;
	hdmach->xfercount += size;
	return ercd;
}

/*
 *  DMAメモリ転送
 *  parameter1  DMAチャネルハンドラへのポインタ
 *  parameter2  転送元メモリアドレス
 *  parameter3  転送先メモリアドレス
 *  parameter4  転送メモリフレームサイズ
 *  return      正常終了時、E_OK
 */
ER
dma_memory_to_memory(DMACH_Handle_t * hdmach, void * dst, void * src, uint32_t size)
{
	ER ercd = E_OK;

	if(hdmach == NULL)
		return E_PAR;
	/*
	 *  転送データなし、イベントなしの場合、正常終了
	 */
	if(size == 0 && (hdmach->evt_req != DMA_EVENT_NONE))
		return E_OK;

    /*
	 *  PROGRAM開始処理
	 */
	ercd = dma_program_start(hdmach);

	if(size != 0){
		flushinvalidatedcache_by_addr((uint8_t *)dst, size);
		flushinvalidatedcache_by_addr((uint8_t *)src, size);
		hdmach->pbuff     = (uint8_t *)dst;
		hdmach->phandle   = src;
		hdmach->xfersize  = size;
		hdmach->xfercount = 0;

        /*
		 *  メモリ・オーバーラップ判定
		 */
		if((uintptr_t)dst > (uintptr_t)src){
			if((uintptr_t)src + size - 1 > (uintptr_t)dst)
				return E_PAR;
		}
		else{
			if((uintptr_t)dst + size - 1 > (uintptr_t)src)
				return E_PAR;
		}

		while(hdmach->xfercount < hdmach->xfersize && ercd == E_OK){
            /*
             *  セグメント・メモリ転送
             */
			ercd = dma_memory_to_memory_segment(hdmach);
		}
	}

    /*
	 *  PROGRAM終了処理
	 */
	if(ercd == E_OK)
		ercd = dma_program_end(hdmach, true);

    /*
	 *  PROGRAM実行
	 */
	if(ercd == E_OK)
        ercd = dma_channel_exec(hdmach);
	return ercd;
}

/*
 *  セグメント単位DMAメモリゼロフィル
 */
static ER
dma_zero_to_memory_segment(DMACH_Handle_t * hdmach)
{
	ER ercd = E_OK;
	uintptr_t bufpa = (uintptr_t)hdmach->pbuff;
	uint32_t size = hdmach->xfersize - hdmach->xfercount;
	uint32_t burstlength, burstcount, loop0, loop1;
	uint32_t sizeleft;

	sizeleft = size;
	syslog_3(LOG_INFO, "dma_zero_to_memory_segment:buf[%08x] size(%d) align(%d)",
				bufpa, size, (bufpa & 0x7));
	ercd = dma_program_move(hdmach, DMA_PROGRAM_REG_DAR, (uint32_t)bufpa);
	if(ercd != E_OK)
		return ercd;

    /*
	 *  先頭８バイトアラインに入らない、領域のクリア
	 */
	if(bufpa & 0x7){
		uint32_t aligncount = MIN(8 - (bufpa & 0x7), sizeleft);

		sizeleft -= aligncount;
        /*
		 *  DMAコード作成
		 *  - DS8   : Destination burst size of 1-byte
		 *  - DBx   : Destination burst length of [aligncount] transfer(s)
		 *  - 他はメモリフィル設定値
		 */
		ercd = dma_program_move(hdmach, DMA_PROGRAM_REG_CCR,
								(  DMA_CCR_OPT_SB1 | DMA_CCR_OPT_SS8 | DMA_CCR_OPT_SAI
								 | ((aligncount - 1) << 18) | DMA_CCR_OPT_DS8 | DMA_CCR_OPT_DAI	/* DB */
								 | DMA_CCR_ZEROMEM ));
		if(ercd != E_OK)
			return ercd;
		ercd = dma_program_mode(hdmach, DMA_PROGRAM_INST_MOD_NONE, DMA_PROGRAM_INST_STZ);
		if(ercd != E_OK)
			return ercd;
    }

    /*
	 *  ８バイトバーストサイズ計算、８アラインに入らない終了サイズを計算
	 */
	burstcount  = sizeleft >> 3;
	burstlength = burstcount >> 4;
	burstcount &= 0xf;
	sizeleft   &= 0x7;
	syslog_3(LOG_INFO, "dma_zero_to_memory_segment:burstlength(%d) burstcount(%d) sizeleft(%d)",
		burstlength, burstcount, sizeleft);

    /*
	 *  １２８バイトクリアデータがある場合、ここでクリア
	 */
	if(burstlength > 0){
		/*
		 *  DMAコード作成
		 *  - DS64  : Destination burst size of 8-byte
		 *  - DB16  : Destination burst length of 16 transfers
		 *  - 他はメモリフィル設定値
		 */
		ercd = dma_program_move(hdmach, DMA_PROGRAM_REG_CCR,
								(  DMA_CCR_OPT_SB1 | DMA_CCR_OPT_SS8 | DMA_CCR_OPT_SAI
								 | DMA_CCR_OPT_DB16 | DMA_CCR_OPT_DS64 | DMA_CCR_OPT_DAI
								 | DMA_CCR_ZEROMEM ));

		while(burstlength > 0){
			if(burstlength > 256) {
				loop0 = MIN(burstlength/256,256);
				loop0 = (loop0) ? loop0 : 1;
				loop1 = 256;
			}
			else{
				loop0 = 1;
				loop1 = burstlength;
			}

			if(ercd != E_OK)
				break;

			burstlength -= loop0 * loop1;
			if(ercd == E_OK)
                ercd = dma_program_lpstart(hdmach, loop0);
			if(ercd == E_OK)
				ercd = dma_program_lpstart(hdmach, loop1);
			if(ercd == E_OK)
				ercd = dma_program_mode(hdmach, DMA_PROGRAM_INST_MOD_NONE, DMA_PROGRAM_INST_STZ);
			if(ercd == E_OK)
				ercd = dma_program_lpend(hdmach, DMA_PROGRAM_INST_MOD_NONE, loop1);
			if(ercd == E_OK)
				ercd = dma_program_lpend(hdmach, DMA_PROGRAM_INST_MOD_NONE, loop0);
		}
		if(ercd != E_OK)
			return ercd;
	}

    /*
	 *  １２８バイトクリアに入らない８バイト単位クリア
	 */
	if(burstcount){
		/*
		 *  DMAコード作成
		 *  - DS64  : Destination burst size of 8-byte
		 *  - DBx   : Destination burst length of [burstlength] transfer(s)
		 *  - 他はメモリフィル設定値
		 */
		ercd = dma_program_move(hdmach, DMA_PROGRAM_REG_CCR,
								(  DMA_CCR_OPT_SB1 | DMA_CCR_OPT_SS8 | DMA_CCR_OPT_SAI
								 | ((burstcount - 1) << 18) | DMA_CCR_OPT_DS64 | DMA_CCR_OPT_DAI	/* DB */
								 | DMA_CCR_ZEROMEM ));
		if(ercd != E_OK)
			return ercd;
		ercd = dma_program_mode(hdmach, DMA_PROGRAM_INST_MOD_NONE, DMA_PROGRAM_INST_STZ);
		if(ercd != E_OK)
			return ercd;
	}

	/*
	 *  ８バイトアラインに入らない残り領域のクリア
	 */
	if(sizeleft){
		/*
		 *  DMAコード作成
		 *  - DS8   : Destination burst size of 1-byte
		 *  - DBx   : Destination burst length of [sizeleft] transfer(s)
		 *  - 他はメモリフィル設定値
		 */
		ercd = dma_program_move(hdmach, DMA_PROGRAM_REG_CCR,
								(  DMA_CCR_OPT_SB1 | DMA_CCR_OPT_SS8 | DMA_CCR_OPT_SAI
								 | ((sizeleft - 1) << 18) | DMA_CCR_OPT_DS8 | DMA_CCR_OPT_DAI	/* DB */
								 | DMA_CCR_ZEROMEM ));
		if(ercd != E_OK)
			return ercd;
		ercd = dma_program_mode(hdmach, DMA_PROGRAM_INST_MOD_NONE, DMA_PROGRAM_INST_STZ);
		if(ercd != E_OK)
			return ercd;
	}
	hdmach->pbuff += size;
	hdmach->xfercount += size;
	return ercd;
}

/*
 *  DMAメモリゼロフィル
 *  parameter1  DMAチャネルハンドラへのポインタ
 *  parameter2  転送先メモリアドレス
 *  parameter3  転送メモリフレームサイズ
 *  return      正常終了時、E_OK
 */
ER
dma_zero_to_memory(DMACH_Handle_t * hdmach, void * buf, uint32_t size)
{
	ER ercd = E_OK;

	if(hdmach == NULL)
		return E_PAR;
	/*
	 *  サイズ・ゼロ、イベントなしなら正常終了
	 */
	if((size == 0) && (hdmach->evt_req != DMA_EVENT_NONE))
		return E_PAR;

    /*
	 *  PROGRAM開始処理
	 */
	ercd = dma_program_start(hdmach);

	if(size != 0){
		flushinvalidatedcache_by_addr((uint8_t *)buf, size);
		hdmach->pbuff     = (uint8_t *)buf;
		hdmach->xfersize  = size;
		hdmach->xfercount = 0;

		while(hdmach->xfercount < hdmach->xfersize && ercd == E_OK){
            /*
             *  セグメント・ゼロフィル実行
             */
			ercd = dma_zero_to_memory_segment(hdmach);
		}
	}

    /*
	 *  PROGRAM終了処理
	 */
	if(ercd == E_OK)
		ercd = dma_program_end(hdmach, false);

    /*
	 *  PROGRAM実行
	 */
	if(ercd == E_OK)
        ercd = dma_channel_exec(hdmach);
	return ercd;
}

/*
 *  DMAチャネル実行
 *  parameter1  DMAチャネルハンドラへのポインタ
 *  return      正常終了時、E_OK
 */
ER
dma_channel_exec(DMACH_Handle_t *hdmach)
{
	uint8_t channel;
	int     state;

	if(hdmach == NULL)
		return E_PAR;
	if(hdmach->dmach_id <= 0 && hdmach->dmach_id > NUM_DMACHANNEL)
		return E_PAR;

	channel = INDEX_DMACH(hdmach->dmach_id);

	state = dma_channel_state(hdmach);
	if(state != DMACH_STATE_STOPPED)
		return E_OBJ;

    /*
	 *  DMAコード領域確認
	 */
    if(hdmach->code_size == 0)
		return E_SYS;

    /*
	 *  DMAコードLOOPエンドフラグを確認
	 */
    if(hdmach->flag & DMA_PROGRAM_FLAG_LOOP_ALL)
		return E_SYS;

    /*
	 *  DMAコード設定終了フラグを確認
	 */
	if((hdmach->flag & DMA_PROGRAM_FLAG_ENDED) == 0)
        return E_SYS;

    /*
	 *  DMAコード領域のキャッシュフラッシュ
	 */
	flushinvalidatedcache_by_addr(hdmach->program, hdmach->code_size);

    /*
     *  DMAコードを実行
     */
	sil_wrw_mem((uint32_t *)(TADR_DMA_BASE+TOFF_DMA_DBGINST0), (0xA0<<16) | (channel<<24));
	sil_wrw_mem((uint32_t *)(TADR_DMA_BASE+TOFF_DMA_DBGINST1), (uint32_t)hdmach->program);
	sil_wrw_mem((uint32_t *)(TADR_DMA_BASE+TOFF_DMA_DBGCMD), 0);
    return E_OK;
}

/*
 *  DMA チャネル状態取出し
 *  parameter1  DMAチャネルハンドラへのポインタ
 *  return      正常終了時、E_OK
 */
ER_UINT
dma_channel_state(DMACH_Handle_t *hdmach)
{
	uint8_t no;

	if(hdmach == NULL)
		return E_PAR;
	no = INDEX_DMACH(hdmach->dmach_id);
	if(DMACh_Handle[no] == NULL)
		return E_PAR;
	return sil_rew_mem((uint32_t *)(TADR_DMA_BASE+TOFF_DMA_CSR+no*8)) & 15;
}

/*
 *  DMAチャネル割込みサービスルーチン
 */
void
dma_hps_isr(intptr_t exinf)
{
	uint32_t ino = (int)exinf - 1;
	DMACH_Handle_t *hdmach;
	uint32_t istatus = sil_rew_mem((uint32_t *)(TADR_DMA_BASE+TOFF_DMA_INTMIS));
	uint32_t estatus = sil_rew_mem((uint32_t *)(TADR_DMA_BASE+TOFF_DMA_INT_EVENT_RIS));
	uint32_t fstatus = sil_rew_mem((uint32_t *)(TADR_DMA_BASE+TOFF_DMA_FSRD));
	uint32_t cstatus;

	cstatus = sil_rew_mem((uint32_t *)(TADR_DMA_BASE+TOFF_DMA_FTR+ino*4));
	syslog_5(LOG_INFO, "dma_hps_isr(%d) i[%08x] e[%08x] f[%08x] c[%08x]", (int)exinf, istatus, estatus, fstatus, cstatus);
	hdmach = DMACh_Handle[ino];
	if(hdmach != NULL){
		if((fstatus & DMA_MANAGER_FAULT_UNDEF_INSTR) != 0)
			hdmach->errorcode |= DMA_ERROR_MG_UNDEFINST;
		if((fstatus & DMA_MANAGER_FAULT_OPERAND_INVALID) != 0)
			hdmach->errorcode |= DMA_ERROR_MG_OPEINVALID;
		if((fstatus & DMA_MANAGER_FAULT_DMAGO_ERR) != 0)
			hdmach->errorcode |= DMA_ERROR_MG_DMAGOERR;
		if((fstatus & DMA_MANAGER_FAULT_MGR_EVNT_ERR) != 0)
			hdmach->errorcode |= DMA_ERROR_MG_EVNTERR;
		if((fstatus & DMA_MANAGER_FAULT_INSTR_FETCH_ERR) != 0)
			hdmach->errorcode |= DMA_ERROR_MG_INSTFETERR;
		if((fstatus & DMA_MANAGER_FAULT_DBG_INSTR) != 0)
			hdmach->errorcode |= DMA_ERROR_MG_DBGINST;

		if((cstatus & DMA_CHANNEL_FAULT_UNDEF_INSTR) != 0)
			hdmach->errorcode |= DMA_ERROR_CH_UNDEFINST;
		if((cstatus & DMA_CHANNEL_FAULT_OPERAND_INVALID) != 0)
			hdmach->errorcode |= DMA_ERROR_CH_OPEINVALID;
		if((cstatus & DMA_CHANNEL_FAULT_CH_EVNT_ERR) != 0)
			hdmach->errorcode |= DMA_ERROR_CH_EVTERR;
		if((cstatus & DMA_CHANNEL_FAULT_CH_PERIPH_ERR) != 0)
			hdmach->errorcode |= DMA_ERROR_CH_PERIPHERR;
		if((cstatus & DMA_CHANNEL_FAULT_CH_RDWR_ERR) != 0)
			hdmach->errorcode |= DMA_ERROR_CH_RDWRERR;
		if((cstatus & DMA_CHANNEL_FAULT_MFIFO_ERR) != 0)
			hdmach->errorcode |= DMA_ERROR_CH_MFIFOERR;
		if((cstatus & DMA_CHANNEL_FAULT_ST_DATA_UNAVAILABLE) != 0)
			hdmach->errorcode |= DMA_ERROR_CH_DATAUNAVAL;
		if((cstatus & DMA_CHANNEL_FAULT_INSTR_FETCH_ERR) != 0)
			hdmach->errorcode |= DMA_ERROR_CH_INSTFETERR;
		if((cstatus & DMA_CHANNEL_FAULT_DATA_WRITE_ERR) != 0)
			hdmach->errorcode |= DMA_ERROR_CH_DWRITEERR;
		if((cstatus & DMA_CHANNEL_FAULT_DATA_READ_ERR) != 0)
			hdmach->errorcode |= DMA_ERROR_CH_DREADERR;
		if((cstatus & DMA_CHANNEL_FAULT_DBG_INSTR) != 0)
			hdmach->errorcode |= DMA_ERROR_CH_DBGINST;
		if((cstatus & DMA_CHANNEL_FAULT_LOCKUP_ERR) != 0)
			hdmach->errorcode |= DMA_ERROR_CH_LOCKUPERR;
		if(hdmach->dma_end != NULL)
			hdmach->dma_end(hdmach);
	}
	sil_wrw_mem((uint32_t *)(TADR_DMA_BASE+TOFF_DMA_INTCLR), 1 << ino);
}


/*
 *  DMA PROGRAM初期化
 */
ER
dma_program_start(DMACH_Handle_t *hdmach)
{
    uint32_t buffer;
	uint32_t offset;

	if(hdmach == NULL)
		return E_PAR;
	buffer = (uint32_t)hdmach->buffer;
	offset = ((buffer + CACHE_LINE_SIZE - 1) & ~(CACHE_LINE_SIZE - 1)) - buffer;
	hdmach->program = hdmach->buffer + offset;

    /*
	 *  プログラム領域初期化
	 */
    hdmach->flag      = 0;
    hdmach->code_size = 0;
	return E_OK;
}

/*
 *  DMA PROGRAM MOVE設定
 */
ER
dma_program_move(DMACH_Handle_t *hdmach, uint8_t chan_reg, uint32_t val)
{
    uint8_t * buffer;
    uint8_t rd_mask = 0;

	if(hdmach == NULL)
		return E_PAR;
	if((hdmach->code_size + 6) > DMA_PROGRAM_PROVISION_BUFFER_SIZE)
        return E_OBJ;
	if(chan_reg > DMA_PROGRAM_REG_DAR)
		return E_PAR;

    /* Verify channel register; construct rd mask value */
    switch (chan_reg){
    case DMA_PROGRAM_REG_SAR:
        rd_mask = 0;
		/*
		 *  SARが設定されたら設定情報を格納
		 */
        if (!(hdmach->flag & DMA_PROGRAM_FLAG_SAR)){
            hdmach->flag |= DMA_PROGRAM_FLAG_SAR;
            hdmach->sar = hdmach->code_size + 2;
        }
        break;

    case DMA_PROGRAM_REG_CCR:
        rd_mask = 1;
        break;

    case DMA_PROGRAM_REG_DAR:
    default:
        rd_mask = 2;
		/*
		 *  DARが設定されたら設定情報を格納
		 */
        if (!(hdmach->flag & DMA_PROGRAM_FLAG_DAR)){
            hdmach->flag |= DMA_PROGRAM_FLAG_DAR;
            hdmach->dar = hdmach->code_size + 2;
        }
        break;
    }

    /*
	 *  DMAMOV作成
	 */
    buffer = hdmach->program + hdmach->code_size;
    buffer[0] = 0xbc;
    buffer[1] = rd_mask;
    buffer[2] = (uint8_t)((val >>  0) & 0xff);
    buffer[3] = (uint8_t)((val >>  8) & 0xff);
    buffer[4] = (uint8_t)((val >> 16) & 0xff);
    buffer[5] = (uint8_t)((val >> 24) & 0xff);

    /*
	 *  コードサイズアップデート
	 */
    hdmach->code_size += 6;
    return E_OK;
}

/*
 *  DMA PROGRAM MODE設定
 */
ER
dma_program_mode(DMACH_Handle_t *hdmach, uint8_t mode, uint8_t inst)
{
	uint8_t * buffer;

	if(hdmach == NULL)
		return E_PAR;
	if(mode > DMA_PROGRAM_INST_MOD_BURST)
		return E_PAR;

    /*
	 *  バッファ領域を確認
	 */
 	if((hdmach->code_size + 1) > DMA_PROGRAM_PROVISION_BUFFER_SIZE)
		return E_OBJ;

	/*
	 *  モードデータ(DMALD/DMAST/DMASTZ)作成
	 */
	buffer = hdmach->program + hdmach->code_size;
	buffer[0] = inst | mode;

    /*
	 *  コードサイズアップデート
	 */
    hdmach->code_size += 1;
	return E_OK;
}

/*
 *  DMA PROGRAM LPスタート
 */
ER
dma_program_lpstart(DMACH_Handle_t *hdmach, uint32_t iterations)
{
    uint8_t * buffer;
    uint8_t lc_mask = 0;

	if(hdmach == NULL)
		return E_PAR;
	if(iterations == 1)
		return E_OK;

    /*
	 *  バッファ領域を確認
	 */
	if((hdmach->code_size + 2) > DMA_PROGRAM_PROVISION_BUFFER_SIZE)
		return E_PAR;

    /*
	 *  LOOPレンジ確認
	 */
	if(iterations == 0 || iterations > 256)
		return E_PAR;

	/*
	 *  LOOPフラグ確認
	 */
	switch(hdmach->flag & DMA_PROGRAM_FLAG_LOOP_ALL){
	case 0:							/* LOOPなし、LOOP0を使用 */
		hdmach->flag |= DMA_PROGRAM_FLAG_LOOP0;
        hdmach->loop0 = hdmach->code_size + 2;	/* DMAL00P0開始位置を保存 */
		lc_mask = 0x0;
		break;

	case DMA_PROGRAM_FLAG_LOOP0:	/* LOOP0使用中、LOOP1を使用 */
		hdmach->flag |= DMA_PROGRAM_FLAG_LOOP1;
		hdmach->loop1 = hdmach->code_size + 2;	/* DMAL00P1開始位置を保存 */
		lc_mask = 0x2;
		break;

	case DMA_PROGRAM_FLAG_LOOP_ALL:	/* すべてのLOOPを使用、エラー */
	default:						/* LOOPフラグエラー */
		return E_OBJ;
	}

	/*
	 *  DMALP作成
	 */
	buffer = hdmach->program + hdmach->code_size;
	buffer[0] = 0x20 | lc_mask;
	buffer[1] = (uint8_t)(iterations - 1);

    /*
	 *  コードサイズアップデート
	 */
	hdmach->code_size += 2;
	return E_OK;
}

/*
 *  DMA PROGRAM LPエンド
 */
ER
dma_program_lpend(DMACH_Handle_t *hdmach, uint8_t mode, uint32_t iterations)
{
	uint8_t * buffer;
	uint8_t lc_mask = 0;
	uint8_t nf_mask = 0;
	uint16_t backwards_jump = 0;

	if(hdmach == NULL)
		return E_PAR;
	if(iterations == 1)
		return E_OK;

	if(mode > DMA_PROGRAM_INST_MOD_BURST)
		return E_PAR;

    /*
	 *  バッファ領域を確認
	 */
	if((hdmach->code_size + 2) > DMA_PROGRAM_PROVISION_BUFFER_SIZE)
		return E_OBJ;

	/*
	 *  LOOPフラグ確認
	 */
	switch(hdmach->flag & DMA_PROGRAM_FLAG_LOOP_ALL){
	case DMA_PROGRAM_FLAG_LOOP0:	/* LOOP0のみ使用中 */
		backwards_jump = hdmach->code_size - hdmach->loop0;
		hdmach->flag &= ~DMA_PROGRAM_FLAG_LOOP0;
		hdmach->loop0 = 0;
		lc_mask = 0x0;

		if(hdmach->flag & DMA_PROGRAM_FLAG_LOOP0_IS_FE)
			hdmach->flag &= ~DMA_PROGRAM_FLAG_LOOP0_IS_FE;
		else
			nf_mask = 0x10;
        break;

	case DMA_PROGRAM_FLAG_LOOP_ALL:	/* LOOP0,1を使用中 */
		backwards_jump = hdmach->code_size - hdmach->loop1;
		hdmach->flag &= ~DMA_PROGRAM_FLAG_LOOP1;
		hdmach->loop1 = 0;
		lc_mask = 0x4;

		if(hdmach->flag & DMA_PROGRAM_FLAG_LOOP1_IS_FE)
			hdmach->flag &= ~DMA_PROGRAM_FLAG_LOOP1_IS_FE;
		else
			nf_mask = 0x10;
		break;

	case 0:							/* LOOP未使用エラー */
	default:						/* LOOPフラグエラー */
		return E_OBJ;
	}

	/*
	 *  LOOPジャンプ先レンジオーバー
	 */
	if(backwards_jump > 255)
		return E_OBJ;

	/*
	 *  DMALPEND作成
	 */
	buffer = hdmach->program + hdmach->code_size;
	buffer[0] = 0x28 | nf_mask | lc_mask | mode;
	buffer[1] = (uint8_t)(backwards_jump);

    /*
	 *  コードサイズアップデート
	 */
	hdmach->code_size += 2;
	return E_OK;
}

/*
 *  DMA PROGRAM ペリフェラル設定
 */
ER
dma_program_periph(DMACH_Handle_t *hdmach, uint8_t mode, bool_t in)
{
	uint8_t *buffer;

	if(hdmach == NULL)
		return E_PAR;
	if(mode == DMA_PROGRAM_INST_MOD_NONE || mode > DMA_PROGRAM_INST_MOD_BURST)
		return E_PAR;

    /*
	 *  バッファ領域を確認
	 */
	if((hdmach->code_size + 7) > DMA_PROGRAM_PROVISION_BUFFER_SIZE)
		return E_OBJ;

	/*
	 *  DMAFLUSHP作成
	 */
	buffer = hdmach->program + hdmach->code_size;
	buffer[0] = 0x35;
	buffer[1] = (uint8_t)(hdmach->periph_info) << 3;

    /*
	 *  DMAWFP作成
	 */
    buffer[2] = 0x30 | (mode - 1);
    buffer[3] = (uint8_t)(hdmach->periph_info) << 3;

	if(in){
	    buffer[4] = 0x04 | mode;		/* DMALD作成 */
	    buffer[5] = 0x29 | (mode - 1);	/* DMASTP作成 */
		buffer[6] = (uint8_t)(hdmach->periph_info) << 3;
	}
	else{
	    buffer[4] = 0x25 | (mode - 1);	/* DMALDP作成 */
		buffer[5] = (uint8_t)(hdmach->periph_info) << 3;
	    buffer[6] = 0x08 | mode;		/* DMAST作成 */
	}

    /*
	 *  コードサイズアップデート
	 */
    hdmach->code_size += 7;
	return E_OK;
}

/*
 *  DMA PROGRAM終了
 */
ER
dma_program_end(DMACH_Handle_t *hdmach, bool_t wmb)
{
    uint8_t * buffer;
	uint32_t  len = 1;

	if(hdmach->evt_req != DMA_EVENT_NONE){
		len += 2;
		if(wmb)
			len++;
	}

    /*
	 *  バッファ領域を確認
	 */
	if((hdmach->code_size + len) > DMA_PROGRAM_PROVISION_BUFFER_SIZE){
    	hdmach->flag      = 0;
	    hdmach->code_size = 0;
		return E_OBJ;
	}

    /*
	 *  DMASEVを作成
	 */
    buffer = hdmach->program + hdmach->code_size;
	if(hdmach->evt_req != DMA_EVENT_NONE){
		if(wmb)
			*buffer++ = 0x13;
		*buffer++ = 0x34;
		*buffer++ = (uint8_t)(hdmach->event) << 3;
	}

    /*
	 *  DMAENDを作成
	 */
    *buffer++ = 0x00;

    /*
	 *  コードサイズアップデート
	 */
    hdmach->code_size += len;

    /*
	 *  エンドマーク設定
	 */
    hdmach->flag |= DMA_PROGRAM_FLAG_ENDED;
	return E_OK;
}


/*
 * LED接続ポート
 * 
 * 拡張I/OボードのLED1-3はプログラマブル入出力ポート0に
 * にインバータを介して接続されている．
 */

#define TADR_LED_BASE   TADR_GPIO1_BASE
#define	PIN_START       PINPOSITION24
#define PIN_END         PINPOSITION24


#if defined(PIN_START)
/*
 *  LED接続ポート初期化
 */ 
void
led_init(intptr_t exinf)
{
	GPIO_Init_t GPIO_Init_Data;
	unsigned char pinpos;

	for(pinpos = PIN_START ; pinpos <= PIN_END ; pinpos++){
		if(((1<<pinpos) & LED_MASK) == 0)
			continue;
		/* モード設定 */
		GPIO_Init_Data.mode      = GPIO_MODE_OUTPUT;
		/* Debounce設定 */
		GPIO_Init_Data.debounce  = GPIO_NODEBOUNCE;
		/* 初期データ:消灯 */
		GPIO_Init_Data.data      = 0;	
		gpio_setup(TADR_LED_BASE, &GPIO_Init_Data, pinpos);
	}
}

/*
 *  LED接続ポート読み出し
 */
uint_t
led_in(void)
{
	return sil_rew_mem((uint32_t *)(TADR_LED_BASE+TOFF_GPIO_SWPORTA_DR));
}

/*
 *  LED接続ポート書き込み
 */ 
void
led_out(unsigned int led_data)
{
	sil_modw_mem((uint32_t *)(TADR_LED_BASE+TOFF_GPIO_SWPORTA_DR), LED_MASK, led_data);
}

/*
 * LEDとスイッチの個別設定・読み込み関数群
 */

/*
 *  LEDの状態保存用変数
 */	
unsigned short LedState;


/*
 *  LED点灯制御
 */
void 
set_led_state(unsigned short led, unsigned char state){
	if (state == ON) {
		LedState = LedState | led;
	} else {
		LedState = LedState & ~led;
	}
	led_out(LedState);
}

#endif


#define TADR_PSW_BASE   TADR_GPIO1_BASE
#define PSW_INDEX       1
#define	PSW_PIN         PINPOSITION25


#if defined(PSW_PIN)
/*
 *  PUSHスイッチ接続ポート初期化
 */
void
switch_push_init(intptr_t exinf)
{
	GPIO_Init_t GPIO_Init_Data;

	/* 入力ポート、モード設定 */
	GPIO_Init_Data.mode      = GPIO_MODE_IT_RISING;
	/* Debounce設定 */
	GPIO_Init_Data.debounce  = GPIO_DEBOUNCE;
	/* 初期データ */
	GPIO_Init_Data.data      = 0;
	gpio_setup(TADR_PSW_BASE, &GPIO_Init_Data, PSW_PIN);
}

/*
 *  PUSHスイッチ割込みクリア
 */
void
switch_push_clear(void)
{
	uint32_t status = sil_rew_mem((uint32_t *)(TADR_PSW_BASE+TOFF_GPIO_INTSTAT));
	volatile uint32_t tmp;

	sil_wrw_mem((uint32_t *)(TADR_PSW_BASE+TOFF_GPIO_PORTA_EOI), status);
	sil_wrw_mem((uint32_t *)(TADR_PSW_BASE+TOFF_GPIO_INTSTAT), 0);
	tmp = sil_rew_mem((uint32_t *)(TADR_PSW_BASE+TOFF_GPIO_INTSTAT));
	((void)(tmp));
}


/*
 *  PUSHスイッチコールバック関数設定
 */
void
setup_sw_func(intptr_t exinf)
{
	gpioi_func[PSW_INDEX][PSW_PIN] = (void (*)(void))exinf;
}

#endif


/*
 *  DIPSWの取出し
 */
uint_t
switch_dip_sense(void)
{
	return dipsw_value;
}

