/*
 *  TOPPERS/FMP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Flexible MultiProcessor Kernel
 * 
 *  Copyright (C) 2008-2011 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2016-2018 by TOPPERS PROJECT Educational Working Group.
 * 
 *  上記著作権者は，以下の(1)～(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 * 
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 * 
 *  @(#) $Id: emac.c 698 2018-03-05 18:37:59Z roi $
 */
/*
 * 
 *  EMAC(LWIP用)ドライバ関数群
 *
 */
#include "kernel_impl.h"
#include <t_syslog.h>
#include <t_stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sil.h>
#include <target_syssvc.h>
#include <string.h>
#include "device.h"
#include "phyreg.h"
#include "emac.h"

/*
 *  SIL関数のマクロ定義
 */
#define sil_orw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) | (b))
#define sil_andw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) & ~(b))
#define sil_modw_mem(a, b, c)	sil_wrw_mem((a), (sil_rew_mem(a) & (~b)) | (c))

#define MIN(a, b) ((a) > (b) ? (b) : (a))

#define CACHE_INVALIDATE(a, b)      invalidatedcache_by_addr((uint8_t *)(a), (b))
#define CACHE_FLUSHINVALIDATE(a, b) flushinvalidatedcache_by_addr((uint8_t *)(a), (b))

/*
 *  EMACポートIDから管理ブロックを取り出すためのマクロ
 */
#define INDEX_EMAC(emacid)      ((uint_t)((emacid) - 1))

/*
 *  AUTO NEGOTIATION判定
 */
#define is_autonego(h)          (((h)->linkvalue & EMAC_LINK_AUTONEGO) != 0)

/* MDIO の Ready 待ち時間(msec) */
#define EMAC_MDIO_READY_WAIT_MSEC   50

#define MII_CLKRANGE_150_250M	(0x10)


/*
 *  MIIモード定義
 */
#define EMAC_CTRL_PHYSEL_GMII_MII 0x0
#define EMAC_CTRL_PHYSEL_RGMII  0x1
#define EMAC_CTRL_PHYSEL_RMII   0x2

/*
 *  DESCRIPTOR定義
 */
/*  TX DESCRIPTERステータスビット定義 */
#define DESC_TXSTS_OWNBYDMA     (1 << 31)
#define DESC_TXSTS_TXINT        (1 << 30)
#define DESC_TXSTS_TXLAST       (1 << 29)
#define DESC_TXSTS_TXFIRST      (1 << 28)
#define DESC_TXSTS_TXCRCDIS     (1 << 27)

#define DESC_TXSTS_TXPADDIS     (1 << 26)
#define DESC_TXSTS_TXCHECKINSCTRL (3 << 22)
#define DESC_TXSTS_TXRINGEND    (1 << 21)
#define DESC_TXSTS_TXCHAIN      (1 << 20)
#define DESC_TXSTS_MSK          (0x1FFFF << 0)

/*  RX DESCRIPTERステータスビット定義 */
#define DESC_RXSTS_OWNBYDMA     (1 << 31)
#define DESC_RXSTS_DAFILTERFAIL (1 << 30)
#define DESC_RXSTS_FRMLENMSK    (0x3FFF << 16)
#define DESC_RXSTS_FRMLENSHFT   (16)

#define DESC_RXSTS_ERROR        (1 << 15)
#define DESC_RXSTS_RXTRUNCATED  (1 << 14)
#define DESC_RXSTS_SAFILTERFAIL (1 << 13)
#define DESC_RXSTS_RXIPC_GIANTFRAME	(1 << 12)
#define DESC_RXSTS_RXDAMAGED    (1 << 11)
#define DESC_RXSTS_RXVLANTAG    (1 << 10)
#define DESC_RXSTS_RXFIRST      (1 << 9)
#define DESC_RXSTS_RXLAST       (1 << 8)
#define DESC_RXSTS_RXIPC_GIANT  (1 << 7)
#define DESC_RXSTS_RXCOLLISION  (1 << 6)
#define DESC_RXSTS_RXFRAMEETHER (1 << 5)
#define DESC_RXSTS_RXWATCHDOG   (1 << 4)
#define DESC_RXSTS_RXMIIERROR   (1 << 3)
#define DESC_RXSTS_RXDRIBBLING  (1 << 2)
#define DESC_RXSTS_RXCRC        (1 << 1)

/*  TX DESCRIPTOR 制御ビット定義 */
#define DESC_TXCTRL_SIZE1SHFT   (0)
#define DESC_TXCTRL_SIZE1MASK   (0x1FFF << DESC_TXCTRL_SIZE1SHFT)
#define DESC_TXCTRL_SIZE2SHFT   (16)
#define DESC_TXCTRL_SIZE2MASK   (0x1FFF << DESC_TXCTRL_SIZE2SHFT)

/*  RX DESCRIPTOR 制御ビット定義 */
#define DESC_RXCTRL_RXINTDIS    (1 << 31)
#define DESC_RXCTRL_RXRINGEND   (1 << 15)
#define DESC_RXCTRL_RXCHAIN     (1 << 14)

#define DESC_RXCTRL_SIZE1SHFT   (0)
#define DESC_RXCTRL_SIZE1MASK   (0x1FFF << DESC_RXCTRL_SIZE1SHFT)
#define DESC_RXCTRL_SIZE2SHFT   (16)
#define DESC_RXCTRL_SIZE2MASK   (0x1FFF << DESC_RXCTRL_SIZE2SHFT)

/*
 *  GMAC制御設定
 */
#define GMAC_CORE_INIT          (GMAC_CONTROL_JD | GMAC_CONTROL_PS | GMAC_CONTROL_ACS | \
			GMAC_CONTROL_BE | GMAC_CONTROL_DCRS)

/*
 *  GMAC割り込み設定
 */
#define	GMAC_INT_DEFAULT_MASK	(GMAC_INT_MASK_TIMESTAMP | GMAC_INT_MASK_RGMII |\
				 GMAC_INT_MASK_PCSLINK | GMAC_INT_MASK_PCSAN)

/*
 *  GMAC-DMA割り込み設定
 */
#define DMA_INTER_RXSTAT_ERROR  (DMA_STATUS_OVF | DMA_STATUS_RPS | DMA_STATUS_RWT)
#define DMA_INTR_NORMAL	(DMA_INTR_ENA_NIE | DMA_INTR_ENA_RIE | DMA_INTR_ENA_TIE)
#define DMA_INTR_ABNORMAL	(DMA_INTR_ENA_AIE | DMA_INTR_ENA_FBE | \
				DMA_INTR_ENA_UNE | DMA_INTER_RXSTAT_ERROR)
#define DMA_INTR_DEFAULT_MASK	(DMA_INTR_NORMAL | DMA_INTR_ABNORMAL)

/*
 *  PHY FEATURES 定義
 */
#define ADVERTISE_MASK1         (ADVERTISE_10HALF   | ADVERTISE_10FULL    | \
							     ADVERTISE_100HALF  | ADVERTISE_100FULL   | \
								 ADVERTISE_100BASE4 | ADVERTISE_PAUSE_CAP | ADVERTISE_PAUSE_ASYM)
#define ADVERTISE_MASK2         (ADVERTISE_1000HALF | ADVERTISE_1000FULL)

/*
 *  EMACハードウェア設定構造体
 */
typedef struct _I2C_PortControlBlock{
	uint32_t              base;
	uint32_t              resetbit;
	uint8_t               intno;
	uint8_t               physel_shift;
	uint8_t               phy_sel;
	uint8_t               phy_addr;
} EMAC_PortControlBlock;

/*
 *  EMACポート設定テーブル
 */
static const EMAC_PortControlBlock emac_pcb[NUM_MACPORT] = {
  {	TADR_EMAC0_BASE, RSTMGR_PERMODRST_EMAC0, INTNO_EMAC0_IRQ, 0, EMAC_CTRL_PHYSEL_RGMII, 0},
  {	TADR_EMAC1_BASE, RSTMGR_PERMODRST_EMAC1, INTNO_EMAC1_IRQ, 2, EMAC_CTRL_PHYSEL_RGMII, 1}
};

/*
 *  EMAC ハンドラ
 */
static EMAC_Handle_t EMacHandle[NUM_MACPORT];

/*
 *  MDIOのREADY待ち
 *  parameter1 base  EMACレジスタのベースアドレス
 *  return     E_OKで正常終了
 */
static ER
emac_mdio_wait_ready(uint32_t base)
{
	int     i;

	for (i = 0; i < (EMAC_MDIO_READY_WAIT_MSEC*10); i++) {
		if((sil_rew_mem((uint32_t *)(base+TOFF_GMAC_MII_ADDR)) & GMAC_MII_ADDR_BUSY) == 0)
			return E_OK;
		sil_dly_nse(100*1000);
	}
	syslog_0(LOG_ERROR, "EMAC:Error mdio timeout!");
	return E_TMOUT;
}

/*
 *  PHYレジスタへのデータ書き込み
 *  parameter1 hmac  EMACハンドラへのポインタ
 *  parameter2 reg   レジスタ番号
 *  parameter3 val   書き込み値
 *  return     E_OKで正常終了
 */
static ER
emac_mdio_write(EMAC_Handle_t *hmac, int reg, int val)
{
	uint32_t miiaddr;
	ER       ercd = E_TMOUT;

	if(hmac->Init.semid != 0)
		wai_sem(hmac->Init.semid);

	miiaddr = ((hmac->phy_addr << 11) & GMAC_MII_ADDR_ADDRMSK) | \
		  ((reg << 6) & GMAC_MII_ADDR_REGMSK) | GMAC_MII_ADDR_WRITE;

	/*
	 *  READY待ち
	 */
	if(emac_mdio_wait_ready(hmac->base) == E_OK){
		/*
		 *  データ設定後、PHYアドレスセット
		 */
		sil_wrw_mem((uint32_t *)(hmac->base+TOFF_GMAC_MII_DATA), val);
		sil_wrw_mem((uint32_t *)(hmac->base+TOFF_GMAC_MII_ADDR), (miiaddr | MII_CLKRANGE_150_250M | GMAC_MII_ADDR_BUSY));

		/*
		 *  READY待ち
		 */
		if (emac_mdio_wait_ready(hmac->base) == E_OK)
			ercd = E_OK;
	}

	if(hmac->Init.semid != 0)
		sig_sem(hmac->Init.semid);
	return ercd;
}

/*
 *  PHYレジスタからのデータ読み込み
 *  parameter1 hmac  EMACハンドラへのポインタ
 *  parameter2 reg   レジスタ番号
 *  return     0以上でレジスタ値
 */
static int
emac_mdio_read(EMAC_Handle_t *hmac, int reg)
{
	uint32_t miiaddr;
	int     val = -1;

	if(hmac->Init.semid != 0)
		wai_sem(hmac->Init.semid);

	miiaddr = ((hmac->phy_addr << 11) & GMAC_MII_ADDR_ADDRMSK) | ((reg << 6) & GMAC_MII_ADDR_REGMSK);

	/*
	 *  READY待ち
	 */
	if(emac_mdio_wait_ready(hmac->base) == E_OK){
		/*
		 *  PHYアドレスセット
		 */
		sil_wrw_mem((uint32_t *)(hmac->base+TOFF_GMAC_MII_ADDR), (miiaddr | MII_CLKRANGE_150_250M | GMAC_MII_ADDR_BUSY));

		/*
		 *  READY待ちの後、データ取得
		 */
		if(emac_mdio_wait_ready(hmac->base) == E_OK)
			val = sil_rew_mem((uint32_t *)(hmac->base+TOFF_GMAC_MII_DATA)) & 0x0000ffffUL;
	}

	if(hmac->Init.semid != 0)
		sig_sem(hmac->Init.semid);
	return val;
}

/*
 *  PHYの初期設定
 *  parameter1 hmac   EMACハンドラへのポインタ
 *  parameter2 phyadr PHYアドレス(32以上ならば自動センス)
 *  return     E_OKで正常終了
 */
static ER
emac_phy_init(EMAC_Handle_t *hmac, uint8_t phy_addr)
{
	int  val;

	if(phy_addr < 0x20){
		hmac->phy_addr = phy_addr;
		val = emac_mdio_read(hmac, MII_PHYSID1);
		if((val > 0) && (val != 0xffff))
			return E_OK;
		else{
			syslog_1(LOG_ERROR, "EMAC:emac_phy_init Error unkown phy address(%d)", phy_addr);
			return E_ID;
		}
	}

	/*
	 *  自動センスモード
	 */
	for(phy_addr = 0; phy_addr < 0x20; phy_addr++){
		hmac->phy_addr = phy_addr;
		val = emac_mdio_read(hmac, MII_PHYSID1);
		if((val > 0) && (val != 0xffff))
			return E_OK;
	}
	syslog_0(LOG_ERROR, "EMAC:emac_phy_init Error unkown phy address");
	return E_ID;
}

/*
 *  PHYをリセットする
 *  parameter1 hmac  EMACハンドラへのポインタ
 *  return     E_OKで正常終了
 */
static ER
emac_phy_reset(EMAC_Handle_t *hmac)
{
	int i, val;

	val = emac_mdio_read(hmac, MII_BMCR);
	val &= ~(BMCR_ISOLATE | BMCR_ANENABLE);
	val |= BMCR_RESET;
	emac_mdio_write(hmac, MII_BMCR, val);

	for(i = 0 ; i < 1000; i++){
		val = emac_mdio_read(hmac, MII_BMCR);
		if (val < 0)
			break;
		if ((val & BMCR_RESET) == 0) {
			if (val & BMCR_ISOLATE) {
				val &= (~BMCR_ISOLATE);
				emac_mdio_write(hmac, MII_BMCR, val);
			}
			return E_OK;
		}
		dly_tsk(1);
	}
	syslog_0(LOG_ERROR, "EMAC:emac_phy_reset Error spin reset timeout");
	return E_TMOUT;
}

/*
 *  PHYの状態を取り出す
 *  parameter1 hmac  EMACハンドラへのポインタ
 *  return     PHYの状態値
 */
static uint32_t
emac_phy_features(EMAC_Handle_t *hmac)
{
	uint32_t   bmsr, esr;
	uint32_t   mode = ADVERTISE_PAUSE_CAP | ADVERTISE_PAUSE_ASYM;

	bmsr = emac_mdio_read(hmac, MII_BMSR);
	if((bmsr & BMSR_10HALF) != 0)
		mode |= ADVERTISE_10HALF;
	if((bmsr & BMSR_10FULL) != 0)
		mode |= ADVERTISE_10FULL;
	if((bmsr & BMSR_100HALF) != 0)
		mode |= ADVERTISE_100HALF;
	if((bmsr & BMSR_100FULL) != 0)
		mode |= ADVERTISE_100FULL;
	if((bmsr & BMSR_ESTATEN) != 0){
		esr = emac_mdio_read(hmac, MII_ESTATUS);
		if((esr & ESTATUS_1000_TFULL) != 0)
			mode |= (ADVERTISE_1000FULL<<16);
		if((esr & ESTATUS_1000_THALF) != 0)
			mode |= (ADVERTISE_1000HALF<<16);
	}
	return mode;
}

/*
 *  PHY処理開始要求
 *  parameter1 hmac  EMACハンドラへのポインタ
 *  parameter2 link  PHY設定値
 *  return     E_OKで正常終了
 */
static ER
emac_phy_start(EMAC_Handle_t *hmac, uint32_t link)
{
	ER       ercd;
	uint32_t id;
	uint32_t mode, val;
	int      speed = 0, duplex = 0;

	/*
	 *  PHYをリセットする
	 */
	if((ercd = emac_phy_reset(hmac)) != E_OK)
		return ercd;

	/*
	 *  PHY情報を取得する
	 */
	hmac->phy_features = emac_phy_features(hmac);
	syslog_1(LOG_INFO, "EMAC:emac_phy_start phy featurs[%08x]", hmac->phy_features);

	/*
	 *  PHYのチップIDの確認
	 */
	id  = (emac_mdio_read(hmac, MII_PHYSID1) << 16);
	id |= (emac_mdio_read(hmac, MII_PHYSID2) & 0xffff);

	syslog_2(LOG_INFO, "EMAC:emac_phy_start PHY-ID[OUI&Model:0x%x, Rev:0x%x]", id & ~0xf, id & 0xf);

	/*
	 *  オプション設定があれば、実行
	 */
	if(hmac->phy_opt_init)
		hmac->phy_opt_init(hmac);

	/*
	 *  LINKパラメータを設定
	 */
	link &= ~EMAC_LINK_CONNECT;
	hmac->linkvalue = link;

	switch(link){
	case EMAC_LINK_10BASE_FULL:
		duplex = BMCR_FULLDPLX;
	case EMAC_LINK_10BASE_HALF:
		speed = BMCR_SPEED10;
		break;
	case EMAC_LINK_100BASE_FULL:
		duplex = BMCR_FULLDPLX;
	case EMAC_LINK_100BASE_HALF:
		speed = BMCR_SPEED100;
		break;
	case EMAC_LINK_1000BASE_FULL:
		duplex = BMCR_FULLDPLX;
	case EMAC_LINK_1000BASE_HALF:
		speed = BMCR_SPEED1000;
		break;
	case EMAC_LINK_AUTONEGO:
	default:
		/* Setup Auto-negotiation */
		hmac->linkvalue |= EMAC_LINK_AUTONEGO;
		mode = hmac->phy_features;
		if (mode & ((ADVERTISE_1000FULL<<16) | (ADVERTISE_1000HALF<<16))) {
			val = emac_mdio_read(hmac, MII_CTRL1000);
			val &= ~ADVERTISE_MASK2;
			val |= (mode >> 16) & ADVERTISE_MASK2;
			emac_mdio_write(hmac, MII_CTRL1000, val);
		}
		val = emac_mdio_read(hmac, MII_ADVERTISE);
		val &= ~ADVERTISE_MASK1;
		val |= mode & ADVERTISE_MASK1;
		emac_mdio_write(hmac, MII_ADVERTISE, val);
		break;
	}

	val = emac_mdio_read(hmac, MII_BMCR);
	val &= ~(BMCR_FULLDPLX | BMCR_SPEED100 | BMCR_SPEED1000 | BMCR_ANENABLE);
	val |= is_autonego(hmac) ? (BMCR_ANENABLE | BMCR_ANRESTART) : (speed | duplex);
	emac_mdio_write(hmac, MII_BMCR, val);
	return E_OK;
}

/*
 *  AUTO-NEGOTIATION設定時のLINK状態の取得
 *  parameter1 hmac  EMACハンドラへのポインタ
 *  return     LINK状態
 */
static uint32_t
emac_phy_link_autonego(EMAC_Handle_t *hmac)
{
	uint32_t  link_detect;
	uint32_t  glpa = 0, lpa, adv;

	lpa = emac_mdio_read(hmac, MII_LPA) & emac_mdio_read(hmac, MII_ADVERTISE);
	if(hmac->phy_features & ((ADVERTISE_1000FULL<<16) | (ADVERTISE_1000HALF<<16))){
		adv  = emac_mdio_read(hmac, MII_CTRL1000);
		glpa = emac_mdio_read(hmac, MII_STAT1000);
		glpa &= adv << 2;
	}

	if (glpa & LPA_1000FULL)
		link_detect = EMAC_LINK_1000BASE_FULL;
	else if (glpa & LPA_1000HALF)
		link_detect = EMAC_LINK_1000BASE_HALF;
	else if (lpa & LPA_100FULL)
		link_detect = EMAC_LINK_100BASE_FULL;
	else if (lpa & LPA_100HALF)
		link_detect = EMAC_LINK_100BASE_HALF;
	else if (lpa & LPA_10FULL)
		link_detect = EMAC_LINK_10BASE_FULL;
	else
		link_detect = EMAC_LINK_10BASE_HALF;
	return link_detect;
}

/*
 *  マニュアル設定時のLINK状態の取得
 *  parameter1 hmac  EMACハンドラへのポインタ
 *  return     LINK状態
 */
static uint32_t
emac_phy_link_manual(EMAC_Handle_t *hmac)
{
	uint32_t    link_detect;
	int     bmcr;

	bmcr = emac_mdio_read(hmac, MII_BMCR);

	if(bmcr & BMCR_SPEED1000)
		link_detect = EMAC_SPEED_1000BASE;
	else if (bmcr & BMCR_SPEED100)
		link_detect = EMAC_SPEED_100BASE;
	else
		link_detect = EMAC_SPEED_10BASE;
	link_detect |= (bmcr & BMCR_FULLDPLX) ? EMAC_DUPLEX_FULL : EMAC_DUPLEX_HALF;
	return link_detect;
}

/*
 *  DESCRIPTER領域の初期化
 *  parameter1 hmac  EMACハンドラへのポインタ
 */
static void
emac_bd_set(EMAC_Handle_t *hmac)
{
	uint32_t   *desc, *next;
	uint8_t    *pdata = hmac->Init.databuffer;
	uint32_t   quecount = hmac->Init.rxquecount + hmac->Init.txquecount;
	int        i;

	/*
	 *  DESCRIPTOR/BUFFER領域ののキャッシュをクリア
	 */
	CACHE_INVALIDATE(pdata, (quecount * hmac->Init.paketbufsize));
	CACHE_INVALIDATE(hmac->Init.quebuffer, (quecount * EMAC_QUE_SIZE));

	desc  = (uint32_t *)hmac->Init.quebuffer;
	next  = (uint32_t *)(desc + DESC_TXRX_SIZE);

	hmac->rx_top_desc = desc;
	for(i = 0 ; i < hmac->Init.rxquecount ; i++) {
		desc[DESC_TXRX_BUFFER] = (uint32_t)pdata;
		desc[DESC_TXRX_NEXT]   = (uint32_t)next;
		desc[DESC_TXRX_CONTROL] = 
			(hmac->Init.paketbufsize & DESC_RXCTRL_SIZE1MASK) | \
				      DESC_RXCTRL_RXCHAIN;
		desc[DESC_TXRX_STATUS] = DESC_RXSTS_OWNBYDMA;

		pdata += hmac->Init.paketbufsize;
		desc  += DESC_TXRX_SIZE;
		next  += DESC_TXRX_SIZE;
	}
	*(desc - DESC_TXRX_SIZE + DESC_TXRX_NEXT)
			= (uint32_t)(hmac->rx_top_desc);

	hmac->tx_top_desc = desc;
	for(i = 0 ; i < hmac->Init.txquecount ; i++){
		desc[DESC_TXRX_BUFFER] = (uint32_t)pdata;
		desc[DESC_TXRX_NEXT]   = (uint32_t)next;
		desc[DESC_TXRX_STATUS] &= ~(DESC_TXSTS_TXINT | DESC_TXSTS_TXLAST |
				DESC_TXSTS_TXFIRST | DESC_TXSTS_TXCRCDIS | \
				DESC_TXSTS_TXCHECKINSCTRL | \
				DESC_TXSTS_TXRINGEND | DESC_TXSTS_TXPADDIS);
		desc[DESC_TXRX_STATUS] |= DESC_TXSTS_TXCHAIN;
		desc[DESC_TXRX_CONTROL] = 0;
		desc[DESC_TXRX_STATUS] &= ~(DESC_TXSTS_MSK | DESC_TXSTS_OWNBYDMA);

		pdata += hmac->Init.paketbufsize;
		desc  += DESC_TXRX_SIZE;
		next  += DESC_TXRX_SIZE;
	}
	*(desc - DESC_TXRX_SIZE + DESC_TXRX_NEXT)
			= (uint32_t)(hmac->tx_top_desc);
	CACHE_FLUSHINVALIDATE(hmac->Init.quebuffer, (quecount * EMAC_QUE_SIZE));
}


/*
 *  EMACデバイスの初期化
 *  parameter1 portid EMACのポートID
 *  parameter2 ini    初期化構造体へのポインタ
 *  return            ハンドラへのポインタ
 */
EMAC_Handle_t *
emac_init(ID portid, EMAC_Init_t *ini)
{
	const EMAC_PortControlBlock *pcb;
	EMAC_Handle_t *hmac;
	uint32_t no;

	if(portid == 0 || portid > NUM_MACPORT)
		return NULL;

	no = INDEX_EMAC(portid);
	pcb = &emac_pcb[no];
	hmac = &EMacHandle[no];

	hmac->base     = pcb->base;
	hmac->intno    = pcb->intno;
	memcpy(&hmac->Init, ini, sizeof(EMAC_Init_t));

	dis_int(hmac->intno);
	sil_orw_mem((uint32_t *)(TADR_RSTMGR_BASE+TOFF_RSTMGR_PERMODRST), pcb->resetbit);
	sil_andw_mem((uint32_t *)(TADR_SYSMGR_BASE+TOFF_SYSMGR_EMAC_CTL), 3<<(pcb->physel_shift));
	sil_orw_mem((uint32_t *)(TADR_SYSMGR_BASE+TOFF_SYSMGR_EMAC_CTL), (pcb->phy_sel)<<(pcb->physel_shift));
	sil_andw_mem((uint32_t *)(TADR_RSTMGR_BASE+TOFF_RSTMGR_PERMODRST), pcb->resetbit);

	/*
	 *  DESCRIPTORの初期化
	 */
	emac_bd_set(hmac);

	if(emac_phy_init(hmac, pcb->phy_addr) != E_OK)
		return NULL;
	hmac->macid = portid;
	return hmac;
}

/*
 *  EMAC終了処理
 *  parameter1 hmac  EMACハンドラへのポインタ
 *  return     E_OKで正常終了
 */
ER
emac_deinit(EMAC_Handle_t *hmac)
{
	const EMAC_PortControlBlock *pcb;

	if(hmac == NULL || hmac->macid == 0 || hmac->macid > NUM_MACPORT)
		return E_OBJ;
	pcb = &emac_pcb[INDEX_EMAC(hmac->macid)];
	dis_int(hmac->intno);
	sil_orw_mem((uint32_t *)(TADR_RSTMGR_BASE+TOFF_RSTMGR_PERMODRST), pcb->resetbit);
	hmac->macid = 0;
	return E_OK;
}

/*
 *  EMAC送信要求
 *  parameter1 hmac  EMACハンドラへのポインタ
 *  parameter2 num   送信パケット数
 */
static void
emac_send_req(EMAC_Handle_t *hmac, int num)
{
	volatile uint32_t *desc, *top;
	uint8_t  *cdesc;
	uint32_t base = hmac->base;

	if(num > 0){
		top = desc = hmac->tx_cur_desc;
		desc = (uint32_t *)(desc[DESC_TXRX_NEXT]);
		while (--num) {
			/*
			 *  連続DESCRIPTORの送信DMA KICK
			 */
			desc[DESC_TXRX_STATUS] |= DESC_TXSTS_OWNBYDMA;
			cdesc = (uint8_t *)desc;
			desc = (uint32_t *)(desc[DESC_TXRX_NEXT]);
			CACHE_FLUSHINVALIDATE(cdesc, EMAC_QUE_SIZE);
		}
		/*
		 *  先頭のDESCRIPTORの送信DMA KICK
		 */
		*(top + DESC_TXRX_STATUS) |= DESC_TXSTS_OWNBYDMA;
		CACHE_FLUSHINVALIDATE(top, EMAC_QUE_SIZE);
		hmac->tx_cur_desc = desc;
	}
	sil_wrw_mem((uint32_t *)(base+TOFF_DMA_TX_POLL_DEMAND), 0xFFFFFFFF);
	sil_orw_mem((uint32_t *)(base+TOFF_GMAC_CONTROL), GMAC_CONTROL_TE);
}

/*
 *  EMAC送信処理
 *  parameter1 hmac  EMACハンドラへのポインタ
 *  parameter2 pktp  送信パケット構造体(LWIP型)
 *  return     E_OKで正常終了
 */
ER
emac_send(EMAC_Handle_t *hmac, struct emac_pkt *pktp)
{
	int      num, len, tlen, retry, i;
	int      start = 1;
	int      actlen = 0;
	ER       ercd = E_OK;
	volatile uint32_t *desc;
	uint32_t info_word = 0;
	volatile uint8_t  *pdata;
	volatile uint8_t  *s, *d;

	if(hmac == NULL || pktp == NULL)
		return E_PAR;

	desc = hmac->tx_cur_desc;
	s = pktp->payload;
	len = pktp->len;

	for(num = 0 ; num < hmac->Init.txquecount && pktp != NULL ; num++){
		if(pktp->len == 0)
			break;

		/*
		 *  送信データをバッファに書き込む
		 */
		pdata = d = (volatile uint8_t *)(desc[DESC_TXRX_BUFFER]);
		while(actlen < hmac->Init.paketbufsize){
			tlen = MIN(len, (hmac->Init.paketbufsize - actlen));
			for(i = 0 ; i < tlen ; i++)
				*d++ = *s++;
			actlen += tlen;
			len    -= tlen;
			if(len == 0){
				if((pktp = pktp->next) == NULL)
					break;
				if(pktp->len == 0)
					break;
				s = pktp->payload;
				len = pktp->len;
			}
		}

		CACHE_FLUSHINVALIDATE(pdata, actlen);
		syslog_4(LOG_DEBUG, "EMAC:emac_send(%d) desc[%08x] buf[%08x] actlen[%04x]", num, desc, pdata, actlen);

		/*
		 *  送信DESCRIPTORの開放待ち
		 */
		CACHE_INVALIDATE(desc, EMAC_QUE_SIZE);
		if((desc[DESC_TXRX_STATUS] & DESC_TXSTS_OWNBYDMA) != 0){
			for (retry = 0; retry < 3; retry++) {
				/* 送信ディスクリプタがフルのとき、開放されるのを待つ */
				dly_tsk(10);
				/* 再度チェックを行う */
				CACHE_INVALIDATE(desc, EMAC_QUE_SIZE);
				if((desc[DESC_TXRX_STATUS] & DESC_TXSTS_OWNBYDMA) == 0)
					break;
				/* 念のため、送信の開始を再要求する */
				emac_send_req(hmac, 0);
			}
			CACHE_INVALIDATE(desc, EMAC_QUE_SIZE);
			if((desc[DESC_TXRX_STATUS] & DESC_TXSTS_OWNBYDMA) != 0){
				syslog_1(LOG_ERROR, "EMAC:emac_send retry(%d) no descriptor", retry+1);
				ercd = E_TMOUT;
				break;
			}
		}

		/*
		 *  送信サイズと送信モードをセット
		 */
		desc[DESC_TXRX_CONTROL] = (actlen << DESC_TXCTRL_SIZE1SHFT) & DESC_TXCTRL_SIZE1MASK;

		info_word  = desc[DESC_TXRX_STATUS];
		info_word &= ~(DESC_TXSTS_MSK | DESC_TXSTS_TXFIRST | DESC_TXSTS_TXLAST);
		info_word |= DESC_TXSTS_TXINT;

		if (start) {
			start = 0;
			info_word |= DESC_TXSTS_TXFIRST;
		}
		/*
		 *  次のデータ確認と終了マーク設定
		 */
		if(pktp == NULL || pktp->len == 0)
			info_word |= DESC_TXSTS_TXLAST;

		desc[DESC_TXRX_STATUS] = info_word;
		CACHE_FLUSHINVALIDATE(desc, EMAC_QUE_SIZE);

		/*
		 *  次のDESCRIPTORへ更新する
		 */
		desc = (uint32_t *)(desc[DESC_TXRX_NEXT]);
	}

	/* パケットの終端までセットしていないときはエラー応答 */
	if((info_word & DESC_TXSTS_TXLAST) == 0){
		syslog_0(LOG_ERROR, "EMAC:emac_send last error !");
		return -1;
	}

	/* 送信を要求する */
	if(num > 0){
		hmac->tx_wait = 1;
		emac_send_req(hmac, num);
	}
	return ercd;
}

/*
 *  次の受信DESCRIPTORを取得する
 *  parameter1 hmac  EMACハンドラへのポインタ
 *  return     次のDESCRIPTORへのポインタ
 */
volatile static uint32_t *
emac_recv_desc_next(EMAC_Handle_t *hmac)
{
	volatile uint32_t *desc;
	uint8_t  *cdesc, *pdata;

	desc = hmac->rx_cur_desc;
	cdesc = (uint8_t *)desc;
	pdata = (uint8_t *)(desc[DESC_TXRX_BUFFER]);
	CACHE_INVALIDATE(cdesc, EMAC_QUE_SIZE);
	CACHE_INVALIDATE(pdata, hmac->Init.paketbufsize);
	desc[DESC_TXRX_CONTROL] = 
		(hmac->Init.paketbufsize & DESC_RXCTRL_SIZE1MASK) | \
			      DESC_RXCTRL_RXCHAIN;
	desc[DESC_TXRX_STATUS]  = DESC_RXSTS_OWNBYDMA;
	desc = (uint32_t *)(desc[DESC_TXRX_NEXT]);
	CACHE_FLUSHINVALIDATE(cdesc, EMAC_QUE_SIZE);
	hmac->rx_cur_desc = desc;
	return desc;
}

/*
 *  受信の先頭パケットのDESCRIPTORを探す
 *  parameter1 hmac  EMACハンドラへのポインタ
 *  return     NULLでなければ、受信の先頭DESCRIPTORへのポインタ
 */
volatile static uint32_t *
emac_recv_sop_find(EMAC_Handle_t *hmac)
{
	int      num;
	volatile uint32_t *desc = hmac->rx_cur_desc;
	uint32_t info_word, length;

	for(num = 0 ; num < hmac->Init.rxquecount ; num++) {
		CACHE_INVALIDATE(desc, EMAC_QUE_SIZE);
		info_word = desc[DESC_TXRX_STATUS];

		/*
		 *  受信データなしのときは、終了する
		 */
		if ((info_word & DESC_RXSTS_OWNBYDMA) != 0)
			break;

		/*
		 * 受信データあり場合、Start of Packet のチェックを行う
		 */
		length = (info_word & DESC_RXSTS_FRMLENMSK) >> DESC_RXSTS_FRMLENSHFT;
		if (info_word & DESC_RXSTS_RXFIRST) {
			if(length >= ETHER_HDR_SIZE)
				return desc;
		}
		else
			hmac->ifShortPkts++;

		/*
		 * 次のDESCRIPTORを取得する
		 */
		desc = emac_recv_desc_next(hmac);
	}
	return NULL;
}

/*
 *  受信データ長の取り出し
 *  parameter1 hmac  EMACハンドラへのポインタ
 *  return     0でなければ、受信パケット長
 */
int
emac_recv_length(EMAC_Handle_t *hmac)
{
	int      num, length = 0;
	volatile uint32_t *desc;
	uint32_t info_word;

	/*
	 *  受信パケットの確認し、ない場合は終了
	 */
	if((desc = emac_recv_sop_find(hmac)) == NULL)
		return 0;

	for(num = 0 ; num < hmac->Init.rxquecount ; num++){
		CACHE_INVALIDATE(desc, EMAC_QUE_SIZE);
		info_word = desc[DESC_TXRX_STATUS];

		/*  何も受信していないときは、終了する */
		if ((info_word & DESC_RXSTS_OWNBYDMA) != 0)
			break;

		/*  受信データ長を取得する */
		length += (desc[DESC_TXRX_STATUS] & DESC_RXSTS_FRMLENMSK) >> DESC_RXSTS_FRMLENSHFT;

		/*
		 *  End of Packetならば、終了
		 */
		if(info_word & DESC_RXSTS_RXLAST)
			break;

		/*
		 *  次のDESCRIPTORを取得する
		 */
		desc = (uint32_t *)(desc[DESC_TXRX_NEXT]);
	}
	return length;
}

/*
 *  受信パケットの取得
 *  parameter1 hmac  EMACハンドラへのポインタ
 *  parameter2 pktp  受信データ格納パケット構造体(LWIP型)
 *  return     E_OKならば、受信パケット長
 */
ER
emac_recv(EMAC_Handle_t *hmac, struct emac_pkt *pktp)
{
	int      num, actlen, i;
	volatile uint32_t *desc;
	volatile uint8_t  *b, *pdata;
	uint32_t info_word, len, tlen;

	if(pktp != NULL){
		b = pktp->payload;
		len = pktp->len;
	}
	else{
		b = NULL;
		len = 0;
	}

	sil_wrw_mem((uint32_t *)(hmac->base+TOFF_DMA_STATUS), DMA_STATUS_RI | DMA_INTER_RXSTAT_ERROR);

	/*
	 *  受信パケットの確認し、ない場合はエラー終了
	 */
	if((desc = emac_recv_sop_find(hmac)) == NULL)
		return E_NOMEM;

	for(num = 0 ; num < hmac->Init.rxquecount ; num++){
		CACHE_INVALIDATE(desc, EMAC_QUE_SIZE);
		info_word = desc[DESC_TXRX_STATUS];

		/*
		 *  受信データがない場合、終了する
		 */
		if((info_word & DESC_RXSTS_OWNBYDMA) != 0)
			break;

		/*
		 *  パケット格納領域がある場合は受信パケットを取得する
		 */
		if(b != NULL){
			/* 受信データを取得する */
			pdata = (uint8_t *)(desc[DESC_TXRX_BUFFER]);
			actlen = (desc[DESC_TXRX_STATUS] & DESC_RXSTS_FRMLENMSK) >> DESC_RXSTS_FRMLENSHFT;
			CACHE_INVALIDATE(pdata, actlen);
			syslog_5(LOG_DEBUG, "EMAC:emac_recv num(%d) desc[%08x] b[%08x] pdata[%08x] actlen[%08x] !", num, desc, b, pdata, actlen);
			syslog_4(LOG_DEBUG, "  12[%02x] 13[%02x] 14[%02x] 15[%02x]", pdata[12], pdata[13], pdata[14], pdata[15]);
			while(actlen > 0){
				tlen = MIN(len, actlen);
				for(i = 0 ; i < tlen ; i++)
					*b++ = *pdata++;
				actlen -= tlen;
				len    -= tlen;
				if(len == 0 /*&& actlen > 0*/){
					if((pktp = pktp->next) == NULL){
						b = NULL;
						break;
					}
					b = pktp->payload;
					if((len = pktp->len) == 0){
						b = NULL;
						break;
					}
				}
			}
		}
		/*
		 *  次のDESCRIPTORの取得と更新
		 */
		desc = emac_recv_desc_next(hmac);

		/*
		 *  End of Packetの場合、正常終了
		 */
		if((info_word & DESC_RXSTS_RXLAST) != 0)
			return E_OK;
	}
	return E_NOMEM;
}

/*
 *  DESCRIPTORのリセット
 *  parameter1 hmac  EMACハンドラへのポインタ
 */
static void
emac_desc_reset(EMAC_Handle_t *hmac)
{
	volatile uint32_t *desc;
	uint32_t base = hmac->base;
	int      i;

	/*
	 *  送信ディスクリプタの初期化
	 */
	desc = hmac->tx_top_desc;
	hmac->tx_cur_desc = desc;
	for(i = 0 ; i < hmac->Init.txquecount ; i++) {
		desc[DESC_TXRX_STATUS] &= ~(DESC_TXSTS_TXINT | DESC_TXSTS_TXLAST |
				DESC_TXSTS_TXFIRST | DESC_TXSTS_TXCRCDIS | \
				DESC_TXSTS_TXCHECKINSCTRL | \
				DESC_TXSTS_TXRINGEND | DESC_TXSTS_TXPADDIS);
		desc[DESC_TXRX_STATUS] |= DESC_TXSTS_TXCHAIN;
		desc[DESC_TXRX_CONTROL] = 0;
		desc[DESC_TXRX_STATUS] &= ~(DESC_TXSTS_MSK | DESC_TXSTS_OWNBYDMA);
		desc += DESC_TXRX_SIZE;
	}
	CACHE_FLUSHINVALIDATE(hmac->tx_top_desc, (hmac->Init.txquecount * EMAC_QUE_SIZE));

	/*
	 *  受信ディスクリプタの初期化
	 */
	desc = hmac->rx_top_desc;
	hmac->rx_cur_desc = desc;
	for (i = 0 ; i < hmac->Init.rxquecount ; i++) {
		desc[DESC_TXRX_CONTROL] = 
			(hmac->Init.paketbufsize & DESC_RXCTRL_SIZE1MASK) | DESC_RXCTRL_RXCHAIN;
		desc[DESC_TXRX_STATUS] = DESC_RXSTS_OWNBYDMA;
		desc += DESC_TXRX_SIZE;
	}
	CACHE_FLUSHINVALIDATE(hmac->rx_top_desc, (hmac->Init.rxquecount * EMAC_QUE_SIZE));

	/* Descriptor Address のセット */
	sil_wrw_mem((uint32_t *)(base+TOFF_DMA_TX_BASE_ADDR), (uint32_t)hmac->tx_top_desc);
	sil_wrw_mem((uint32_t *)(base+TOFF_DMA_RX_BASE_ADDR), (uint32_t)hmac->rx_top_desc);
}

/*
 *  ETHERNET通信を開始する
 *  parameter1 hmac  EMACハンドラへのポインタ
 *  parameter2 link  通信モード
 *  return     E_OKで正常終了
 */
ER
emac_start(EMAC_Handle_t *hmac, uint32_t link)
{
	uint32_t   base;

	if(hmac == NULL)
		return E_PAR;
	base = hmac->base;

	/*
	 *  PHYスタート
	 */
	emac_phy_start(hmac, link);

	/*
	 *  送受信を有効設定
	 */
	sil_orw_mem((uint32_t *)(base+TOFF_DMA_CONTROL), DMA_CONTROL_SR);

	/*
	 *  割り込みマスク設定
	 */
	sil_wrw_mem((uint32_t *)(base+TOFF_GMAC_INT_MASK), GMAC_INT_DEFAULT_MASK);
	sil_orw_mem((uint32_t *)(base+TOFF_DMA_INTR_ENA), DMA_INTR_DEFAULT_MASK);

	/*
	 *  割り込み許可
	 */
	ena_int(hmac->intno);

	sil_wrw_mem((uint32_t *)(base+TOFF_DMA_STATUS), 0x0001FFFF);
	sil_orw_mem((uint32_t *)(base+TOFF_DMA_CONTROL), DMA_CONTROL_ST);
	sil_orw_mem((uint32_t *)(base+TOFF_GMAC_CONTROL), (GMAC_CONTROL_TE | GMAC_CONTROL_RE));
	return E_OK;
}

/*
 *  ETHERNET通信を開始する
 *  parameter1 hmac  EMACハンドラへのポインタ
 *  parameter2 disable 送受信の無効化設定(1で無効化)
 *  return     E_OKで正常終了
 */
ER
emac_stop(EMAC_Handle_t *hmac, int disable)
{
	uint32_t base;
	ER       ercd = E_OK;

	if(hmac == NULL)
		return E_PAR;
	base = hmac->base;

	/*
	 * 割り込み停止
	 */
	dis_int(hmac->intno);
	sil_wrw_mem((uint32_t *)(base+TOFF_DMA_INTR_ENA), 0);

	if(disable == 1) {
		/*
		 *  送受信を無効に設定
		 */
		sil_andw_mem((uint32_t *)(base+TOFF_GMAC_CONTROL), GMAC_CONTROL_RE);
		dly_tsk(200);

		/*
		 *  受信済みパケットを破棄
		 */
		while(ercd == E_OK){
			ercd = emac_recv(hmac, NULL);
			dly_tsk(1);
			syslog_1(LOG_NOTICE, "EMAC:emac_stop flush result(%d) !", ercd);
		}
		sil_andw_mem((uint32_t *)(base+TOFF_DMA_CONTROL), DMA_CONTROL_SR);
		sil_andw_mem((uint32_t *)(base+TOFF_GMAC_CONTROL), GMAC_CONTROL_TE);
		sil_andw_mem((uint32_t *)(base+TOFF_DMA_CONTROL), DMA_CONTROL_ST);
	}
	return E_OK;
}

/*
 *  LINK状態を取得する
 *  parameter1 hmac  EMACハンドラへのポインタ
 *  return     LINK状態(0で未接続)
 */
uint32_t
emac_link_detect(EMAC_Handle_t *hmac)
{
	uint32_t link_detect = 0;
	int      bmsr;

	/* 最初は空読み */
	emac_mdio_read(hmac, MII_BMSR);
	bmsr = emac_mdio_read(hmac, MII_BMSR);
	syslog_2(LOG_INFO, "EMAC:phy_link_detect[%08x] phydev->auto_nego(%d)", bmsr, is_autonego(hmac));

	if ((bmsr >= 0) && (bmsr & BMSR_LSTATUS)) {
		link_detect = is_autonego(hmac)
				? emac_phy_link_autonego(hmac)
				: emac_phy_link_manual(hmac);
	}
	return link_detect;
}

/*
 *  EMACの制御設定を行う
 *  parameter1 hmac  EMACハンドラへのポインタ
 *  parameter2 link  EMAC設定モード
 *  return     E_OKで正常終了
 */
ER
emac_link_mode_set(EMAC_Handle_t *hmac, uint32_t link)
{
	uint32_t base, conf;

	if(hmac == NULL || link == 0)
		return E_PAR;
	base = hmac->base;
	conf = sil_rew_mem((uint32_t *)(base+TOFF_GMAC_CONTROL));

	if((link & EMAC_SPEED_MASK) != EMAC_SPEED_1000BASE){
		conf |= GMAC_CONTROL_PS;
		if((link & EMAC_SPEED_MASK) != EMAC_SPEED_10BASE)
			conf |= GMAC_CONTROL_FES;
		else
			conf &= ~GMAC_CONTROL_FES;
	}
	else
		conf &= ~GMAC_CONTROL_PS;

	if((link & EMAC_DUPLEX_MASK) == EMAC_DUPLEX_FULL)
		conf |= GMAC_CONTROL_DM;
	else
		conf &= ~GMAC_CONTROL_DM;
	sil_wrw_mem((uint32_t *)(base+TOFF_GMAC_CONTROL), conf);
	return E_OK;
}

/*
 *  EMACをリセットする
 *  parameter1 hmac   EMACハンドラへのポインタ
 *  parameter2 macadr MACアドレスデータへのポインタ
 *  return     E_OKで正常終了
 */
ER
emac_reset(EMAC_Handle_t *hmac, uint8_t *mac_addr)
{
	uint32_t  timeout = 3000;
	uint32_t  base, conf;
	uint32_t  macid_hi, macid_lo;
	uint8_t   *b = (uint8_t *)&mac_addr[0];

	if(hmac == NULL || mac_addr == NULL)
		return E_PAR;
	base = hmac->base;

	/*
	 *  MACをリセット
	 */
	sil_wrw_mem((uint32_t *)(base+TOFF_DMA_BUS_MODE), DMA_BUS_MODE_SRST);
	conf = GMAC_CORE_INIT;
	if(hmac->Init.paketbufsize > 1500)
		conf |= GMAC_CONTROL_2K;
	sil_wrw_mem((uint32_t *)(base+TOFF_GMAC_CONTROL), conf);
	while(timeout > 0){
		if((sil_rew_mem((uint32_t *)(base+TOFF_DMA_BUS_MODE)) & DMA_BUS_MODE_SRST) == 0)
			break;
		dly_tsk(1);
		timeout--;
	}

	/*
	 *  送受信の停止
	 */
	emac_stop(hmac, 0);

	/*
	 *  MACアドレスの登録
	 */
	macid_lo = b[0] + (b[1] << 8) + (b[2] << 16) + (b[3] << 24);
	macid_hi = b[4] + (b[5] << 8);
	sil_wrw_mem((uint32_t *)(base+TOFF_GMAC_ADDR_HIGH), macid_hi);
	sil_wrw_mem((uint32_t *)(base+TOFF_GMAC_ADDR_LOW),  macid_lo);

	/*
	 *  MACの設定
	 */
	sil_wrw_mem((uint32_t *)(base+TOFF_DMA_BUS_MODE), (DMA_BUS_MODE_FIXEDBURST | DMA_BUS_MODE_PRIORXTX_41 | DMA_BUS_MODE_BURST_16));
	sil_orw_mem((uint32_t *)(base+TOFF_DMA_CONTROL), (DMA_CONTROL_FTF |  DMA_CONTROL_STOREFORWARD |DMA_CONTROL_TXSECONDFRAME));

	/*
	 *  送受信DESCRIPTERリセット
	 */
	emac_desc_reset(hmac);

	sil_orw_mem((uint32_t *)(base+TOFF_DMA_CONTROL), DMA_CONTROL_SR);
	sil_orw_mem((uint32_t *)(base+TOFF_DMA_CONTROL), DMA_CONTROL_ST);
	sil_orw_mem((uint32_t *)(base+TOFF_GMAC_CONTROL), (GMAC_CONTROL_RE | GMAC_CONTROL_TE));
	return E_OK;
}

/*
 *  EMAC用割込み割り込み処理
 */
static void
emac_interrupt(EMAC_Handle_t *hmac, uint32_t status)
{
	uint32_t base = hmac->base;
	uint32_t tmp;

	/*
	 *  MMC割り込み：未使用
	 */
	if((status & GMAC_INT_STATUS_MMCTIS) != 0)
		hmac->mmc_tx_irq_n++;
	if((status & GMAC_INT_STATUS_MMCRIS) != 0)
		hmac->mmc_rx_irq_n++;
	if((status & GMAC_INT_STATUS_MMCCSUM) != 0)
		hmac->mmc_rx_csum_offload_irq_n++;
	if((status & GMAC_INT_STATUS_PMT) != 0) {
		/*
		 *  PMT ビット5,6クリア(reading the PMT status regによる)
		 */
		tmp = sil_rew_mem((uint32_t *)(base+TOFF_GMAC_PMT));
		hmac->irq_receive_pmt_irq_n++;
	}

	/*
	 *  MAC TX/RX EEE LPI割り込み
	 */
	if((status & GMAC_INT_STATUS_LPIIS) != 0) {
		/*
		 *  LPI割り込みクリア
		 */
		tmp = sil_rew_mem((uint32_t *)(base+TOFF_LPI_CTRL_STATUS));

		if(tmp & LPI_CTRL_STATUS_TLPIEN)
			hmac->irq_tx_path_in_lpi_mode_n++;
		if(tmp & LPI_CTRL_STATUS_TLPIEX)
			hmac->irq_tx_path_exit_lpi_mode_n++;
		if(tmp & LPI_CTRL_STATUS_RLPIEN)
			hmac->irq_rx_path_in_lpi_mode_n++;
		if(tmp & LPI_CTRL_STATUS_RLPIEX)
			hmac->irq_rx_path_exit_lpi_mode_n++;
	}
}

/*
 *  EMAC DMA用割り込み処理
 */
static int
emac_dma_interrupt(EMAC_Handle_t *hmac, uint32_t status)
{
	int  event = 0;

	if((status & 0x1FFFF) == 0)
		return 0;
	event = 0;

	/*
	 *  ABNORMAL割り込み
	 */
	if((status & DMA_STATUS_AIS) != 0) {
		if ((status & DMA_STATUS_UNF) != 0) {
			hmac->tx_undeflow_irq++;
			event |= 0x0001;
		}
		if((status & DMA_STATUS_TJT) != 0)
			hmac->tx_jabber_irq++;
		if((status & DMA_STATUS_OVF) != 0)
			hmac->rx_overflow_irq++;
		if((status & DMA_STATUS_RU) != 0)
			hmac->rx_buf_unav_irq++;
		if((status & DMA_STATUS_RPS) != 0)
			hmac->rx_process_stopped_irq++;
		if((status & DMA_STATUS_RWT) != 0)
			hmac->rx_watchdog_irq++;
		if((status & DMA_STATUS_ETI) != 0)
			hmac->tx_early_irq++;
		if((status & DMA_STATUS_TPS) != 0) {
			hmac->tx_process_stopped_irq++;
			event |= 0x0002;
		}
		if ((status & DMA_STATUS_FBI) != 0) {
			hmac->fatal_bus_error_irq++;
			event |= 0x0004;
		}
	}

	/*
	 *  TX/RX NORMAL割り込み
	 */
	if((status & DMA_STATUS_NIS) != 0){
		hmac->normal_irq_n++;
		if((status & DMA_STATUS_RI) != 0) {
			uint32_t value = sil_rew_mem((uint32_t *)(hmac->base+TOFF_DMA_INTR_ENA));
			/* to schedule NAPI on real RIE event. */
			if((value & DMA_INTR_ENA_RIE) != 0){
				hmac->rx_normal_irq_n++;
				if(hmac->emacevent != NULL)
					hmac->emacevent(hmac, 0, RECV_EVENT_DETECT);
			}
		}
		if((status & DMA_STATUS_TI) != 0){
			/*
			 * 送信割り込みを通知する
			 */
			if(hmac->tx_wait != 0){
				if(hmac->emacevent != NULL)
					hmac->emacevent(hmac, 0, SEND_EVENT_DETECT);
				hmac->tx_wait = 0;
			}
		}
		if((status & DMA_STATUS_ERI) != 0)
			hmac->rx_early_irq++;
	}
	/* Optional hardware blocks, interrupts should be disabled */
	if((status &
		     (DMA_STATUS_GPI | DMA_STATUS_GMI | DMA_STATUS_GLI)) != 0)
		syslog_2(LOG_ERROR, "%s: unexpected status %08x\n", __func__, status);
	return event;
}

/*
 *  EMAC割り込みサービスルーチン
 */
void
emac_hps_isr(intptr_t exinf)
{
	EMAC_Handle_t *hmac = &EMacHandle[INDEX_EMAC((uint32_t)exinf)];
	uint32_t base, istatus, dstatus, imask;
	uint32_t event;

	base = hmac->base;
	istatus = sil_rew_mem((uint32_t *)(base+TOFF_GMAC_INT_STATUS));
	dstatus = sil_rew_mem((uint32_t *)(base+TOFF_DMA_STATUS));
	imask   = sil_rew_mem((uint32_t *)(base+TOFF_GMAC_INT_MASK));

	if((istatus & ~imask) != 0){
		emac_interrupt(hmac, istatus);
	}
	if(dstatus != 0){
		event = emac_dma_interrupt(hmac, dstatus);
		sil_wrw_mem((uint32_t *)(base+TOFF_DMA_STATUS), (dstatus & 0x0001FFFF));
		/*
		 *  DMAエラーがあった場合、通知
		 */
		if(event != 0)
			syslog_2(LOG_ERROR, "EMAC:emac_isr id(%d) DMA error[%08x] !", (int)exinf, event);
	}
}

