/*
 *  TOPPERS/FMP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Flexible MultiProcessor Kernel
 * 
 *  Copyright (C) 2008-2011 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2016-2018 by TOPPERS PROJECT Educational Working Group.
 * 
 *  上記著作権者は，以下の(1)～(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 * 
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 * 
 *  @(#) $Id: i2c.c 698 2018-05-22 20:23:32Z roi $
 */
/*
 * 
 *  I2Cドライバ関数群
 *
 */
#include "kernel_impl.h"
#include <t_syslog.h>
#include <t_stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sil.h>
#include <target_syssvc.h>
#include "i2c.h"

#define sil_orw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) | (b))
#define sil_andw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) & ~(b))
#define sil_modw_mem(a, b, c)	sil_wrw_mem((a), (sil_rew_mem(a) & (~b)) | (c))

/*
 *  I2CポートIDから管理ブロックを取り出すためのマクロ
 */
#define INDEX_I2C(i2cid)            ((uint_t)((i2cid) - 1))

#define MIN(a, b) ((a) > (b) ? (b) : (a))

#ifndef EXT_CLK_CLKSC1_FREQCUR
#define EXT_CLK_CLKSC1_FREQCUR                 25000000
#endif
#ifndef EXT_CLK_CLKSC2_FREQCUR
#define EXT_CLK_CLKSC2_FREQCUR	               25000000
#endif
#ifndef EXT_CLK_PERIPH_FREQCUR
#define EXT_CLK_PERIPH_FREQCUR                 0
#endif

/*
 *  CLOCK SOURCE
 */
#define CLKMGR_PERPLL_VCO_PSRC_E_EOSC1          0x0	/* eosc1_clk */
#define CLKMGR_PERPLL_VCO_PSRC_E_EOSC2          0x1	/* eosc2_clk */
#define CLKMGR_PERPLL_VCO_PSRC_E_F2S_PERIPH_REF 0x2	/* f2s_periph_ref_clk */

#define CLKMGR_MAINPLL_MAINDIV_L4SPCLK_E_DIV16  0x4


#define I2C_DEFAULT_INT_STARUS  (I2C_INTR_STAT_RX_UNDER | I2C_INTR_STAT_RX_OVER | I2C_INTR_STAT_TX_OVER | I2C_INTR_STAT_R_TX_ABRT | I2C_INTR_STAT_RX_FULL)

#define I2C_RESET_TMO_INIT      100			/* I2Cリセット実行タイミング設定 */
#define I2C_MAX_T_POLL_COUNT    100			/* I2Cディーブルタイミング設定 */

#define I2C_SS_MIN_SPEED        8000		/* SS最小クロック値 */
#define I2C_SS_MAX_SPEED        100000		/* SS最大クロック値 */
#define I2C_FS_MIN_SPEED        100000		/* FS最小クロック値 */
#define I2C_FS_MAX_SPEED        400000		/* FS最大クロック値 */

#define I2C_DIFF_LCNT_HCNT      70			/* SCL-LCNTとSCL-HCLの差分 */

#define I2C_SLV_RESERVE_ADDR_S_1 0x00		/* スレーブ予約スレーブアドレス1最小値 */
#define I2C_SLV_RESERVE_ADDR_F_1 0x0E		/* スレーブ予約スレーブアドレス1最大値 */
#define I2C_SLV_RESERVE_ADDR_S_2 0xF0		/* スレーブ予約スレーブアドレス2最小値 */
#define I2C_SLV_RESERVE_ADDR_F_2 0xFE		/* スレーブ予約スレーブアドレス2最大値 */

/*
 *  I2Cポート設定テーブル
 */
static const I2C_PortControlBlock i2c_pcb[NUM_I2CPORT] = {
  {	TADR_I2C1_BASE, INTNO_I2C0_IRQ, RSTMGR_PERMODRST_I2C0 },
  {	TADR_I2C2_BASE, INTNO_I2C1_IRQ, RSTMGR_PERMODRST_I2C1 },
  {	TADR_I2C3_BASE, INTNO_I2C2_IRQ, RSTMGR_PERMODRST_I2C2 },
  {	TADR_I2C4_BASE, INTNO_I2C3_IRQ, RSTMGR_PERMODRST_I2C3 }
};

/*
 * I2C ハンドラ
 */
static I2C_Handle_t I2cHandle[NUM_I2CPORT];


/*
 *  I2Cクロック有効判定
 */
static bool_t
i2c_clk_is_enabled(void)
{
	if((sil_rew_mem((uint32_t *)(TADR_CLKMGR_MPLL_BASE+TOFF_CLKMGR_MPLL_EN)) & CLKMGR_MPLL_EN_L4SPCLK) != 0)
		return true;
	else
		return false;
}

/*
 *  I2Cクロック取得
 */
static uint32_t
i2c_clock(void)
{
	uint32_t temp = 0;
	uint64_t temp1 = 0;
	uint64_t numer = 0;
	uint64_t denom = 1;
	uint64_t denom2 = 1;

	temp = (sil_rew_mem((uint32_t *)(TADR_CLKMGR_MPLL_BASE+TOFF_CLKMGR_MPLL_MAINDIV)) & CLKMGR_MPLL_MDV_L4SPCLK) >> 7;
	if (temp <= CLKMGR_MAINPLL_MAINDIV_L4SPCLK_E_DIV16)
		temp <<= 1;
	else
		return 0;

	denom = (uint64_t) temp;
	if((sil_rew_mem((uint32_t *)(TADR_CLKMGR_MPLL_BASE+TOFF_CLKMGR_MPLL_L4SRC)) & CLKMGR_MPLL_L4SRC_L4SP) == 0){
		uint64_t numer2;
		temp = sil_rew_mem((uint32_t *)(TADR_CLKMGR_MGR_BASE+TOFF_CLKMGR_BYPASS));
		if((temp & CLKMGR_BYPASS_MAINPLL) != 0)
			numer = (uint64_t)EXT_CLK_CLKSC1_FREQCUR;
		else{
			temp = sil_rew_mem((uint32_t *)(TADR_CLKMGR_MPLL_BASE+TOFF_CLKMGR_MPLL_VCO));
			numer2 = (temp & CLKMGR_MPLL_VCO_NUMER) >> 3;
			denom2 = (temp & CLKMGR_MPLL_VCO_DENOM) >> 16;
			temp1 = (uint64_t) EXT_CLK_CLKSC1_FREQCUR;
			temp1 *= (numer2 + 1);
			temp1 /= (denom2 + 1);

			if(temp1 <= UINT32_MAX)
				numer = temp1;
			else
				return 0;
		}

		temp = (((sil_rew_mem((uint32_t *)(TADR_CLKMGR_MPLL_BASE+TOFF_CLKMGR_MPLL_MAINCLK)) & CLKMGR_MPLL_MAINCLK_CNT) >> 7) +1) *
						((sil_rew_mem((uint32_t *)(TADR_CLKMGR_MISC_MAINCLK_BASE+TOFF_CLKMGR_MISC_MAINCLK)) & CLKMGR_MISC_MAINCLK_CNT) + 1);
		denom = denom * (uint64_t) temp;
	}
	else{
		temp = sil_rew_mem((uint32_t *)(TADR_CLKMGR_MGR_BASE+TOFF_CLKMGR_BYPASS));
		if((temp & CLKMGR_BYPASS_PERPLL) != 0){
			if((temp & CLKMGR_BYPASS_PERPLLSRC) == 0)
				temp = EXT_CLK_CLKSC1_FREQCUR;
			else
				temp = EXT_CLK_PERIPH_FREQCUR;
		}
		else{
			temp = sil_rew_mem((uint32_t *)(TADR_CLKMGR_PERPLL_BASE+TOFF_CLKMGR_PERPLL_VCO));
			numer = (temp & CLKMGR_PERPLL_VCO_NUMER) >> 3;
			denom2 = (temp & CLKMGR_PERPLL_VCO_DENOM) >> 16;
			temp = (temp & CLKMGR_PERPLL_VCO_PSRC) >> 22;
			if (temp == CLKMGR_PERPLL_VCO_PSRC_E_EOSC1)
				temp1 = (uint64_t)EXT_CLK_CLKSC1_FREQCUR;
			else if (temp == CLKMGR_PERPLL_VCO_PSRC_E_EOSC2)
				temp1 = (uint64_t)EXT_CLK_CLKSC2_FREQCUR;
			else if (temp == CLKMGR_PERPLL_VCO_PSRC_E_F2S_PERIPH_REF)
				temp1 = (uint64_t)EXT_CLK_PERIPH_FREQCUR;

			if(temp1 != 0){
				temp1 *= (numer + 1);
				temp1 /= (denom2 + 1);
				if (temp1 <= UINT32_MAX)
					temp = (uint32_t)temp1;
				else
					return 0;
			}
		}
		numer = (uint64_t)temp;
		temp = (sil_rew_mem((uint32_t *)(TADR_CLKMGR_PERPLL_BASE+TOFF_CLKMGR_PERPLL_PERBASECLK)) & CLKMGR_PERPLL_PERBASECLK_CNT) + 1;
		denom = denom * (uint64_t)temp;
	}

	if(denom > 0){
		numer /= denom;
		if(numer <= UINT32_MAX)
			return (uint32_t)numer;
	}
    return 0;
}

/*
 *  I2Cモジュールリセット
 */
static ER
i2c_rstmgr_reset(I2C_Handle_t *hi2c, uint32_t timeout)
{
    uint32_t rst_mask = i2c_pcb[INDEX_I2C(hi2c->PortID)].resetbit;

	sil_orw_mem((uint32_t *)(TADR_RSTMGR_BASE+TOFF_RSTMGR_CTL), rst_mask);
	if(timeout == 0)
		return E_OK;

    /*
	 *  リセット時間待ち
	 */
    while (--timeout){
		sil_dly_nse(1000);
	}

	sil_andw_mem((uint32_t *)(TADR_RSTMGR_BASE+TOFF_RSTMGR_CTL), rst_mask);
	return E_OK;
}

/*
 *  TX FIFOレベル取得
 */
static uint32_t
i2c_tx_fifo_level_get(I2C_Handle_t *hi2c)
{
	return (sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_TXFLR)) & I2C_TXFLR_MSK);
}

/*
 *  RX FIFOレベル取得
 */
static uint32_t
i2c_rx_fifo_level_get(I2C_Handle_t *hi2c)
{
	return (sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_RXFLR)) & I2C_RXFLR_MSK);
}

/*
 *  マスタースレーブアドレス設定
 */
static ER
i2c_master_target_set(I2C_Handle_t *hi2c, uint16_t target_addr)
{
	uint32_t timeout = 200;

	while(hi2c->XferSize != 0){
		dly_tsk(1);
		if(--timeout == 0)
			return E_TMOUT;
	}

	if(target_addr != hi2c->last_target){
		while((sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_STAT)) & I2C_STAT_TFE) == 0){
			dly_tsk(1);
			if(--timeout == 0)
				return E_TMOUT;
		}

        /*
		 * ターゲットスレーブアドレス設定
		 */
		sil_modw_mem((uint32_t *)(hi2c->base+TOFF_I2C_TAR), I2C_TAR_IC_TAR, target_addr);
		hi2c->last_target = target_addr;
    }
	return E_OK;
}

/*
 *  I2Cモジュールイネーブルチェック
 */
static bool_t
i2c_is_enabled(I2C_Handle_t * hi2c)
{
	if((sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_EN_STAT)) & 1) != 0)
		return true;
    else
		return false;
}

/*
 *  I2Cモジュールイネーブル
 */
static void
i2c_enable(I2C_Handle_t *hi2c)
{
    /*
	 *  DMA設定有効
	 */
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DMA_CR), (I2C_DMA_CR_RDMAE | I2C_DMA_CR_TDMAE));

	sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_EN), I2C_EN_EN_SET);
}

/*
 *  I2Cモジュールディセーブル
 */
static ER
i2c_disable(I2C_Handle_t *hi2c)
{
    uint32_t timeout = I2C_MAX_T_POLL_COUNT;
	uint32_t temp;

	if(!i2c_is_enabled(hi2c))
		return E_OK;

    /*
	 *  I2C有効設定クリア
	 */
	sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_EN), I2C_EN_EN_SET);

	/*
	 *  I2Cディセーブル設定待ち
	 */
	while (i2c_is_enabled(hi2c)){
		sil_dly_nse(1000);
		if (--timeout == 0){
			return E_TMOUT;
		}
	}

	/*
	 *  割込みクリア
	 */
	temp = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_CLR_INTR));
	(void)(temp);
	return E_OK;
}


/*
 *  I2Cの要因別割込み許可
 *  parameter1 hi2c  I2Cハンドラへのポインタ
 *  parameter2 mask  I2C割込み要因ビットマップ
 *  return           正常終了時、E_OK
 */
ER
i2c_int_enable(I2C_Handle_t *hi2c, const uint32_t mask)
{
    if(hi2c == NULL)
		return E_PAR;
	sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_INTR_MSK), mask);
	return E_OK;
}

/*
 *  I2Cの要因別割込み禁止
 *  parameter1 hi2c  I2Cハンドラへのポインタ
 *  parameter2 mask  I2C割込み要因ビットマップ
 *  return           正常終了時、E_OK
 */
ER
i2c_int_disable(I2C_Handle_t *hi2c, const uint32_t mask)
{
    if(hi2c == NULL)
		return E_PAR;
	sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_INTR_MSK), mask);
	return E_OK;
}


/*
 *  I2Cデバイスの初期化
 *  parameter1 hi2c  I2Cハンドラへのポインタ
 *  parameter2 ii2c  初期化構造体へのポインタ
 *  return           ハンドラへのポインタ
 */
I2C_Handle_t *
i2c_init(ID port, I2C_Init_t *ii2c)
{
	I2C_Handle_t *hi2c;
	ER ercd = E_OK;
	uint32_t no;

	if(port < I2C1_PORTID || port > NUM_I2CPORT)
		return NULL;
	if(ii2c == NULL)
		return NULL;

	if(ii2c->OwnAddress1 == 0){
		if(ii2c->ClockSpeed > I2C_FS_MAX_SPEED || ii2c->ClockSpeed < I2C_SS_MIN_SPEED)
			return NULL;
	}
	else{
		if ((ii2c->OwnAddress1 > I2C_SAR_IC_SAR)
    	    || (ii2c->OwnAddress1 < I2C_SLV_RESERVE_ADDR_F_1)
        	|| ((ii2c->OwnAddress1 > I2C_SLV_RESERVE_ADDR_S_2) && (ii2c->OwnAddress1 < I2C_SLV_RESERVE_ADDR_F_2))){
			return NULL;
	    }
	}

	if(ii2c->AddressingMode != I2C_ADDRESSINGMODE_7BIT && ii2c->AddressingMode != I2C_ADDRESSINGMODE_10BIT)
		return NULL;

    /*
	 *  I2Cハンドラ設定
	 */
	no = INDEX_I2C(port);
	hi2c = &I2cHandle[no];
	if(hi2c->Status != I2C_STATUS_RESET)
		return NULL;
	hi2c->base = i2c_pcb[no].base;

	if(!i2c_clk_is_enabled()){
		syslog_0(LOG_ERROR, "i2c_init:i2c clock disable !");
		return NULL;
	}

	hi2c->clock_freq = i2c_clock();
	if(hi2c->clock_freq == 0){
		syslog_1(LOG_ERROR, "i2c_init:i2c clock error(%d) !", hi2c->clock_freq);
		return NULL;
	}

	memcpy(&hi2c->Init, ii2c, sizeof(I2C_Init_t));
	hi2c->PortID = port;

	hi2c->Status = I2C_STATUS_READY;
    /*
	 * I2Cモジュールリセット
	 */
	ercd = i2c_reset(hi2c);

    /*
	 *  I2Cモジュール有効化、割込み許可
	 */
    if(ercd == E_OK){
		i2c_enable(hi2c);
		i2c_int_enable(hi2c, I2C_DEFAULT_INT_STARUS);
    }
	return hi2c;
}

/*
 *  I2Cデバイスの無効化
 *  parameter1 hi2c  I2Cハンドラへのポインタ
 *  return           正常終了時、E_OK
 */
ER
i2c_deinit(I2C_Handle_t * hi2c)
{
	ER ercd = E_OK;

	if(hi2c == NULL){
		return E_PAR;
	}

    /*
	 *  I2Cモジュール無効化
	 */
	ercd = i2c_disable(hi2c);

    /*
	 * I2Cモジュールリセット
	 */
    if(ercd == E_OK){
		ercd = i2c_rstmgr_reset(hi2c, 0);
	}
	hi2c->Status = I2C_STATUS_RESET;
    return ercd;
}


/*
 *  I2Cデバイスのリセット
 *  parameter1 hi2c  I2Cハンドラへのポインタ
 *  return           正常終了時、E_OK
 */
ER
i2c_reset(I2C_Handle_t * hi2c)
{
	ER ercd = E_OK;
	bool_t already_enabled = false;
	uint32_t temp;

	if(hi2c->Status == I2C_STATUS_RESET)
		return E_OK;

    already_enabled = i2c_is_enabled(hi2c);
    if (already_enabled){
        /* Temporarily disable controller */
        ercd = i2c_disable(hi2c);
        if (ercd != E_OK)
        {
            return ercd;
        }
    }

    /*
	 *  モジュールマネージャによるI2Cリセット
	 */
    i2c_rstmgr_reset(hi2c, I2C_RESET_TMO_INIT);

	if(hi2c->Init.OwnAddress1 == 0){	/* setup master */
		uint32_t scl_lcnt, scl_hcnt, speed_mode;

		if(hi2c->Init.ClockSpeed > I2C_FS_MIN_SPEED)
			speed_mode = I2C_CON_SPEED_FAST;
		else
			speed_mode = I2C_CON_SPEED_STANDARD;

	    /*
		 *  LCOUNT値を計算
		 */
	    scl_lcnt = hi2c->clock_freq / (hi2c->Init.ClockSpeed << 1);

		/*
		 *  HCOUNT値を計算
		 */
		scl_hcnt = scl_lcnt - I2C_DIFF_LCNT_HCNT;

		/*
		 *  マスターモードのレジスタ設定
		 */
		sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CON), (I2C_CON_IC_SLV_DIS | I2C_CON_MST_MOD));
		sil_modw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CON), I2C_CON_SPEED, speed_mode);
		sil_modw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CON), I2C_CON_IC_RESTART_EN, hi2c->Init.restart_enable);
		sil_modw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CON), I2C_CON_IC_10BITADDR_MST, (hi2c->Init.AddressingMode<<4));
		sil_modw_mem((uint32_t *)(hi2c->base+TOFF_I2C_TAR), I2C_TAR_IC_10BITADDR, (hi2c->Init.AddressingMode<<12));
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_FS_SPKLEN), (hi2c->Init.fs_spklen & I2C_FS_SPKLEN_MSK));
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_SS_SCL_LCNT), (scl_lcnt & I2C_SS_SCL_LCNT_MSK));
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_SS_SCL_HCNT), (scl_hcnt & I2C_SS_SCL_HCNT_MSK));
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_FS_SCL_LCNT), (scl_lcnt & I2C_FS_SCL_LCNT_MSK));
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_FS_SCL_HCNT), (scl_hcnt & I2C_FS_SCL_HCNT_MSK));

		/*
		 *  General Call ACK設定
		 */
		if(hi2c->Init.GeneralCallMode == 0)
			sil_modw_mem((uint32_t *)(hi2c->base+TOFF_I2C_TAR), (I2C_TAR_GC_OR_START | I2C_TAR_SPECIAL), I2C_TAR_GC_OR_START);
		else
			sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_TAR), (I2C_TAR_GC_OR_START | I2C_TAR_SPECIAL));
	}
	else{
		/*
		 *  スレーブモードのレジスタ設定
		 */
		sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CON), (I2C_CON_IC_SLV_DIS | I2C_CON_MST_MOD));
		sil_modw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CON), I2C_CON_IC_10BITADDR_SLV, (hi2c->Init.AddressingMode<<3));
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_SAR), ((hi2c->Init.OwnAddress1/2) & I2C_SAR_IC_SAR));
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_SLV_DATA_NACK), (hi2c->Init.nack_enable & I2C_SLV_DATA_NACK_ONLY));
	}

	/*
	 *  SDAホールド時間設定
	 */
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_SDA_HOLD), (hi2c->Init.SdaHoldClock & I2C_SDA_HOLD_IC_SDA));

    /*
	 *  ターゲットスレーブアドレス保存値をリセット
	 */
    hi2c->last_target = 0xffff;

    /*
	 *  割込みマスク解除と割込みクリア
	 */
	sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_INTR_MSK), I2C_INTR_STAT_ALL);
	temp = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_CLR_INTR));
	(void)(temp);

	/*
	 *  I2Cモジュール有効化と割込み設定
	 */
    if(already_enabled){
        i2c_enable(hi2c);
    	i2c_int_enable(hi2c, I2C_DEFAULT_INT_STARUS);
    }
	hi2c->Status = I2C_STATUS_READY;
	hi2c->ErrorCode = I2C_ERROR_NONE;
    return ercd;
}

/*
 *  I2CモジュールのBUSYチェック
 *  parameter1 hi2c:   I2Cハンドラへのポインタ
 *  return             BUSYならtrue
 */
bool_t
i2c_is_busy(I2C_Handle_t *hi2c)
{
    if(hi2c == NULL || hi2c->Status != I2C_STATUS_READY)
		return true;

	if((sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_STAT)) & I2C_STAT_ACTIVITY) != 0)
		return true;
	else
		return false;
}

/*
 *  I2Cバルク送信セットと待ち
 */
static ER
i2c_master_transmit_wait(I2C_Handle_t * hi2c)
{
	ER ercd = E_OK;
	uint32_t timeout = hi2c->XferSize * 20;

	i2c_int_enable(hi2c, I2C_INTR_STAT_TX_EMPTY);
	while(hi2c->XferCount < hi2c->XferSize){
		if(hi2c->Init.semid != 0)
			twai_sem(hi2c->Init.semid, 1);
		else
			dly_tsk(1);
		if(--timeout == 0){
			ercd = E_TMOUT;
			break;
		}
		if(hi2c->ErrorCode != I2C_ERROR_NONE){
			ercd = E_SYS;
			break;
		}
	}
    return ercd;
}

/*
 *  I2Cスレーブ受信
 *  parameter1 hi2c:   I2Cハンドラへのポインタ
 *  parameter2 pData:  読み出しバッファへのポインタ
 *  parameter3 Size:   読み出しサイズ
 *  return             正常終了ならE_OK
 */
ER
i2c_slaveread(I2C_Handle_t *hi2c, uint8_t *pData, const uint16_t Size)
{
	ER ercd = E_OK;
	uint32_t timeout = Size * 20;

	if(pData == NULL || Size == 0 || hi2c == NULL)
		return E_PAR;
	if(!i2c_is_enabled(hi2c))
		return E_OBJ;

	hi2c->Status = I2C_STATUS_BUSY_RX;
	hi2c->pBuffPtr  = (uint8_t *)pData;
	hi2c->XferSize  = Size;
	hi2c->XferCount = 0;
	hi2c->XferFifo  = 0;
	hi2c->issue_restart = false;
	hi2c->issue_stop = false;
	i2c_int_enable(hi2c, I2C_INTR_STAT_TX_EMPTY);
	while(hi2c->XferCount < hi2c->XferSize){
		if(hi2c->Init.semid != 0)
			twai_sem(hi2c->Init.semid, 1);
		else
			dly_tsk(1);
		if(--timeout == 0){
			ercd = E_TMOUT;
			break;
		}
	}
	hi2c->XferSize  = 0;
	hi2c->XferCount = 0;
    return ercd;
}


/*
 *  I2Cスレーブ送信
 *  parameter1 hi2c:   I2Cハンドラへのポインタ
 *  parameter2 pData:  読み出しバッファへのポインタ
 *  parameter3 Size:   読み出しサイズ
 *  return             正常終了ならE_OK
 */
ER
i2c_slavewrite(I2C_Handle_t *hi2c, uint8_t *data, const uint16_t Size)
{
	ER ercd = E_OK;

	if(data == NULL || Size == 0 || hi2c == NULL)
		return E_PAR;

	/*
	 *  書き込みロック
	 */
	if(hi2c->Init.semlock != 0)
		wai_sem(hi2c->Init.semlock);

	if(!i2c_is_enabled(hi2c)){
		if(hi2c->Init.semlock != 0)
			sig_sem(hi2c->Init.semlock);
		return E_OBJ;
	}

	hi2c->pBuffPtr  = (uint8_t *)data;
	hi2c->XferSize  = Size;
	hi2c->XferCount = 0;
	hi2c->XferFifo  = 0;
	hi2c->issue_restart = false;
	hi2c->issue_stop = false;
	ercd = i2c_master_transmit_wait(hi2c);

	/*
	 *  書き込みロック解除
	 */
	if(hi2c->Init.semlock != 0)
		sig_sem(hi2c->Init.semlock);
    return ercd;
}


/*
 *  I2Cデータリード
 *  parameter1 hi2c:       I2Cハンドラへのポインタ
 *  parameter2 DevAddress: スレーブアドレス
 *  parameter3 MemAddress: メモリアドレス
 *  parameter4 MemAddSize: メモリアドレスサイズ
 *  parameter5 pData:      読み出しバッファへのポインタ
 *  parameter6 Size:       読み出しサイズ
 *  return                 正常終了ならE_OK
 */
ER
i2c_memread(I2C_Handle_t *hi2c, uint16_t devaddr, uint16_t addr, uint16_t addrsize, void * data, uint32_t Size)
{
	uint8_t sendbuffer[4];
	ER      ercd = E_OK;
	uint8_t issue_restart = true;
	uint32_t timeout = Size * 20;

	if(data == NULL || Size == 0 || hi2c == NULL)
		return E_PAR;

	/*
	 *  書き込みロック
	 */
	if(hi2c->Init.semlock != 0)
		wai_sem(hi2c->Init.semlock);

	if(!i2c_is_enabled(hi2c)){
		if(hi2c->Init.semlock != 0)
			sig_sem(hi2c->Init.semlock);
		return E_OBJ;
	}

	hi2c->Status = I2C_STATUS_BUSY_RF;
	/*
	 *  スレーブアドレス設定
	 */
	ercd = i2c_master_target_set(hi2c, devaddr/2);
	if(ercd != E_OK){
		if(hi2c->Init.semlock != 0)
			sig_sem(hi2c->Init.semlock);
		return ercd;
	}

	/*
	 *  アドレス設定があれば送信
	 */
	if(addrsize > 0){
		if(addrsize == 1)
			sendbuffer[0] = addr;
		else{
			sendbuffer[0] = addr>>8;
			sendbuffer[1] = addr & 0xff;
		}
		hi2c->pBuffPtr  = (uint8_t *)sendbuffer;
		hi2c->XferSize  = addrsize;
		hi2c->XferCount = 0;
		hi2c->XferFifo  = 0;
		hi2c->issue_restart = issue_restart;
		hi2c->issue_stop = true;
		ercd = i2c_master_transmit_wait(hi2c);
		issue_restart = false;
		if(ercd != E_OK){
			if(hi2c->Init.semlock != 0)
				sig_sem(hi2c->Init.semlock);
			return ercd;
		}
	}

	/*
	 *  受信設定と受信終了待ち
	 */
	hi2c->Status = I2C_STATUS_BUSY_RX;
	hi2c->pBuffPtr  = (uint8_t *)data;
	hi2c->XferSize  = Size;
	hi2c->XferCount = 0;
	hi2c->XferFifo  = 0;
	hi2c->issue_restart = issue_restart;
	hi2c->issue_stop = true;
	i2c_int_enable(hi2c, I2C_INTR_STAT_TX_EMPTY);
	while(hi2c->XferCount < hi2c->XferSize){
		if(hi2c->Init.semid != 0)
			twai_sem(hi2c->Init.semid, 1);
		else
			dly_tsk(1);
		if(--timeout == 0){
			ercd = E_TMOUT;
			break;
		}
	}
	hi2c->XferSize  = 0;
	hi2c->XferCount = 0;

	/*
	 *  書き込みロック解除
	 */
	if(hi2c->Init.semlock != 0)
		sig_sem(hi2c->Init.semlock);
	return ercd;
}

/*
 *  I2Cデータライト
 *  parameter1 hi2c:       I2Cハンドラへのポインタ
 *  parameter2 DevAddress: スレーブアドレス
 *  parameter3 MemAddress: メモリアドレス
 *  parameter4 MemAddSize: メモリアドレスサイズ
 *  parameter5 pData:      書込みバッファへのポインタ
 *  parameter6 Size:       書込みサイズ
 *  return                 正常終了ならE_OK
 */
ER
i2c_memwrite(I2C_Handle_t *hi2c, uint16_t devaddr, uint16_t addr, uint16_t addrsize, const void * data, uint32_t Size)
{
	uint8_t sendbuffer[4];
    ER ercd = E_OK;
	uint8_t issue_restart = true;

	if(data == NULL || Size == 0 || hi2c == NULL)
		return E_PAR;

	/*
	 *  書き込みロック
	 */
	if(hi2c->Init.semlock != 0)
		wai_sem(hi2c->Init.semlock);

	if(!i2c_is_enabled(hi2c)){
		if(hi2c->Init.semlock != 0)
			sig_sem(hi2c->Init.semlock);
        return E_OBJ;
	}

	hi2c->Status = I2C_STATUS_BUSY_TF;
	/*
	 *  スレーブアドレス設定
	 */
	ercd = i2c_master_target_set(hi2c, devaddr/2);
	if(ercd != E_OK){
		if(hi2c->Init.semlock != 0)
			sig_sem(hi2c->Init.semlock);
		return ercd;
	}

	/*
	 *  アドレス設定があれば送信
	 */
	if(addrsize > 0){
		if(addrsize == 1)
			sendbuffer[0] = addr;
		else{
			sendbuffer[0] = addr>>8;
			sendbuffer[1] = addr & 0xff;
		}
		hi2c->pBuffPtr  = (uint8_t *)sendbuffer;
		hi2c->XferSize  = addrsize;
		hi2c->XferCount = 0;
		hi2c->XferFifo  = addrsize;
		hi2c->issue_restart = true;	/*issue_restart*/
		hi2c->issue_stop = false;
		ercd = i2c_master_transmit_wait(hi2c);
		issue_restart = false;
	}

	/*
	 *  送信設定と送信終了待ち
	 */
	hi2c->Status = I2C_STATUS_BUSY_TX;
	if(ercd == E_OK){
		hi2c->pBuffPtr  = (uint8_t *)data;
		hi2c->XferSize  = Size;
		hi2c->XferCount = 0;
		hi2c->XferFifo  = Size;
		hi2c->issue_restart = issue_restart;
		hi2c->issue_stop = true;
		ercd = i2c_master_transmit_wait(hi2c);
	}

	/*
	 *  書き込みロック解除
	 */
	if(hi2c->Init.semlock != 0)
		sig_sem(hi2c->Init.semlock);
	return ercd;
}

/*
 *  I2Cマスターバルク送信
 */
static void
i2c_master_isr_transmit(I2C_Handle_t * hi2c)
{
	uint32_t temp;

	if(hi2c->XferSize == 1){
		temp = *hi2c->pBuffPtr | (hi2c->issue_restart << 10) | (hi2c->issue_stop << 9);
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA_CMD), temp);
		++hi2c->pBuffPtr;
		++hi2c->XferCount;
	}
    else{
        /* First byte */
		if(hi2c->XferCount == 0){
			temp = *hi2c->pBuffPtr | (hi2c->issue_restart << 10) | (0 << 9);
			sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA_CMD), temp);
			++hi2c->pBuffPtr;
			++hi2c->XferCount;
		}
		else if((hi2c->XferCount+1) < hi2c->XferSize){	/* Middle byte(s) */
			uint32_t count = hi2c->XferSize - hi2c->XferCount;
			uint32_t i;
			uint32_t space = ALT_I2C_TX_FIFO_NUM_ENTRIES - i2c_tx_fifo_level_get(hi2c);
			space = MIN(space, (count-1));
			for(i = 0; i < space; ++i){
				temp = *hi2c->pBuffPtr | (0 << 10) | (0 << 9);
				sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA_CMD), temp);
				++hi2c->pBuffPtr;
			}
			hi2c->XferCount += space;
        }
		else{		/* Last byte */
			temp = *hi2c->pBuffPtr | (0 << 10) | (hi2c->issue_stop << 9);
			sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA_CMD), temp);
			++hi2c->pBuffPtr;
			++hi2c->XferCount;
		}
    }
}

/*
 *  I2Cマスター受信用コマンド送信
 */
static void
i2c_master_isr_command(I2C_Handle_t *hi2c)
{
	uint32_t temp;

	if(hi2c->XferSize == 1){
		temp = I2C_DATA_CMD_CMD | (hi2c->issue_restart << 10) | (hi2c->issue_stop << 9);
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA_CMD), temp);
		++hi2c->XferFifo;
	}
    else{
        /* First byte */
		if(hi2c->XferFifo == 0){
			temp = I2C_DATA_CMD_CMD | (hi2c->issue_restart << 10) | (0 << 9);
			sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA_CMD), temp);
			++hi2c->XferFifo;
		}
		else if((hi2c->XferFifo+1) < hi2c->XferSize){	/* Middle byte(s) */
			uint32_t count = hi2c->XferSize - hi2c->XferFifo;
			uint32_t i;
			uint32_t space = ALT_I2C_TX_FIFO_NUM_ENTRIES - i2c_tx_fifo_level_get(hi2c);
			space = MIN(space, (count-1));
			for(i = 0; i < space; ++i){
				temp = I2C_DATA_CMD_CMD | (0 << 10) | (0 << 9);
				sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA_CMD), temp);
			}
			hi2c->XferFifo += space;
        }
		else{		/* Last byte */
			temp = I2C_DATA_CMD_CMD | (0 << 10) | (hi2c->issue_stop << 9);
			sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA_CMD), temp);
			++hi2c->XferFifo;
		}
    }
	if(hi2c->XferFifo >= hi2c->XferSize)
		i2c_int_disable(hi2c, I2C_INTR_STAT_TX_EMPTY);
}

/*
 *  I2Cスレーブバルク送信
 */
static void
i2c_slave_isr_transmit(I2C_Handle_t * hi2c)
{
	uint32_t count = hi2c->XferSize - hi2c->XferCount;
	uint32_t i, temp;
	uint32_t space = ALT_I2C_TX_FIFO_NUM_ENTRIES - i2c_tx_fifo_level_get(hi2c);
	space = MIN(space, (count-1));
	for(i = 0; i < space; ++i){
		temp = *hi2c->pBuffPtr | (hi2c->issue_restart << 10) | (hi2c->issue_stop << 9);
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA_CMD), temp);
		++hi2c->pBuffPtr;
	}
	hi2c->XferCount += space;
}

static void
i2c_isr_receive(I2C_Handle_t *hi2c)
{
 	uint32_t size = hi2c->XferSize - hi2c->XferCount;
	uint32_t level = i2c_rx_fifo_level_get(hi2c);
	uint32_t i;

	level = MIN(size, level);
	for(i = 0 ; i < level ; ++i){
		*hi2c->pBuffPtr++ = (uint8_t)(sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA_CMD)));
	}
	hi2c->XferCount += level;
}

void
i2c_hps_isr(intptr_t exinf)
{
	ER ercd = E_OK;
	uint32_t i2c_int = 0;
	uint32_t no = INDEX_I2C((uint32_t)exinf);
	uint32_t temp;
	I2C_Handle_t *device;

	device = &I2cHandle[no];
	i2c_int = sil_rew_mem((uint32_t *)(device->base+TOFF_I2C_INTR_STAT));

    /*
	 *  割込みハンドル処理
	 */
	if(i2c_int & I2C_INTR_STAT_RX_UNDER){
		/* Reading the RX FIFO when it is empty.(internal) */
		temp = sil_rew_mem((uint32_t *)(device->base+TOFF_I2C_CLR_RX_UNDER));
		device->ErrorCode |= I2C_ERROR_RXUNDR;
		ercd = E_SYS;
		syslog_1(LOG_ERROR, "i2c_hps_isr: RX_UNDER[%08x] !", i2c_int);
	}
	if(i2c_int & I2C_INTR_STAT_RX_OVER){
		/*
		 *  RX FIFOオーバーフロー
		 */
		temp = sil_rew_mem((uint32_t *)(device->base+TOFF_I2C_CLR_RX_OVER));
		device->ErrorCode |= I2C_ERROR_RXOVR;
		ercd = E_SYS;
		syslog_1(LOG_ERROR, "i2c_hps_isr: RX_OVER[%08x] !", i2c_int);
	}
	if(i2c_int & I2C_INTR_STAT_TX_OVER){
		/*
		 *  TX FIFOオーバーフロー(internal)
		 */
		temp = sil_rew_mem((uint32_t *)(device->base+TOFF_I2C_CLR_TX_OVER));
		device->ErrorCode |= I2C_ERROR_TXOVR;
		ercd = E_SYS;
		syslog_1(LOG_ERROR, "i2c_hps_isr: TX_OVER[%08x] !", i2c_int);
	}
	if(i2c_int & I2C_INTR_STAT_R_TX_ABRT){
		/*
		 *  送信アボート(internal)
		 */
		temp = sil_rew_mem((uint32_t *)(device->base+TOFF_I2C_CLR_TX_ABRT));
		device->ErrorCode |= I2C_ERROR_TXABOUT;
		ercd = E_SYS;
		syslog_1(LOG_ERROR, "i2c_hps_isr: TX_ABRT[%08x] !", i2c_int);
	}
	if(i2c_int & I2C_INTR_STAT_TX_EMPTY){
		if(device->XferCount < device->XferSize){
			if(device->Init.OwnAddress1 == 0)		/* master mode */
				if(device->Status == I2C_STATUS_BUSY_RX)
					i2c_master_isr_command(device);
				else
					i2c_master_isr_transmit(device);
			else									/* slave mode */
				i2c_slave_isr_transmit(device);
		}
		else{
			if(device->XferSize > 0){
				if(device->Status == I2C_STATUS_BUSY_TX){
					device->Status = I2C_STATUS_READY;
					if(device->writecallback != NULL)
						device->writecallback(device);
				}
				device->XferSize  = 0;
				device->XferCount = 0;
			}
			if(device->Init.semid != 0){
				ER ercd = isig_sem(device->Init.semid);
				if(ercd != E_OK && ercd != E_QOVR)
					syslog_1(LOG_ERROR, "i2c_hps_isr[I2C_INTR_STAT_TX_EMPTY] id(%d) isig_sem error(%d)", ercd);
			}
			i2c_int_disable(device, I2C_INTR_STAT_TX_EMPTY);
		}
	}
	if(i2c_int & I2C_INTR_STAT_RX_FULL){
		if(device->Status == I2C_STATUS_BUSY_RX && device->XferCount < device->XferSize){
			i2c_isr_receive(device);
			if(device->XferSize == device->XferCount){
				device->Status = I2C_STATUS_READY;
				if(device->readcallback != NULL)
					device->readcallback(device);
			}
			if(device->Init.semid != 0){
				ER ercd = isig_sem(device->Init.semid);
				if(ercd != E_OK && ercd != E_QOVR)
					syslog_1(LOG_ERROR, "i2c_hps_isr[I2C_INTR_STAT_RX_FULL] isig_sem error(%d)", ercd);
			}
		}
		else{
			uint32_t temp;
			temp = (uint8_t)(sil_rew_mem((uint32_t *)(device->base+TOFF_I2C_DATA_CMD)));
			(void)(temp);
		}
	}
	if(i2c_int & I2C_INTR_STAT_R_RD_REQ){
		/*
		 *  スレーブ側データ受信要求
		 */
		syslog_2(LOG_ERROR, "INFO [ISR]: I2C_INTR_STAT_R_RD_REQ(%d) int(%d)", device->XferCount, (int)exinf);
		temp = sil_rew_mem((uint32_t *)(device->base+TOFF_I2C_CLR_RD_REQ));
	}

	if(i2c_int & I2C_INTR_STAT_R_RX_DONE){
		syslog_2(LOG_ERROR, "INFO [ISR]: I2C_INTR_STAT_R_RX_DONE level(%d) int(%d)", i2c_rx_fifo_level_get(device), (int)exinf);
		temp = sil_rew_mem((uint32_t *)(device->base+TOFF_I2C_CLR_RX_DONE));
	}
	if(i2c_int & I2C_INTR_STAT_R_ACTIVITY){
		syslog_2(LOG_NOTICE, "INFO [ISR]: I2C_INTR_STAT_R_ACTIVITY busy(%d) int(%d)", i2c_is_busy(device), (int)exinf);
		temp = sil_rew_mem((uint32_t *)(device->base+TOFF_I2C_CLR_ACTIVITY));
	}
	if(i2c_int & I2C_INTR_STAT_R_STOP_DET){
		syslog_1(LOG_NOTICE, "INFO [ISR]: I2C_INTR_STAT_R_STOP_DET int(%d)", (int)exinf);
		temp = sil_rew_mem((uint32_t *)(device->base+TOFF_I2C_CLR_STOP_DET));
	}
	if(i2c_int & I2C_INTR_STAT_R_START_DET){
		syslog_1(LOG_NOTICE, "INFO [ISR]: I2C_INTR_STAT_R_START_DET int(%d)", (int)exinf);
		temp = sil_rew_mem((uint32_t *)(device->base+TOFF_I2C_CLR_START_DET));
	}
	if(i2c_int & I2C_INTR_STAT_R_GEN_CALL){
		syslog_1(LOG_NOTICE, "INFO [ISR]: I2C_INTR_STAT_R_GEN_CALL int(%d)", (int)exinf);
		temp = sil_rew_mem((uint32_t *)(device->base+TOFF_I2C_CLR_GEN_CALL));
	}
	(void)(temp);

	if(ercd != E_OK){
		device->Status = I2C_STATUS_BUSY;
		if(device->errorcallback != NULL)
			device->errorcallback(device);
	}
}


