/*
 *  TOPPERS/FMP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Flexible MultiProcessor Kernel
 * 
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2016-2017 by TOPPERS PROJECT Educational Working Group.
 * 
 *  上記著作権者は，以下の(1)～(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 * 
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 * 
 *  $Id: qspi.c 2416 2016-12-20 18:14:20Z roi $
 */

/* 
 *  QSPIドライバの本体
 */

#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include <sil.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include "socfpga_cv.h"
#include <string.h>
#include <qspi.h>

#define sil_orw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) | (b))
#define sil_andw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) & ~(b))
#define sil_modw_mem(a, b, c)	sil_wrw_mem((a), (sil_rew_mem(a) & (~b)) | (c))

/*
 *  QSPIポートIDから管理ブロックを取り出すためのマクロ
 */
#define INDEX_QSPI(qspiid)      ((uint_t)((qspiid) - 1))


#define MIN(a, b) ((a) > (b) ? (b) : (a))
#define GET_MULTIPLIER(ns, mhz) ((((ns) * (mhz)) + 999) / 1000)
#define GET_SR_WIP(value)       ((value >> 0) & 0x1)
#define GET_STIG_RDID_JEDECID(value)        ((value >>  0) & 0xff)
#define GET_STIG_RDID_TYPE_CAPACITY(value)  ((value >>  8) & 0xffff)

/* qspi_clk operating frequency range. */
#define QSPI_CLK_FREQ_MIN     0
#define QSPI_CLK_FREQ_MAX     432000000

#ifndef EXT_CLK_CLKSC1_FREQCUR
#define EXT_CLK_CLKSC1_FREQCUR                 25000000
#endif
#ifndef EXT_CLK_CLKSC2_FREQCUR
#define EXT_CLK_CLKSC2_FREQCUR	               25000000
#endif
#ifndef EXT_CLK_PERIPH_FREQCUR
#define EXT_CLK_PERIPH_FREQCUR                 0
#endif

/*
 *  CLOCK SOURCE
 */
#define CLKMGR_PERPLL_SRC_QSPI_E_F2S_PERIPH_REF_CLK 0x0	/* f2s_periph_ref_clk */
#define CLKMGR_PERPLL_SRC_QSPI_E_MAIN_QSPI_CLK      0x1	/* main_qspi_clk */
#define CLKMGR_PERPLL_SRC_QSPI_E_PERIPH_QSPI_CLK    0x2	/* periph_qspi_clk */

#define CLKMGR_PERPLL_VCO_PSRC_E_EOSC1              0x0	/* eosc1_clk */
#define CLKMGR_PERPLL_VCO_PSRC_E_EOSC2              0x1	/* eosc2_clk */
#define CLKMGR_PERPLL_VCO_PSRC_E_F2S_PERIPH_REF     0x2	/* f2s_periph_ref_clk */


#define LOG_BURST_TYPE_SIZE   2		/* 4 */
#define LOG_SINGLE_TYPE_SIZE  2		/* 4 */


/* The set of all valid QSPI controller interrupt status mask values. */
#define QSPI_INT_ERROR_ALL  (QSPI_IRQSTAT_UNDERFLOWDET  | \
        QSPI_IRQSTAT_INDRDREJECT              | \
        QSPI_IRQSTAT_PROTWRATTEMPT            | \
        QSPI_IRQSTAT_ILLEGALACC               | \
        QSPI_IRQSTAT_RXOVER)

#define QSPI_INT_READ_ALL  (QSPI_IRQSTAT_INDOPDONE | \
        QSPI_IRQSTAT_RXTHRESHCMP | QSPI_IRQSTAT_RXFULL)


#define MAX_DUMMYBYTES                  31
#define ALT_QSPI_TIMEOUT_INFINITE (0xffffffff)

QSPI_Handle_t QSPIHandle[NUM_QSPIPORT];


/*
 *  QSPIのクロック検証
 */
static bool_t
qspi_clk_is_enabled(void)
{
	uint32_t enable = sil_rew_mem((uint32_t *)(TADR_CLKMGR_PERPLL_BASE+TOFF_CLKMGR_PERPLL_EN));
	if((enable & CLKMGR_PERPLL_EN_QSPICLK) != 0)
		return true;
	else
		return false;
}

/*
 *  QSPIデバイスコンフィギュレーション
 */
static ER
qspi_configure_device(QSPI_Handle_t *hqspi, uint32_t qspi_clk_mhz)
{
	QSPI_Init_t *cfg = &hqspi->Init;
    ER ercd = E_OK;
    uint32_t temp, rdid;
	uint16_t manuf_id, type_capacity;

    /*
	 *  デバイスIDを取り出す
	 */
	ercd = qspi_stig_rd_cmd(hqspi, QSPI_STIG_OPCODE_RDID, 0, 4, &rdid, 10000);
	if(ercd != E_OK){
		syslog_0(LOG_ERROR, "qspi_configure_device: get_device_id failed.");
		return ercd;
	}

	manuf_id = GET_STIG_RDID_JEDECID(rdid);
	type_capacity = GET_STIG_RDID_TYPE_CAPACITY(rdid);
	syslog_2(LOG_NOTICE, "qspi_configure_device: manuf_id[%08x] type_capacity[%08x] ##", manuf_id, type_capacity);

	if(cfg->manuf_id != manuf_id || cfg->type_capacity != type_capacity)
		ercd = E_PAR;
	temp = GET_MULTIPLIER(cfg->tshsl_ns, qspi_clk_mhz);
	if(temp > 127)
		ercd = E_PAR;
    temp = GET_MULTIPLIER(cfg->tsd2d_ns, qspi_clk_mhz);
    if(temp > 127)
		ercd = E_PAR;
	temp = GET_MULTIPLIER(cfg->tchsh_ns, qspi_clk_mhz);
	if(temp > 127)
		ercd = E_PAR;
	temp = GET_MULTIPLIER(cfg->tslch_ns, qspi_clk_mhz);
	if(temp > 127)
		ercd = E_PAR;
	if(cfg->clk_phase > 1)
		ercd = E_PAR;
	if(cfg->clk_pol > 1)
		ercd = E_PAR;
    if(cfg->read_op_code > QSPI_DEVRD_RDOPCODE)
        ercd = E_PAR;
	if(cfg->inst_type > MAX_QSPI_MODE)
        ercd = E_PAR;
	if(cfg->read_addr_xfer_type > MAX_QSPI_MODE)
        ercd = E_PAR;
	if(cfg->read_data_xfer_type > MAX_QSPI_MODE)
        ercd = E_PAR;
    if(cfg->read_dummy_cycles > 31)
        ercd = E_PAR;
	if(cfg->write_op_code > QSPI_DEVWR_WROPCODE)
		ercd = E_PAR;
	if(cfg->inst_type > 2)
		ercd = E_PAR;
	if(cfg->write_addr_xfer_type > MAX_QSPI_MODE)
		ercd = E_PAR;
	if(cfg->write_data_xfer_type > MAX_QSPI_MODE)
		ercd = E_PAR;
	if(cfg->write_dummy_cycles > 31)
		ercd = E_PAR;
    if(cfg->page_size > 4095) 
		ercd = E_PAR;
    if(cfg->addr_size > QSPI_DEVSZ_NUMADDRBYTES)
		ercd = E_PAR;
	if(cfg->device_size == 0)
		ercd = E_PAR;
    return ercd;
}

/*
 *  QSPIのアイドル判定
 */
static bool_t
qspi_is_idle(QSPI_Handle_t *hqspi) 
{
	uint32_t cfgreg = sil_rew_mem((uint32_t *)(hqspi->base+TOFF_QSPI_CFG));
	if((cfgreg & QSPI_CFG_IDLE) != 0)
		return true;
	else
		return false;
}

/*
 *  QSPIクロック取得
 */
static uint32_t qspi_clock(void)
{
	uint32_t temp = 0;
	uint64_t temp1 = 0;
	uint64_t numer = 0;
	uint64_t denom = 1;
	uint64_t denom2 = 1;

	temp = (sil_rew_mem((uint32_t *)(TADR_CLKMGR_PERPLL_BASE+TOFF_CLKMGR_PERPLL_SRC)) & CLKMGR_PERPLL_SRC_QSPI) >> 4;
	if(temp == CLKMGR_PERPLL_SRC_QSPI_E_F2S_PERIPH_REF_CLK)
		numer = (uint64_t)EXT_CLK_PERIPH_FREQCUR;
	else if (temp == CLKMGR_PERPLL_SRC_QSPI_E_MAIN_QSPI_CLK){
		temp = sil_rew_mem((uint32_t *)(TADR_CLKMGR_MGR_BASE+TOFF_CLKMGR_BYPASS));
		if((temp & CLKMGR_BYPASS_MAINPLL) != 0)
			temp = EXT_CLK_CLKSC1_FREQCUR;
		else{
			temp = sil_rew_mem((uint32_t *)(TADR_CLKMGR_MPLL_BASE+TOFF_CLKMGR_MPLL_VCO));
			numer = (temp & CLKMGR_MPLL_VCO_NUMER) >> 3;
			denom = (temp & CLKMGR_MPLL_VCO_DENOM) >> 16;
			temp1 = (uint64_t)EXT_CLK_CLKSC1_FREQCUR;
			temp1 *= (numer + 1);
			temp1 /= (denom + 1);

			if (temp1 <= UINT32_MAX)
				temp = (uint32_t)temp1;
			else
				return 0;
		}
		numer = (uint64_t)temp;
		temp = (sil_rew_mem((uint32_t *)(TADR_CLKMGR_MPLL_BASE+TOFF_CLKMGR_MPLL_MAINQSPICLK)) & CLKMGR_MPLL_MAINQSPICLK_CNT) + 1;
		denom = (uint64_t) temp;
	}
	else if(temp == CLKMGR_PERPLL_SRC_QSPI_E_PERIPH_QSPI_CLK){
		temp = sil_rew_mem((uint32_t *)(TADR_CLKMGR_MGR_BASE+TOFF_CLKMGR_BYPASS));
		if((temp & CLKMGR_BYPASS_PERPLL) != 0){
			if((temp & CLKMGR_BYPASS_PERPLLSRC) == 0)
				temp = EXT_CLK_CLKSC1_FREQCUR;
			else
				temp = EXT_CLK_PERIPH_FREQCUR;
		}
		else{
			temp = sil_rew_mem((uint32_t *)(TADR_CLKMGR_PERPLL_BASE+TOFF_CLKMGR_PERPLL_VCO));
			numer = (temp & CLKMGR_PERPLL_VCO_NUMER) >> 3;
			denom2 = (temp & CLKMGR_PERPLL_VCO_DENOM) >> 16;
			temp = (temp & CLKMGR_PERPLL_VCO_PSRC) >> 22;
			if (temp == CLKMGR_PERPLL_VCO_PSRC_E_EOSC1){
				temp1 = (uint64_t)EXT_CLK_CLKSC1_FREQCUR;
			}
			else if (temp == CLKMGR_PERPLL_VCO_PSRC_E_EOSC2){
				temp1 = (uint64_t)EXT_CLK_CLKSC2_FREQCUR;
			}
			else if (temp == CLKMGR_PERPLL_VCO_PSRC_E_F2S_PERIPH_REF){
				temp1 = (uint64_t)EXT_CLK_PERIPH_FREQCUR;
			}

			if(temp1 != 0){
				temp1 *= (numer + 1);
				temp1 /= (denom2 + 1);
				if (temp1 <= UINT32_MAX)
					temp = (uint32_t) temp1;
				else
					return 0;
			}
		}
		numer = (uint64_t)temp;
		denom = (uint64_t)(sil_rew_mem((uint32_t *)(TADR_CLKMGR_PERPLL_BASE+TOFF_CLKMGR_PERPLL_PERQSPICLK)) & CLKMGR_PERPLL_PERQSPICLK_CNT) + 1;
	}
	else
		return 0;

	if(denom > 0){
		numer /= denom;
		if(numer <= UINT32_MAX)
			return (uint32_t)numer;
	}
    return 0;
}

/*
 *  QSPIデバイスの初期化
 *  parameter1 hi2c  I2Cハンドラへのポインタ
 *  return           ハンドラへのポインタ
 */
QSPI_Handle_t *
qspi_init(ID port, QSPI_Init_t *pini)
{
	QSPI_Handle_t *hqspi;
	QSPI_Init_t *cfg;
    ER ercd = E_OK;
	uint32_t no;
    uint32_t cfgreg   = 0;
    uint32_t delayreg = 0;
	uint32_t devrd, devwr, dmaper; 
    int qspi_clk_mhz;
	uint8_t div_bits;
	uint32_t device_sclk_mhz, div_actual, temp;

	if(port < QSPI1_PORTID || port > NUM_QSPIPORT)
		return NULL;

	no = INDEX_QSPI(port);
    /* Validate QSPI module input clocks.
    /  - pclk    - l4_mp_clk
    /  - hclk    - l4_mp_clk
    /  - ref_clk - qspi_clk */

    /*
	 *  QSPIクロックチェック
	 */
    if(!qspi_clk_is_enabled()){
		syslog_0(LOG_ERROR, "qspi_init:clock disable !");
		return NULL;
	}

    /*
	 *  QSPIモジュールリセット
	 */
	sil_orw_mem((uint32_t *)(TADR_RSTMGR_BASE+TOFF_RSTMGR_PERMODRST), RSTMGR_PERMODRST_QSPI);
	sil_dly_nse(10000);
	sil_andw_mem((uint32_t *)(TADR_RSTMGR_BASE+TOFF_RSTMGR_PERMODRST), RSTMGR_PERMODRST_QSPI);

	hqspi = &QSPIHandle[no];
	memset(hqspi, 0, sizeof(QSPI_Handle_t));
	hqspi->base     = TADR_QSPI_BASE;
	hqspi->clk_freq = qspi_clock();
	if(hqspi->clk_freq == QSPI_CLK_FREQ_MIN || hqspi->clk_freq >= QSPI_CLK_FREQ_MAX){
		syslog_1(LOG_ERROR, "qspi_init:clock error(%d) !", hqspi->clk_freq);
		return NULL;
	}
    qspi_clk_mhz = hqspi->clk_freq / 1000000;
	cfg = &hqspi->Init;
	memcpy(cfg, pini, sizeof(QSPI_Init_t));

    /*
	 *  ダイレクト・リードアドレスにゼロをセット
	 */
	sil_wrw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_REMAPADDR), 0);

    /*
	 *  割込み禁止
	 */
	sil_wrw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_IRQMSK), 0);

    /*
	 *  QSPI有効化
	 */
	sil_orw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_CFG), QSPI_CFG_EN);

    /*
	 *  フラッシュデバイスチェック
	 */
	ercd = qspi_configure_device(hqspi, qspi_clk_mhz);
	if(ercd != E_OK){
		syslog_0(LOG_NOTICE, "qspi_init: configure_device failed.");
        return NULL;
	}

    if(qspi_is_idle(hqspi) == false){
		syslog_0(LOG_ERROR, "qspi_init: not idle !");
		return NULL;
	}

    /*
	 *  コンフィギュレーション・レジスタとデレィ・レジスタの初期化
	 */
	cfgreg = sil_rew_mem((uint32_t *)(hqspi->base+TOFF_QSPI_CFG));
	cfgreg &= ~(QSPI_CFG_SELCLKPHASE | QSPI_CFG_SELCLKPOL);

	delayreg |= GET_MULTIPLIER(cfg->tshsl_ns, qspi_clk_mhz)<<24;
	delayreg |= (GET_MULTIPLIER(cfg->tsd2d_ns, qspi_clk_mhz)<<16) & QSPI_DELAY_BTWN;
	delayreg |= (GET_MULTIPLIER(cfg->tchsh_ns, qspi_clk_mhz)<<8) & QSPI_DELAY_AFTER;
	delayreg |= GET_MULTIPLIER(cfg->tslch_ns, qspi_clk_mhz) & QSPI_DELAY_INIT;
	cfgreg |= (cfg->clk_phase<<2) | (cfg->clk_pol<<1);
	sil_wrw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_CFG), cfgreg);
	sil_wrw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_DELAY), delayreg);

	/*
	 *  リード・データ・キャプチャレジスタの設定
	 */
	sil_wrw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_RDDATACAP), (1<<1) | QSPI_RDDATACAP_BYP);

    /*
	 *  ボーレート設定
	 */
	device_sclk_mhz = cfg->max_freq_mhz;
	div_actual = (qspi_clk_mhz + (device_sclk_mhz - 1)) / device_sclk_mhz;
	syslog_1(LOG_NOTICE, "qspi_init: div_actual = %08d.", div_actual);
	div_bits = (uint8_t) (((div_actual + 1) / 2) - 1);
	sil_modw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_CFG), QSPI_CFG_BAUDDIV, (div_bits<<19));

    /*
	 *  デバイス依存リード設定
	 */
	devrd = sil_rew_mem((uint32_t *)(hqspi->base+TOFF_QSPI_DEVRD));
	devrd &= ~(QSPI_DEVRD_RDOPCODE | QSPI_DEVRD_INSTWIDTH |
				QSPI_DEVRD_ADDRWIDTH | QSPI_DEVRD_DATAWIDTH |
				QSPI_DEVRD_DUMMYRDCLKS);
	devrd |= (cfg->read_op_code) | (cfg->inst_type<<8) |
			(cfg->read_addr_xfer_type<<12) | (cfg->read_data_xfer_type<<16) |
			(cfg->read_dummy_cycles<<24);
	sil_wrw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_DEVRD), devrd);

    /*
	 *  デバイス依存ライト設定
	 */
	devwr = sil_rew_mem((uint32_t *)(hqspi->base+TOFF_QSPI_DEVWR));
    devwr &= ~(QSPI_DEVWR_WROPCODE | QSPI_DEVWR_ADDRWIDTH |
				QSPI_DEVWR_DATAWIDTH | QSPI_DEVWR_DUMMYWRCLKS);
    devwr |= cfg->write_op_code | (cfg->write_addr_xfer_type<<12) |
             (cfg->write_data_xfer_type<<16) | (cfg->write_dummy_cycles<<24);
	sil_wrw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_DEVWR), devwr);

    /*
	 *  インストラクション幅、タイプ設定
	 */
	sil_modw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_DEVRD), QSPI_DEVRD_INSTWIDTH, cfg->inst_type<<8);

    /*
	 *  デバイス幅とアドレス設定
	 */
    temp = (QSPI_DEVSZ_BYTESPERSUBSECTOR_RESET<<16) | (cfg->page_size<<4) | cfg->addr_size;
	sil_wrw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_DEVSZ), temp);
	sil_andw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_WRPROT), QSPI_WRPROT_EN);

    /*
	 *  DMAパラメータ設定(DMAは未使用)
	 */
	dmaper = sil_rew_mem((uint32_t *)(hqspi->base+TOFF_QSPI_DMAPER));
    dmaper &= ~(QSPI_DMAPER_NUMBURSTREQBYTES | QSPI_DMAPER_NUMSGLREQBYTES);
    dmaper |= (LOG_BURST_TYPE_SIZE<<8) | (LOG_SINGLE_TYPE_SIZE);
	sil_wrw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_DMAPER), dmaper);

    /*
	 *  初期化終了関数の実行
	 */
    if(cfg->init_func != NULL){
		ercd = cfg->init_func(hqspi);
		if(ercd != E_OK){
			syslog_0(LOG_ERROR, "qspi_ini: config.init failed.");
			return NULL;
		}
	}
	sil_wrw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_IRQMSK), QSPI_INT_ERROR_ALL);
	return hqspi;
}

/*
 *  QSPIデバイスの無効化
 *  parameter1 hqspi QSPIハンドラへのポインタ
 *  return           正常終了時、E_OK
 */
ER
qspi_deinit(QSPI_Handle_t *hqspi)
{
	sil_wrw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_IRQMSK), 0);
	sil_andw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_CFG), QSPI_CFG_EN);
	sil_orw_mem((uint32_t *)(TADR_RSTMGR_BASE+TOFF_RSTMGR_PERMODRST), RSTMGR_PERMODRST_QSPI);
    return E_OK;
}


/*
 *  QSPIインダイレクト・リード処理
 */
static ER
qspi_read_exec(QSPI_Handle_t *hqspi, void * dst, uint32_t src, uint32_t size)
{
	QSPI_Init_t *cfg = &hqspi->Init;
	uint32_t level = 0;
	uint32_t water, i;
	ER ercd = E_OK;

    /*
	 *  読み込み最大サイズチェック
	 */
    if(cfg->micron_multi_die){
        uint32_t max_read_size = cfg->die_size - (src % cfg->die_size);
        if(size > max_read_size){
			return E_PAR;
		}
	}

	hqspi->pBuffPtr  = (uint32_t *)dst;
	hqspi->XferSize  = size;
	hqspi->XferCount = 0;
	hqspi->istatus   = 0;
	sil_wrw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_INDRDSTADDR), src);
	sil_wrw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_INDRDCNT), hqspi->XferSize);
	sil_wrw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_INDRD), (QSPI_INDRD_START | QSPI_INDRD_IND_OPS_DONE_STAT));
	water = size / sizeof(uint32_t);
	if(water > QSPI_SRAM_FIFO_ENTRY_COUNT)
		water = QSPI_SRAM_FIFO_ENTRY_COUNT/2;
	sil_wrw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_INDRDWATER), water);
	sil_orw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_IRQMSK), QSPI_INT_READ_ALL);

    /*
	 *  読み込みFIFO取り込み
	 */
	while((sil_rew_mem((uint32_t *)(hqspi->base+TOFF_QSPI_INDRD)) & QSPI_INDRD_IND_OPS_DONE_STAT) == 0){
		if(hqspi->semid != 0)
			twai_sem(hqspi->semid, 10);
		else
			dly_tsk(1);
		if(hqspi->errorcode != QSPI_ERROR_NONE){
			syslog_1(LOG_ERROR, "qspi_read_exec:error[%08x] !", hqspi->errorcode);
			ercd = E_SYS;
			break;
		}
		level = sil_rew_mem((uint32_t *)(hqspi->base+TOFF_QSPI_SRAMFILL)) & QSPI_SRAMFILL_INDRDPART;
		for(i = 0; i < level; i++){
			*hqspi->pBuffPtr++ = sil_rew_mem((uint32_t *)(TADR_QSPI_DATA_BASE));
			hqspi->XferCount += sizeof(uint32_t);
		}
		water = (hqspi->XferSize - hqspi->XferCount) / sizeof(uint32_t);
		if(water > QSPI_SRAM_FIFO_ENTRY_COUNT)
			water = QSPI_SRAM_FIFO_ENTRY_COUNT/2;
		sil_wrw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_INDRDWATER), water);
		sil_orw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_IRQMSK), (QSPI_IRQSTAT_RXTHRESHCMP | QSPI_IRQSTAT_RXFULL));
	}
	return ercd;
}

/*
 *  インダイレクト・フラッシュ・リード
 *  parameter1 hqspi QSPIハンドラへのポインタ
 *  parameter2 dst   読み込みアドレス
 *  parameter3 src   フラッシュROMスタートアドレス
 *  parameter4 size  読み込みサイズ
 *  return           正常終了時、E_OK
 */
ER
qspi_read(QSPI_Handle_t *hqspi, void * dst, uint32_t src, uint32_t size) 
{
	QSPI_Init_t *cfg = &hqspi->Init;
    ER ercd = E_OK;

	if(hqspi == NULL)
		return E_PAR;
	if(src >= cfg->device_size)
		return E_PAR;
	if((src + size - 1) >= cfg->device_size)
		return E_PAR;
	if(size == 0)
		return E_PAR;
	/*
	 *  パラメータアラインチェック
	 */
    if(((uintptr_t)dst & 0x3) || (src & 0x3) || (size & 0x3))
		return E_PAR;

    /*
	 *  読み込み処理実行中チェック
	 */
	if((sil_rew_mem((uint32_t *)(hqspi->base+TOFF_QSPI_INDRD)) & QSPI_INDRD_RD_STAT) != 0)
		return E_OBJ;

	hqspi->errorcode = QSPI_ERROR_NONE;
	if(cfg->micron_multi_die){
		uint32_t crt_src = src;
		uint32_t crt_size = size;
		uint8_t * crt_dst = (uint8_t*) dst;

		while(ercd == E_OK && crt_size != 0){
			uint32_t max_read_size = cfg->die_size - (crt_src % cfg->die_size);
			uint32_t crt_read_size = MIN(crt_size, max_read_size);

            ercd = qspi_read_exec(hqspi, crt_dst, crt_src, crt_read_size);
			if(ercd == E_OK){
				crt_size -= crt_read_size;
				crt_dst += crt_read_size;
				crt_src += crt_read_size;
			}
		}
		if(crt_size != 0)
			ercd = E_SYS;
	} 
	else{
		ercd = qspi_read_exec(hqspi, dst, src, size);
	}
	return ercd;
}

/*
 *  QSPIインダイレクト・ライト処理
 */
static ER
qspi_write_exec(QSPI_Handle_t *hqspi, uint32_t dst, const void * src, uint32_t size)
{
	QSPI_Init_t *cfg = &hqspi->Init;
	ER ercd = E_OK;
	uint32_t capacity, water;

    /*
	 *  最大書き込みサイズチェック
	 */
    if(cfg->micron_multi_die){
		uint32_t max_read_size = cfg->page_size - (dst % cfg->page_size);
		if(size > max_read_size){
			return E_PAR;
		}
    }

	hqspi->pBuffPtr  = (uint32_t *)src;
	hqspi->XferSize  = size;
	hqspi->XferCount = 0;
	sil_wrw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_INDWRSTADDR), dst);
	sil_wrw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_INDWRCNT), hqspi->XferSize);
	sil_wrw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_INDWR), (QSPI_INDWR_START | QSPI_INDWR_INDDONE));
	sil_orw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_IRQMSK), QSPI_IRQSTAT_TXTHRESHCMP);
	capacity = QSPI_SRAM_FIFO_ENTRY_COUNT - (sil_rew_mem((uint32_t *)(hqspi->base+TOFF_QSPI_SRAMPART)) & QSPI_SRAMPART_ADDR);

    /*
	 *  書き込みFIFO設定
	 */
	while(hqspi->XferCount < hqspi->XferSize){
		uint32_t i;
		uint32_t space = capacity - (sil_rew_mem((uint32_t *)(hqspi->base+TOFF_QSPI_SRAMFILL)) >> 16);
		space = MIN(space, (hqspi->XferSize - hqspi->XferCount) / sizeof(uint32_t));
		water = hqspi->XferSize - hqspi->XferCount - space;
		if(water > QSPI_SRAM_FIFO_ENTRY_COUNT)
			sil_wrw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_INDWRWATER), QSPI_SRAM_FIFO_ENTRY_COUNT/2);
		else
			sil_wrw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_INDWRWATER), QSPI_SRAM_FIFO_ENTRY_COUNT-water);

		for (i = 0; i < space; ++i)
			sil_wrw_mem((uint32_t *)(TADR_QSPI_DATA_BASE), *hqspi->pBuffPtr++);
		hqspi->XferCount += space * sizeof(uint32_t);
		if(hqspi->semid != 0)
			twai_sem(hqspi->semid, 10);
		else
			dly_tsk(1);
		if(hqspi->errorcode != QSPI_ERROR_NONE){
			syslog_1(LOG_ERROR, "qspi_write_exec:error[%08x] !", hqspi->errorcode);
			ercd = E_SYS;
		}
	}
	if(ercd == E_OK)
		ercd = hqspi->Init.wait_func(hqspi, ALT_QSPI_TIMEOUT_INFINITE);
	return ercd;
}

/*
 *  インダイレクト・フラッシュ・ライト
 *  parameter1 hqspi QSPIハンドラへのポインタ
 *  parameter2 dst   フラッシュROMスタートアドレス
 *  parameter3 src   書き込みデータアドレス
 *  parameter4 size  書き込みサイズ
 *  return           正常終了時、E_OK
 */
ER
qspi_write(QSPI_Handle_t *hqspi, uint32_t dst, const void * src, uint32_t size) 
{
	QSPI_Init_t *cfg = &hqspi->Init;
    ER ercd = E_OK;

    if(dst >= cfg->device_size)
		return E_PAR;
	if((dst + size - 1) >= cfg->device_size)
		return E_PAR;
	if(size == 0)
		return E_OK;
	if(((uintptr_t) src & 0x3) || (dst & 0x3) || (size & 0x3))
		return E_PAR;

    /*
	 *  書き込み処理実行中チェック
	 */
	if((sil_rew_mem((uint32_t *)(hqspi->base+TOFF_QSPI_INDWR)) & QSPI_INDWR_RDSTAT) != 0)
		return E_OBJ;

	hqspi->errorcode = QSPI_ERROR_NONE;
    if(cfg->micron_multi_die){
        uint32_t crt_dst = dst;
        uint32_t crt_size = size;
        uint8_t * crt_src = (uint8_t*) src;

        while(ercd == E_OK && crt_size != 0){
			uint32_t max_write_size = cfg->page_size - (crt_dst % cfg->page_size);
			uint32_t crt_write_size = MIN(crt_size, max_write_size);

			ercd = qspi_write_exec(hqspi, crt_dst, crt_src, crt_write_size);
			if(ercd == E_OK){
                crt_size -= crt_write_size;
                crt_dst += crt_write_size;
                crt_src += crt_write_size;
            }
        }
		if(ercd == E_OK && crt_size != 0)
            ercd = E_OBJ;
    }
    else
		ercd = qspi_write_exec(hqspi, dst, src, size);
    return ercd;
}

/*
 *  フラッシュ消去実行関数
 *  parameter1 hqspi   QSPIハンドラへのポインタ
 *  parameter2 address 消去領域の先頭アドレス
 *  parameter3 idx     消去実行番号
 *  return             正常終了時、E_OK
 */
ER
qspi_erase_exec(QSPI_Handle_t *hqspi, uint32_t address, uint32_t idx)
{
	QSPI_Init_t *cfg = &hqspi->Init;
	ER ercd = E_OK;
    uint32_t reg;

    /*
	 *  書き込み設定
	 */
    reg = (QSPI_STIG_OPCODE_WREN<<24);
	ercd = qspi_stig_cmd(hqspi, reg, 10000);
	if(ercd != E_OK)
		return E_OK;

    /*
	 *  消去コマンド送信
	 */
	if(cfg->erase_sizes[idx] == cfg->device_size){
		/* CHIP ERASE */
		reg = (cfg->erase_cmds[idx]<<24) | (0<<7);
		ercd = qspi_stig_cmd(hqspi, reg, ALT_QSPI_TIMEOUT_INFINITE);
	}
	else{
		/* SECTOR ERASE */
		reg = (cfg->erase_cmds[idx]<<24) | (0<<7);
		reg |= QSPI_FLSHCMD_ENCMDADDR | (cfg->addr_size<<16);
		sil_wrw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_FLSHCMDADDR), address);
		ercd = qspi_stig_cmd(hqspi, reg, ALT_QSPI_TIMEOUT_INFINITE);
	}

	/*
	 *  消去終了待ち
	 */
	if(ercd == E_OK)
		ercd = cfg->wait_func(hqspi, ALT_QSPI_TIMEOUT_INFINITE);
	return ercd;
}

/*
 *  フラッシュメモリの消去
 *  parameter1 hqspi   QSPIハンドラへのポインタ
 *  parameter2 address 消去領域の先頭アドレス
 *  parameter3 size    消去バイトサイズ
 *  return             正常終了時、E_OK
 */
ER
qspi_erase(QSPI_Handle_t *hqspi, uint32_t address, uint32_t size)
{
	QSPI_Init_t *cfg = &hqspi->Init;
    uint32_t actual_address = address;
    uint32_t actual_size = size;
    uint32_t crt_address;
    uint32_t crt_size;
    uint32_t remainder;
    ER ercd = E_OK;

    /*
	 *  パラメータチェック
	 */
	if(hqspi == NULL)
		return E_PAR;
	if((address + size) > cfg->device_size)
		return E_PAR;
	if((address + size) > cfg->device_size)
		return E_PAR;

    /*
	 *  消去アライン補正
	 */
	remainder = address % cfg->erase_sizes[0];
	if(remainder != 0){
		actual_address -= remainder;
		actual_size += remainder;
	}
	remainder = actual_size % cfg->erase_sizes[0];
	if(remainder != 0){
		remainder = cfg->erase_sizes[0] - remainder;
		actual_size += remainder;
	}

	/*
	 *  補正が必要ならパラメータエラー
	 */
	if((actual_address != address) || (actual_size != size))
		return E_PAR;

    /*Perform erasure, using the largest erase size option available */
    crt_address = address;
    crt_size = size;
    while ((crt_size != 0) && (ercd == E_OK)){
		int idx;
		/* try from largest to smallest erase size option */
		for(idx = cfg->erase_count - 1; idx >= 0; idx--) {
			uint32_t erase_size = cfg->erase_sizes[idx];

			/* skip if current address is not a multiple of erase size */
			if((crt_address % erase_size) != 0)
				continue;

			/* skip if remaining size is smaller than the erase size */
			if (crt_size < erase_size)
				continue;

            /* perform erasure */
			syslog_2(LOG_NOTICE, "DEBUG[QSPI]: INFO: Erasing %d KBytes at 0x%08x.", (int) erase_size / 1024, (int) crt_address);
			ercd = qspi_erase_exec(hqspi, crt_address, idx);

			/* bail out in case of error */
			if(ercd != E_OK)
				break;

			/* adjust current address and remaining size */
			crt_address += erase_size;
			crt_size -= erase_size;

			/* break from the loop to try again larger sizes, maybe now they are aligned */
			break;
		}
	}

	/*
	 *  消去サイズ再チェック
	 */
	if(crt_size != 0)
		ercd = E_SYS;
    return ercd;
}

/*
 *  ダイレクト・アクセス無効化
 *  parameter1 hqspi   QSPIハンドラへのポインタ
 */
void
qspi_direct_disable(QSPI_Handle_t *hqspi)
{
	sil_andw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_CFG), QSPI_CFG_ENAHBREMAP);
	sil_andw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_CFG), QSPI_CFG_ENDIRACC);
}

/*
 *  ダイレクト・アクセス無効化
 *  parameter1 hqspi   QSPIハンドラへのポインタ
 *  parameter2 addr    参照アドレス
 */
void
qspi_direct_enable(QSPI_Handle_t *hqspi, const uint32_t ahb_remap_addr)
{
	sil_andw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_CFG), QSPI_CFG_ENAHBREMAP);
	sil_wrw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_REMAPADDR), ahb_remap_addr);
	sil_orw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_CFG), QSPI_CFG_ENAHBREMAP);
	sil_orw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_CFG), QSPI_CFG_ENDIRACC);
}

/*
 *  ダイレクト・アクセス参照アドレスを取り出す
 *  parameter1 hqspi   QSPIハンドラへのポインタ
 *  return             リマップアドレス
 */
uint32_t
qspi_ahb_remap_address_get(QSPI_Handle_t *hqspi) 
{
	return sil_rew_mem((uint32_t *)(hqspi->base+TOFF_QSPI_REMAPADDR));
}

/*
 *  CS,CSモード設定
 *  parameter1 hqspi     QSPIハンドラへのポインタ
 *  parameter2 pcs       CS(0-15)格納領域へのポインタ
 *  parameter3 cs_mode   CSモード(0,1)格納領域へのぽいんた
 *  return               正常終了時、E_OK
 */
ER
qspi_chip_select_config_get(QSPI_Handle_t *hqspi, uint32_t* pcs, uint8_t* pcs_mode) 
{
    uint32_t cfg;

	if(hqspi == NULL || pcs == NULL || pcs_mode == NULL)
		return E_PAR;
	cfg = sil_rew_mem((uint32_t *)(hqspi->base+TOFF_QSPI_CFG));
    *pcs = (cfg & QSPI_CFG_PERCSLINES) >> 10;
    *pcs_mode = (cfg & QSPI_CFG_PERSELDEC) >> 9;
	return E_OK;
}

/*
 *  CS,CSモード設定
 *  parameter1 hqspi     QSPIハンドラへのポインタ
 *  parameter2 cs        CS(0-15)
 *  parameter3 cs_mode   CSモード(0,1)
 *  return               正常終了時、E_OK
 */
ER
qspi_chip_select_config_set(QSPI_Handle_t *hqspi, const uint32_t cs, const uint8_t cs_mode)
{
    uint32_t cfg; 

	if(hqspi == NULL)
		return E_PAR;
	if(cs > 15)
		return E_PAR;
    if(cs_mode > QSPI_CS_MODE_DECODE)
		return E_PAR;

    cfg = sil_rew_mem((uint32_t *)(hqspi->base+TOFF_QSPI_CFG));
    cfg &= ~(QSPI_CFG_PERCSLINES & QSPI_CFG_PERSELDEC);
    cfg |= (cs<<10) | (cs_mode<<9);
	sil_wrw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_CFG), cfg);
	return E_OK;
}

/*
 *  モードビット無効設定
 *  parameter1 hqspi     QSPIハンドラへのポインタ
 */
void
qspi_mode_bit_disable(QSPI_Handle_t *hqspi)
{
	sil_andw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_DEVRD), QSPI_DEVRD_ENMODBITS);
}

/*
 *  モードビット有効設定
 *  parameter1 hqspi     QSPIハンドラへのポインタ
 */
void
qspi_mode_bit_enable(QSPI_Handle_t *hqspi)
{
	sil_orw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_DEVRD), QSPI_DEVRD_ENMODBITS);
}

/*
 *  コンフィグレーション・レジスタ・モードビット取り出し
 *  parameter1 hqspi     QSPIハンドラへのポインタ
 *  return               モードビット(0-255)
 */
uint32_t
qspi_mode_bit_config_get(QSPI_Handle_t *hqspi) 
{
    /* Return the 8 bit value from the Mode Field of the Mode Bit Configuration
       Register. */
	return sil_rew_mem((uint32_t *)(hqspi->base+TOFF_QSPI_MODBIT)) & QSPI_MODBIT_MOD;
}

/*
 *  コンフィグレーション・レジスタ・モードビット設定
 *  parameter1 hqspi     QSPIハンドラへのポインタ
 *  parameter2 mode_bits モードビット(0-255)
 *  return               正常終了時、E_OK
 */
ER
qspi_mode_bit_config_set(QSPI_Handle_t *hqspi, const uint32_t mode_bits) 
{
	if (qspi_is_idle(hqspi) == false)
		return E_OBJ;
    if(mode_bits > QSPI_MODBIT_MOD) 
		return E_PAR;

	sil_modw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_MODBIT), QSPI_MODBIT_MOD, mode_bits);
	return E_OK;
}

/*
 *  SRAM中のインダイレクト・リード・パーティション数設定
 *  parameter1 hqspi     QSPIハンドラへのポインタ
 *  parameter2 read_part_size パーティション数(0-127)
 *  return               正常終了時、E_OK
 */
ER
qspi_sram_partition_set(QSPI_Handle_t *hqspi, const uint32_t read_part_size) 
{
	if(read_part_size > QSPI_SRAMPART_ADDR)
		return E_PAR;
	sil_modw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_SRAMPART), QSPI_SRAMPART_ADDR, read_part_size);
	return E_OK;
}


/*
 *  書き込み終了待ち
 */
ER
qspi_sr_wait_write(void *h, uint32_t timeout)
{
	QSPI_Handle_t *hqspi = h;
	ER ercd = E_OK;
	uint8_t sr = 0;
	bool_t infinite = (timeout == ALT_QSPI_TIMEOUT_INFINITE);
    uint32_t data = 0;

    do{
		ercd = qspi_stig_rd_cmd(hqspi, QSPI_STIG_OPCODE_RD_STAT_REG, 0, 1, &data, ALT_QSPI_TIMEOUT_INFINITE);
	    sr = (uint8_t)data;
		if(ercd != E_OK)
			break;
        if(!GET_SR_WIP(sr))
			break;
	}while (timeout-- || infinite);

	if(timeout == (uint32_t) -1 && !infinite){
        ercd = E_TMOUT;
	}
    return ercd;
}

/*
 *  STIGコマンド送信
 */
ER
qspi_stig_cmd(QSPI_Handle_t *hqspi, uint32_t reg_value, uint32_t timeout) 
{
    bool_t infinite = (timeout == ALT_QSPI_TIMEOUT_INFINITE);

	sil_wrw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_FLSHCMD), reg_value);
	sil_wrw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_FLSHCMD), reg_value | QSPI_FLSHCMD_EXECCMD);

	do{
		reg_value = sil_rew_mem((uint32_t *)(hqspi->base+TOFF_QSPI_FLSHCMD));
		if (!(reg_value & QSPI_FLSHCMD_CMDEXECSTAT))
			break;

	}while(timeout-- || infinite);

	if(timeout == (uint32_t) -1 && !infinite)
		return E_TMOUT;
	else
		return E_OK;
}

/*
 *  STIG-READコマンド送信
 */
ER
qspi_stig_rd_cmd(QSPI_Handle_t *hqspi, uint8_t opcode, uint32_t dummy, uint32_t num_bytes, uint32_t * output, uint32_t timeout) 
{
	ER ercd = E_OK;
    uint32_t reg_value;

    if(dummy > MAX_DUMMYBYTES)
		return E_PAR;
	if((num_bytes > 8) || (num_bytes == 0))
		return E_PAR;

	reg_value = (opcode<<24) | QSPI_FLSHCMD_ENRDDATA | ((num_bytes - 1)<<20) | (dummy<<7);
    ercd = qspi_stig_cmd(hqspi, reg_value, timeout);
    if(ercd != E_OK)
		return ercd;

    output[0] = sil_rew_mem((uint32_t *)(hqspi->base+TOFF_QSPI_FLSHCMDRDDATALO));
    if(num_bytes > 4)
		output[1] = sil_rew_mem((uint32_t *)(hqspi->base+TOFF_QSPI_FLSHCMDRDDATAUP));
	return E_OK;
}

/*
 *  STIG-WRITEコマンド送信
 */
ER
qspi_stig_wr_cmd(QSPI_Handle_t *hqspi, uint8_t opcode, uint32_t dummy, uint32_t num_bytes, const uint32_t * input, uint32_t timeout)
{
	uint32_t reg_value;

	if(dummy > MAX_DUMMYBYTES)
		return E_PAR;
	if((num_bytes > 8) || (num_bytes == 0)) 
		return E_PAR;
    reg_value = (opcode<<24) | QSPI_FLSHCMD_ENWRDATA | ((num_bytes - 1)<<12) | (dummy<<7);
	sil_wrw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_FLSHCMDWRDATALO), input[0]);

	if (num_bytes > 4)
		sil_wrw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_FLSHCMDWRDATAUP), input[1]);
    return qspi_stig_cmd(hqspi, reg_value, timeout);
}

/*
 *  QSPI割込みハンドラ
 */
void
qspi_hps_handler(ID portid)
{
	QSPI_Handle_t *hqspi = &QSPIHandle[INDEX_QSPI(portid)];
	uint32_t istatus = sil_rew_mem((uint32_t *)(hqspi->base+TOFF_QSPI_IRQSTAT));

	if((istatus & QSPI_INT_ERROR_ALL) != 0)
		hqspi->errorcode |= istatus & QSPI_INT_ERROR_ALL;
	hqspi->istatus |= istatus & ~QSPI_INT_ERROR_ALL;
	if((istatus & QSPI_IRQSTAT_TXTHRESHCMP) != 0){
		if(hqspi->XferCount >= hqspi->XferSize){
			sil_andw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_IRQMSK), QSPI_IRQSTAT_TXTHRESHCMP);
			sil_wrw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_INDWRWATER), 0xFFFFFFFF);
		}
	}
	if((istatus & QSPI_IRQSTAT_RXTHRESHCMP) != 0)
		sil_andw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_IRQMSK), QSPI_IRQSTAT_RXTHRESHCMP);
	if((istatus & QSPI_IRQSTAT_RXFULL) != 0)
		sil_andw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_IRQMSK), QSPI_IRQSTAT_RXFULL);
	if((istatus & QSPI_IRQSTAT_INDOPDONE) != 0)
		sil_andw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_IRQMSK), QSPI_IRQSTAT_INDOPDONE);
	if(hqspi->semid != 0)
		isig_sem(hqspi->semid);

	/*
	 *  割込みクリア
	 */
	sil_wrw_mem((uint32_t *)(hqspi->base+TOFF_QSPI_IRQSTAT), istatus);
}


