/*
 *  TOPPERS/FMP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Flexible MultiProcessor Kernel
 * 
 *  Copyright (C) 2008-2011 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2017 by TOPPERS PROJECT Educational Working Group.
 * 
 *  上記著作権者は，以下の(1)～(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 * 
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 * 
 *  @(#) $Id: spi.c 698 2017-01-13 18:41:16Z roi $
 */
/*
 * 
 *  I2Cドライバ関数群
 *
 */
#include "kernel_impl.h"
#include <t_syslog.h>
#include <t_stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sil.h>
#include <target_syssvc.h>
#include "spi.h"

#define sil_orw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) | (b))
#define sil_andw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) & ~(b))
#define sil_modw_mem(a, b, c)	sil_wrw_mem((a), (sil_rew_mem(a) & (~b)) | (c))

#define MIN(a, b) ((a) > (b) ? (b) : (a))

/*
 *  SPIOポートIDから管理ブロックを取り出すためのマクロ
 */
#define INDEX_SPI(spiid)        ((uint_t)((spiid) - 1))

/*
 *  SPIモジュールタイムアウト値
 */
#define  SPI_RESET_TMO_INIT     8192

#ifndef SPI_WAIT_TIME
#define SPI_WAIT_TIME           5
#endif

#ifndef EXT_CLK_CLKSC1_FREQCUR
#define EXT_CLK_CLKSC1_FREQCUR                 25000000
#endif
#ifndef EXT_CLK_CLKSC2_FREQCUR
#define EXT_CLK_CLKSC2_FREQCUR	               25000000
#endif
#ifndef EXT_CLK_PERIPH_FREQCUR
#define EXT_CLK_PERIPH_FREQCUR                 0
#endif

/*
 *  CLOCK SOURCE
 */
#define CLKMGR_PERPLL_VCO_PSRC_E_EOSC1         0x0	/* eosc1_clk */
#define CLKMGR_PERPLL_VCO_PSRC_E_EOSC2         0x1	/* eosc2_clk */
#define CLKMGR_PERPLL_VCO_PSRC_E_F2S_PERIPH_REF 0x2	/* f2s_periph_ref_clk */

#define CLKMGR_PERPLL_DIV_SPIMCLK_E_DIV16   0x4

#ifndef SPI_TX_FIFO_NUM_ENTRIES
#define SPI_TX_FIFO_NUM_ENTRIES 256
#endif

#ifndef SPI_RX_FIFO_NUM_ENTRIES
#define SPI_RX_FIFO_NUM_ENTRIES 256
#endif

#define SPI_INT_ERROR           (SPIM_ISR_TXOIS | SPIM_ISR_RXUIS | SPIM_ISR_RXOIS)
#define SPI_INT_ALL             (SPIM_ISR_TXEIS | SPI_INT_ERROR | SPIM_ISR_RXFIS | SPIM_ISR_MSTIS)

#define DMA_CCR_PHAPH           (DMA_CCR_OPT_SB1 | DMA_CCR_OPT_SP(0) | DMA_CCR_OPT_DB1 | DMA_CCR_OPT_DP(0) | DMA_CCR_OPT_ES8)
#define DMA_CCR_MEMSPI          (DMA_CCR_OPT_SAI | DMA_CCR_OPT_SC(7) | DMA_CCR_OPT_DAF | DMA_CCR_OPT_DC(0) | DMA_CCR_PHAPH)
#define DMA_CCR_SPIMEM          (DMA_CCR_OPT_SAF | DMA_CCR_OPT_SC(0) | DMA_CCR_OPT_DAI | DMA_CCR_OPT_DC(7) | DMA_CCR_PHAPH)

/*
 *  SPIポート設定テーブル
 */
static const SPI_PortControlBlock spi_pcb[4] = {
  {	TADR_SPIM1_BASE, INTNO_SPI0_IRQ, SPI_MODE_MASTER, RSTMGR_PERMODRST_SPIM0 },
  {	TADR_SPIM2_BASE, INTNO_SPI1_IRQ, SPI_MODE_MASTER, RSTMGR_PERMODRST_SPIM1 },
  {	TADR_I2C3_BASE, INTNO_SPI2_IRQ, SPI_MODE_SLAVE,  RSTMGR_PERMODRST_SPIS0 },
  {	TADR_I2C4_BASE, INTNO_SPI3_IRQ, SPI_MODE_SLAVE,  RSTMGR_PERMODRST_SPIS1 }
};

static SPI_Handle_t SpiHandle[NUM_SPIPORT];


/*
 *  割込み要因クリア(マスタースレーブ共通)
 */
static void
spi_int_clear(SPI_Handle_t *hspi, const uint32_t mask)
{
	uint32_t temp;

    if(mask == SPI_INT_ALL){
		temp = sil_rew_mem((uint32_t *)(hspi->base+TOFF_SPIM_ICR));
        return;
    }

    /*
	 *  個別の割込みクリア
	 */
    if((mask & SPIM_ISR_RXUIS) != 0)
		temp = sil_rew_mem((uint32_t *)(hspi->base+TOFF_SPIM_TXOICR));
    if((mask & SPIM_ISR_RXOIS) != 0)
		temp = sil_rew_mem((uint32_t *)(hspi->base+TOFF_SPIM_RXOICR));
    if((mask & SPIM_ISR_RXUIS) != 0)
		temp = sil_rew_mem((uint32_t *)(hspi->base+TOFF_SPIM_RXUICR));
	(void)(temp);
}

/*
 *  SPI DISABLE設定(マスタースレーブ共通)
 */
static void
spi_disable(SPI_Handle_t * hspi)
{
    /*
	 *  ENABLE設定判定
	 */
    if((sil_rew_mem((uint32_t *)(hspi->base+TOFF_SPIM_SPIENR)) & SPIM_SPIENR_SPI_EN) == 0)
        return;

	sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPIM_SPIENR), 0);

    /* Clear interrapts mask and clear interrupt status */
	sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPIM_IMR), 0);
    spi_int_clear(hspi, SPI_INT_ALL);
}

/*
 *  送信DMA完了コールバック関数
 */
static void
dma_transtx_end(DMACH_Handle_t *hdmach)
{
	SPI_Handle_t *hspi = hdmach->phandle;

	hspi->txxfercount += hdmach->xfercount;
	if(hspi->Init.semid != 0)
		isig_sem(hspi->Init.semid);
}

/*
 *  受信DMA完了コールバック関数
 */
static void
dma_transrx_end(DMACH_Handle_t *hdmach)
{
	SPI_Handle_t *hspi = hdmach->phandle;

	hspi->rxxfercount += hdmach->xfercount;
	if(hspi->Init.semid != 0)
		isig_sem(hspi->Init.semid);
}

/*
 *  SPIマスター DMA送信設定
 */
static ER
dma_memory_to_spim(DMACH_Handle_t * program)
{
	ER ercd = E_OK;
	SPI_Handle_t * handle = program->phandle;

	ercd = dma_program_move(program, DMA_PROGRAM_REG_DAR, (handle->base+TOFF_SPIM_DR));
    if(ercd != E_OK)
		return ercd;

	/*
	 * Transfer out the current segment.
	 */
	ercd = dma_program_move(program, DMA_PROGRAM_REG_SAR, (uint32_t)program->pbuff);
	if(ercd != E_OK)
		return ercd;

	if(handle->Init.DataSize < 9){
		/*
		 *  DMAコード作成
		 *  - SS8   : Source      burst size of 1-byte
	 	 *  - DS8   : Destination burst size of 1-byte
		 *  - 他はメモリtoSPI設定値
		 */
		ercd = dma_program_move(program, DMA_PROGRAM_REG_CCR,
										( DMA_CCR_OPT_SS8 | DMA_CCR_OPT_DS8 | DMA_CCR_MEMSPI ));
	}
	else{
		/*
		 *  DMAコード作成
		 *  - SS16  : Source      burst size of 2-byte
	 	 *  - DS16  : Destination burst size of 2-byte
		 *  - 他はメモリtoSPI設定値
		 */
		ercd = dma_program_move(program, DMA_PROGRAM_REG_CCR,
										( DMA_CCR_OPT_SS16 | DMA_CCR_OPT_DS16 | DMA_CCR_MEMSPI ));
	}

	while(program->xfercount < program->xfersize){
        uint32_t loopcount = MIN((program->xfersize - program->xfercount), 256);
		if(ercd != E_OK)
			break;

		program->pbuff     += loopcount;
		program->xfercount += loopcount;

		ercd = dma_program_lpstart(program, loopcount);
		if(ercd == E_OK){
			dma_program_periph(program, DMA_PROGRAM_INST_MOD_SINGLE, true);
			ercd = dma_program_lpend(program, DMA_PROGRAM_INST_MOD_SINGLE, loopcount);
		}
	}
	return ercd;
}


/*
 *  SPIマスター DMA受信設定
 */
static ER
dma_spim_to_memory(DMACH_Handle_t * program)
{
    ER ercd = E_OK;
	SPI_Handle_t *handle = program->phandle;

    ercd = dma_program_move(program, DMA_PROGRAM_REG_SAR, (handle->base+TOFF_SPIM_DR));
	if(ercd != E_OK)
		return ercd;
    ercd = dma_program_move(program, DMA_PROGRAM_REG_DAR, (uint32_t)program->pbuff);
	if(ercd != E_OK)
		return ercd;


	if(handle->Init.DataSize < 9){
		/*
		 *  DMAコード作成
		 *  - SS8   : Source      burst size of 1-byte
	 	 *  - DS8   : Destination burst size of 1-byte
		 *  - 他はSPItoメモリ設定値
		 */
		ercd = dma_program_move(program, DMA_PROGRAM_REG_CCR,
										( DMA_CCR_OPT_SS8 | DMA_CCR_OPT_DS8 | DMA_CCR_SPIMEM ));
	}
	else{
		/*
		 *  DMAコード作成
		 *  - SS16  : Source      burst size of 2-byte
	 	 *  - DS16  : Destination burst size of 2-byte
		 *  - 他はSPItoメモリ設定値
		 */
		ercd = dma_program_move(program, DMA_PROGRAM_REG_CCR,
										( DMA_CCR_OPT_SS16 | DMA_CCR_OPT_DS16 | DMA_CCR_SPIMEM ));
	}

	while(program->xfercount < program->xfersize){
		uint32_t loopcount = MIN((program->xfersize - program->xfercount), 256);
		if(ercd != E_OK)
			break;

		program->pbuff     += loopcount;
		program->xfercount += loopcount;

		ercd = dma_program_lpstart(program, loopcount);
		if(ercd == E_OK){
			dma_program_periph(program, DMA_PROGRAM_INST_MOD_SINGLE, false);
			ercd = dma_program_lpend(program, DMA_PROGRAM_INST_MOD_SINGLE, loopcount);
		}
	}
	return ercd;
}

/*
 *  送信転送初期化(マスタースレーブ共通)
 */
static void
spi_tx_init(SPI_Handle_t *hspi, void *tx_buf, uint32_t size)
{
	uint32_t isize = size;

	if(hspi->Init.DataSize > 8)
		isize *= 2;
	if(hspi->hdmachtx != NULL){
		hspi->hdmachtx->dma_start = dma_memory_to_spim;
		hspi->hdmachtx->dma_end   = dma_transtx_end;
		sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPIM_DMACR), SPIM_DMACR_TDMAE);
		sil_modw_mem((uint32_t *)(hspi->base+TOFF_SPIM_DMATDLR), SPIM_DMATDLR_DMATDL, 256-1);
		flushinvalidatedcache_by_addr((void*)tx_buf, isize);
	}
	else
		sil_andw_mem((uint32_t *)(hspi->base+TOFF_SPIM_DMACR), SPIM_DMACR_TDMAE);
	hspi->ptxbuff     = (uint8_t *)tx_buf;
	hspi->txxfersize  = size;
	hspi->txxfercount = 0;
}

/*
 *  受信転送初期化(マスタースレーブ共通)
 */
static void
spi_rx_init(SPI_Handle_t *hspi, void *rx_buf, uint32_t size)
{
	uint32_t isize = size;

	if(hspi->Init.DataSize > 8)
		isize *= 2;
	if(hspi->hdmachrx != NULL){
		hspi->hdmachrx->dma_start = dma_spim_to_memory;
		hspi->hdmachrx->dma_end   = dma_transrx_end;
		sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPIM_DMACR), SPIM_DMACR_RDMAE);
		sil_modw_mem((uint32_t *)(hspi->base+TOFF_SPIM_DMARDLR), SPIM_DMARDLR_DMARDL, 256-1);
		flushinvalidatedcache_by_addr((void*)rx_buf, isize);
	}
	else
		sil_andw_mem((uint32_t *)(hspi->base+TOFF_SPIM_DMACR), SPIM_DMACR_RDMAE);
	hspi->prxbuff     = (uint8_t *)rx_buf;
	hspi->rxxfersize  = size;
	hspi->rxxfercount = 0;
}

/*
 *  受信スレッシュホールド指定(マスタースレーブ共通)
 */
static void
spi_set_rxfifothreshold(SPI_Handle_t *hspi, uint32_t size)
{
	if(size <= SPI_RX_FIFO_NUM_ENTRIES)
		size--;
	else
		size = (size / 2) - 1;
	sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPIM_RXFTLR), size);
}

/*
 *  送信データ転送
 */
static void
spi_tx_transfer(SPI_Handle_t *hspi, uint32_t size)
{
	uint32_t i;

	if(hspi->Init.DataSize > 8){
		uint16_t *p = (uint16_t *)hspi->ptxbuff;
		for(i = 0 ; i < size ; i++)
			sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPIM_DR), *p++);
		hspi->ptxbuff = (uint8_t *)p;
	}
	else{
		for(i = 0 ; i < size ; i++)
			sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPIM_DR), *hspi->ptxbuff++);
	}
	hspi->txxfercount += size;
}

/*
 *  受信データ転送
 */
static void
spi_rx_transfer(SPI_Handle_t *hspi, uint32_t size)
{
	uint32_t i;

	if(hspi->Init.DataSize > 8){
		uint16_t *p = (uint16_t *)hspi->prxbuff;
		for(i = 0 ; i < size ; i++)
			*p++ = sil_rew_mem((uint32_t *)(hspi->base+TOFF_SPIM_DR));
		hspi->prxbuff = (uint8_t *)p;
	}
	else{
		for(i = 0 ; i < size ; i++)
			*hspi->prxbuff++ = sil_rew_mem((uint32_t *)(hspi->base+TOFF_SPIM_DR));
	}
    hspi->rxxfercount += size;
}


bool_t
spiclk_is_enabled(void)
{
	uint32_t status = sil_rew_mem((uint32_t *)(TADR_CLKMGR_PERPLL_BASE+TOFF_CLKMGR_PERPLL_EN));
	if((status & CLKMGR_PERPLL_EN_SPIMCLK) != 0)
		return true;
	else
		return false;
}

static uint32_t spi_clock(void)
{
    uint32_t temp = 0;
	uint64_t temp1 = 0;
    uint64_t numer = 0;
    uint64_t denom = 1;
	uint64_t denom2 = 1;

	temp = sil_rew_mem((uint32_t *)(TADR_CLKMGR_MGR_BASE+TOFF_CLKMGR_BYPASS));
	if((temp & CLKMGR_BYPASS_PERPLL) != 0){
		if((temp & CLKMGR_BYPASS_PERPLLSRC) == 0)
			temp = EXT_CLK_CLKSC1_FREQCUR;
		else
			temp = EXT_CLK_PERIPH_FREQCUR;
	}
	else{
		temp = sil_rew_mem((uint32_t *)(TADR_CLKMGR_PERPLL_BASE+TOFF_CLKMGR_PERPLL_VCO));
		numer = (temp & CLKMGR_PERPLL_VCO_NUMER) >> 3;
		denom2 = (temp & CLKMGR_PERPLL_VCO_DENOM) >> 16;
		temp = (temp & CLKMGR_PERPLL_VCO_PSRC) >> 22;
		if (temp == CLKMGR_PERPLL_VCO_PSRC_E_EOSC1)
			temp1 = (uint64_t)EXT_CLK_CLKSC1_FREQCUR;
		else if (temp == CLKMGR_PERPLL_VCO_PSRC_E_EOSC2)
			temp1 = (uint64_t)EXT_CLK_CLKSC2_FREQCUR;
		else if (temp == CLKMGR_PERPLL_VCO_PSRC_E_F2S_PERIPH_REF)
			temp1 = (uint64_t)EXT_CLK_PERIPH_FREQCUR;

		if (temp1 != 0){
			temp1 *= (numer + 1);
			temp1 /= (denom2 + 1);
			if (temp1 <= UINT32_MAX){
				temp = (uint32_t)temp1;
			}
			else
				return 0;
		}
	}

	numer = (uint64_t) temp;
	denom = (uint64_t)(sil_rew_mem((uint32_t *)(TADR_CLKMGR_PERPLL_BASE+TOFF_CLKMGR_PERPLL_PERBASECLK)) & CLKMGR_PERPLL_PERBASECLK_CNT) + 1;

	temp = (sil_rew_mem((uint32_t *)(TADR_CLKMGR_PERPLL_BASE+TOFF_CLKMGR_PERPLL_DIV)) & CLKMGR_PERPLL_DIV_SPIMCLK) >> 3;
	if(temp <= CLKMGR_PERPLL_DIV_SPIMCLK_E_DIV16)
		denom = denom * (uint64_t)(1 << temp);
	else
		return 0;

	if(denom > 0){
		numer /= denom;
		if(numer <= UINT32_MAX)
			return (uint32_t)numer;
	}
    return 0;
}

/*
 *  SPI初期設定
 *  parameter1  port: SPIポート番号
 *  parameter2  spii: SPI初期設定構造体へのポインタ
 *  return SPIハンドラへのポインタ、NULLでエラー
 */
SPI_Handle_t *
spi_init(ID portid, const SPI_Init_t *cfg)

{
	SPI_Handle_t * hspi;
	uint32_t no;

	if(portid > NUM_SPIPORT || portid == 0)
        return NULL;
	no = INDEX_SPI(portid);

    /*
	 *  パラメータチェック
	 */
	if(cfg->DataSize < MIN_SPI_DFS || cfg->DataSize > MAX_SPI_DFS){
		syslog_1(LOG_ERROR, "spi_init: frame size error(%d) !", cfg->DataSize);
		return NULL;
	}
	if(cfg->RxSampleDelay > SPI_RX_SAMPLE_DELAY_MAX){
		syslog_1(LOG_ERROR, "spi_init: Rx sample delay range error(%d) !", cfg->RxSampleDelay);
		return NULL;
	}
	if(cfg->FrameMode == SPI_FRM_MICROWIRE){
		if(cfg->CtlFrameSize < MIN_SPI_MW_CTS_DFS || cfg->CtlFrameSize > MAX_SPI_MW_CTS_DFS){
			syslog_1(LOG_ERROR, "spi_init: Microware Control frame size error(%d) !", cfg->CtlFrameSize);
			return NULL;
		}
	}

	/*
	 *  クロックチェック
	 */
    if(!spiclk_is_enabled()){
		syslog_0(LOG_NOTICE, "spi_init:no spi clock !");
        return NULL;
    }

	hspi = &SpiHandle[no];
	memset(hspi, 0, sizeof(SPI_Handle_t));
    hspi->base = spi_pcb[no].base;
	hspi->op_mode = spi_pcb[no].mode;

	/*
	 *  SPIクロック設定
	 */
	hspi->clock_freq = spi_clock();	/* SPIM_CLOCK; */
	syslog_1(LOG_NOTICE, "## SPIM_CLOCK (%d) ##", hspi->clock_freq);
	hspi->portid = portid;
	memcpy(&hspi->Init, cfg, sizeof(SPI_Init_t));

    /*
	 *  SPIモジュールリセット
	 */
	spi_reset(hspi);

	/*
	 *  SPIをENABLE設定
	 */
	sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPIM_SPIENR), SPIM_SPIENR_SPI_EN);
	return hspi;
}

/*
 *  SPI終了設定
 *  parameter1  hspi: SPIハンドラへのポインタ
 *  return ERコード
 */
ER
spi_deinit(SPI_Handle_t * hspi)
{
	uint32_t no;
    uint32_t rst_mask;

	if(hspi == NULL)
		return E_PAR;
	if(hspi->portid == 0 || hspi->portid > NUM_SPIPORT)
		return E_PAR;

    /*
	 *  SPIを無効化
	 */
    spi_disable(hspi);

    /*
	 *  SPIモジュールリセット
	 */
	no = INDEX_SPI(hspi->portid);
	rst_mask = spi_pcb[no].resetbit;
	sil_orw_mem((uint32_t *)(TADR_RSTMGR_BASE+TOFF_RSTMGR_PERMODRST), rst_mask);

	hspi->base   = 0;
	hspi->portid = 0;
    return E_OK;
}

/*
 *  リセットSPIデバイス
 *  parameter1  hspi: SPIハンドラへのポインタ
 */
void
spi_reset(SPI_Handle_t * hspi)
{
	SPI_Init_t *cfg;
    bool_t already_enabled;
	uint32_t no, div;
    volatile uint32_t timeout = SPI_RESET_TMO_INIT;
    uint32_t rst_mask;
    uint32_t cfgr;

	if(hspi == NULL)
		return;

	already_enabled = sil_rew_mem((uint32_t *)(hspi->base+TOFF_SPIM_SPIENR)) & SPIM_SPIENR_SPI_EN;
	if(already_enabled){	/* 無効化 */
        spi_disable(hspi);
    }

    /*
	 *  SPIモジュールリセット
	 */
	cfg = &hspi->Init;
	no = INDEX_SPI(hspi->portid);
	rst_mask = spi_pcb[no].resetbit;
	sil_orw_mem((uint32_t *)(TADR_RSTMGR_BASE+TOFF_RSTMGR_PERMODRST), rst_mask);

    /*
	 *  SPIモジュールリセット待ち
	 */
    while (timeout--)
        ;

    /*
	 *  SPIモジュールリセット解除
	 */
	sil_andw_mem((uint32_t *)(TADR_RSTMGR_BASE+TOFF_RSTMGR_PERMODRST), rst_mask);

    /*
	 *  データリセット
	 */
	hspi->errorcode = SPI_ERROR_NONE;
    hspi->slave_select = 0xffff;
	hspi->xmode = cfg->XferMode;

    /*
	 *  CR0レジスタ設定
	 */
	cfgr = sil_rew_mem((uint32_t *)(hspi->base+TOFF_SPIM_CTLR0))
					& ~(SPIM_CTLR0_DFS | SPIM_CTLR0_FRF | SPIM_CTLR0_SCPH |
						SPIM_CTLR0_SCPOL | SPIM_CTLR0_TMOD | SPIM_CTLR0_SRL |
						SPIS_CTLR0_SLV_OE);

	cfgr |= (cfg->DataSize - 1) | (cfg->FrameMode & SPIM_CTLR0_FRF)
					| (cfg->CLKPhase & SPIM_CTLR0_SCPH)
					| (cfg->CLKPolarity & SPIM_CTLR0_SCPOL)
					| (cfg->XferMode & SPIM_CTLR0_TMOD)
					| (cfg->LoopBack & SPIM_CTLR0_SRL);

	if(hspi->op_mode == SPI_MODE_MASTER){
		cfgr |= (cfg->LoopBack & SPIM_CTLR0_SRL);
		sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPIM_RX_SMPL_DLY), cfg->RxSampleDelay);
	}
	else{
		cfgr |= (cfg->SlaveOut & SPIS_CTLR0_SLV_OE);
	}
	sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPIM_CTLR0), cfgr);

    /*
	 *  MMCR設定(必要なら)
	 */
    if(cfg->FrameMode == SPI_FRM_MICROWIRE){
	    uint32_t mwcr = 0;
	    uint32_t mwcr_mask = SPIM_MWCR_MWMOD | SPIM_MWCR_MDD;

		mwcr = (cfg->MwMode & SPIM_MWCR_MWMOD) | (cfg->MwDir & SPIM_MWCR_MDD);
	    /*
		 *  マスターのみの設定
		 */
	    if(hspi->op_mode == SPI_MODE_MASTER){
	        mwcr      |= cfg->MwHandshake & SPIM_MWCR_MHS;
	        mwcr_mask |= SPIM_MWCR_MHS;
	    }
		sil_modw_mem((uint32_t *)(hspi->base+TOFF_SPIM_MWCR), mwcr_mask, mwcr);
		sil_modw_mem((uint32_t *)(hspi->base+TOFF_SPIM_CTLR0), SPIM_CTLR0_CFS, (cfg->CtlFrameSize - 1) << 12);
    }

	/*
	 *  転送スピード設定
	 */
    div = hspi->clock_freq / hspi->Init.Speed;
	sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPIM_BAUDR), div);

	/*
	 *  割込み禁止、割込みクリア
	 */
	sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPIM_IMR), 0);
    spi_int_clear(hspi, SPI_INT_ALL);

	/*
	 *  すでにENABLEなら、SPI ENABLE設定
	 */
    if(already_enabled)
		sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPIM_SPIENR), SPIM_SPIENR_SPI_EN);
}

/*
 *  SPI転送本体
 */
static ER
spi_transfer(SPI_Handle_t *hspi, const uint32_t slave_sel, uint32_t transfer_mode)
{
	ER ercd = E_OK;
    bool_t reenable_needed = false;
	uint32_t len = 0;
    uint32_t timeout;

	timeout = hspi->rxxfersize;
	if(timeout == 0)
		timeout = hspi->txxfersize;
	timeout *= SPI_WAIT_TIME;

	/*
	 *  SPI ENABLE判定
	 */
	if((sil_rew_mem((uint32_t *)(hspi->base+TOFF_SPIM_SPIENR)) & SPIM_SPIENR_SPI_EN) == 0)
        return E_OBJ;

	/*
	 *  SPI BUSYのチェック
	 */
	if((sil_rew_mem((uint32_t *)(hspi->base+TOFF_SPIM_SR)) & SPIM_SR_BUSY) != 0)
		return E_OBJ;

    /*
	 *  マスターモード設定
	 */
	if(hspi->op_mode == SPI_MODE_MASTER){
		/*
		 *  スレーブ設定変更
		 */
	    if(slave_sel != hspi->slave_select){
			if(!reenable_needed){
				spi_disable(hspi);
				reenable_needed = true;
			}
			sil_modw_mem((uint32_t *)(hspi->base+TOFF_SPIM_SER), SPIM_SER_SER, slave_sel);
			hspi->slave_select = slave_sel;
	    }

		/*
		 *  転送モード変更
		 */
		if(transfer_mode != hspi->xmode){
			if(!reenable_needed){
				spi_disable(hspi);
				reenable_needed = true;
			}
			sil_modw_mem((uint32_t *)(hspi->base+TOFF_SPIM_CTLR0), SPIM_CTLR0_TMOD, transfer_mode);
			hspi->xmode = transfer_mode;
		}

		/*
		 *  受信フレームサイズ設定
		 */
		if(transfer_mode == SPI_XMOD_RX){
			if(!reenable_needed){
				spi_disable(hspi);
				reenable_needed = true;
			}
			sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPIM_CTLR1), (hspi->rxxfersize - 1) & SPIM_CTLR1_NDF);
		}

		/*
		 *  DISABLEなら、SPI ENABLE設定
		 */
		if(reenable_needed)
			sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPIM_SPIENR), SPIM_SPIENR_SPI_EN);
	}

	/*
	 *  SPIエラー割込み設定
	 */
	hspi->errorcode = SPI_ERROR_NONE;
	sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPIM_IMR), SPI_INT_ERROR);

	/*
	 *  マスターモード受信なら受信有効化
	 */
    if(hspi->op_mode == SPI_MODE_MASTER && transfer_mode == SPI_XMOD_RX){
		sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPIM_DR), 0);
    }

	len = hspi->rxxfersize - hspi->rxxfercount;
	if(len > 0){
		if(hspi->hdmachrx != NULL){
			ercd = dma_periph_transfer(hspi->hdmachrx, hspi->prxbuff, len);
			if(ercd != E_OK)
				return ercd;
		}
		else{
			spi_set_rxfifothreshold(hspi, len);
			sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPIM_IMR), SPIM_ISR_RXFIS);
		}
	}

	len = hspi->txxfersize - hspi->txxfercount;
	if(len > 0){
		if(hspi->hdmachtx != NULL){
			ercd = dma_periph_transfer(hspi->hdmachtx, (void *)hspi->ptxbuff, len);
			if(ercd != E_OK)
				return ercd;
		}
		else{
			sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPIM_IMR), SPIM_ISR_TXEIS);
		}
	}

	/*
	 *  転送終了待ち
	 */
	while(hspi->txxfercount < hspi->txxfersize || hspi->rxxfercount < hspi->rxxfersize){
		if(hspi->Init.semid != 0)
			twai_sem(hspi->Init.semid, 1);
		else
			dly_tsk(1);
		timeout--;
		if(hspi->errorcode != 0){
			ercd = E_SYS;
			break;
		}
		if(timeout == 0){
	        spi_reset(hspi);
			hspi->errorcode |= SPI_ERROR_TIMEOUT;
			ercd = E_TMOUT;
			break;
		}
	}
    return ercd;
}

/*
 *  SPI送受信実行関数
 *  parameter1  hspi: SPIハンドラへのポインタ
 *  parameter2  slave_select: スレーブ指定
 *  parameter3  ptxdata: 送信バッファへのポインタ(DMAなら32バイトアライン)
 *  parameter4  prxdata: 受信バッファへのポインタ(DMAなら32バイトアライン)
 *  parameter5  length: 送信サイズ
 *  return ERコード
 */
ER
spi_transrecv(SPI_Handle_t *hspi, uint32_t slave_select, uint8_t *tx_buf, uint8_t *rx_buf, uint32_t len)
{
	ER ercd = E_OK;

	if(hspi == NULL)
		return E_PAR;
	if(len == 0)
		return E_OK;
	if(slave_select == 0 || slave_select > SPIM_SER_SER)
		return E_PAR;
	if(hspi->Init.semlock != 0)
		wai_sem(hspi->Init.semlock);

	spi_tx_init(hspi, (void *)tx_buf, len);
	spi_rx_init(hspi, (void *)rx_buf, len);

	ercd = spi_transfer(hspi, slave_select, SPI_XMOD_TXRX);
	if(hspi->Init.semlock != 0)
		sig_sem(hspi->Init.semlock);
	return ercd;
}

/*
 *  SPI送信実行関数
 *  parameter1  hspi: SPIハンドラへのポインタ
 *  parameter2  slave_select: スレーブ指定
 *  parameter3  ptxdata: 送信バッファへのポインタ(DMAなら32バイトアライン)
 *  parameter4  length: 送信サイズ
 *  return ERコード
 */
ER
spi_transmit(SPI_Handle_t *hspi, uint32_t slave_select, uint8_t *tx_buf, uint32_t len)
{
	ER ercd = E_OK;

	if(hspi == NULL)
		return E_PAR;
	if(len == 0)
		return E_OK;
	if(slave_select == 0 || slave_select > SPIM_SER_SER)
		return E_PAR;
	if(hspi->Init.semlock != 0)
		wai_sem(hspi->Init.semlock);

	spi_tx_init(hspi, (void *)tx_buf, len);
	hspi->rxxfersize  = 0;
	hspi->rxxfercount = 0;
	ercd = spi_transfer(hspi, slave_select, SPI_XMOD_TX);
	if(hspi->Init.semlock != 0)
		sig_sem(hspi->Init.semlock);
	return ercd;
}

/*
 *  SPI受信実行関数
 *  parameter1  hspi: SPIハンドラへのポインタ
 *  parameter2  slave_select: スレーブ指定
 *  parameter3  pdata: 受信バッファへのポインタ(DMAなら32バイトアライン)
 *  parameter3  length: 受信サイズ
 *  return ERコード
 */
ER
spi_receive(SPI_Handle_t *hspi, uint32_t slave_select, uint8_t *rx_buf, uint32_t len)
{
	ER ercd = E_OK;

	if(hspi == NULL)
		return E_PAR;
	if(len == 0)
		return E_OK;
	if(slave_select == 0 || slave_select > SPIM_SER_SER)
		return E_PAR;
	if(hspi->Init.semlock != 0)
		wai_sem(hspi->Init.semlock);

	hspi->txxfersize  = 0;
	hspi->txxfercount = 0;
	spi_rx_init(hspi, (void *)rx_buf, len);

	ercd = spi_transfer(hspi, slave_select, SPI_XMOD_RX);
	if(hspi->Init.semlock != 0)
		sig_sem(hspi->Init.semlock);
	return ercd;
}

/*
 *  SPI割込みサービスルーチン(マスタースレーブ共通)
 */
void
spim_hps_isr(intptr_t exinf)
{
	SPI_Handle_t *hspi;
	uint32_t no = (int)exinf - 1;
	uint32_t istatus = 0;
	uint32_t level, space;

	hspi = &SpiHandle[no];
	istatus = sil_rew_mem((uint32_t *)(hspi->base+TOFF_SPIM_ISR));
	syslog_2(LOG_INFO, "spim_hps_isr(%d) [%08x]", exinf, istatus);
	if((istatus & SPIM_ISR_TXEIS) != 0){
		uint32_t tx_len = hspi->txxfersize - hspi->txxfercount;

		if(tx_len > 0){
			level = sil_rew_mem((uint32_t *)(hspi->base+TOFF_SPIM_TXFLR)) & SPIM_TXFLR_TXTFL;

			space = SPI_TX_FIFO_NUM_ENTRIES - level;
			space = MIN((tx_len), space);
			spi_tx_transfer(hspi, space);
		}
		else{
			sil_andw_mem((uint32_t *)(hspi->base+TOFF_SPIM_IMR), SPIM_ISR_TXEIS);
			if(hspi->Init.semid != 0)
				isig_sem(hspi->Init.semid);
		}
	}
	if((istatus & SPIM_RXFLR_RXTFL) != 0){
		uint32_t rx_len = hspi->rxxfersize - hspi->rxxfercount;

		if(rx_len > 0){
			level = sil_rew_mem((uint32_t *)(hspi->base+TOFF_SPIM_RXFLR)) & SPIM_RXFLR_RXTFL;
			level = MIN(rx_len, level);
			space = rx_len - level;
			if(space > 0)
				spi_set_rxfifothreshold(hspi, space);
			spi_rx_transfer(hspi, level);
			if(space == 0 && hspi->Init.semid != 0)
				isig_sem(hspi->Init.semid);
		}
		else
			sil_andw_mem((uint32_t *)(hspi->base+TOFF_SPIM_IMR), SPIM_RXFLR_RXTFL);
	}
	if((istatus & SPI_INT_ERROR) != 0){
		hspi->errorcode |= istatus & SPI_INT_ERROR;
	}

	/*
	 *  割込みクリア
	 */
	spi_int_clear(hspi, istatus);
}

/*
 *  SPI初期化情報を表示
 *  parameter1  hspi: SPIハンドラへのポインタ
 *  return ERコード
 */
ER
spi_config_disp(SPI_Handle_t * hspi)
{
	SPI_Init_t *cfg;
	uint32_t cfgr, mwcr, sdelay;

	if(hspi == NULL)
		return E_PAR;
	cfg = &hspi->Init;
	cfgr = sil_rew_mem((uint32_t *)(hspi->base+TOFF_SPIM_CTLR0));
	mwcr = sil_rew_mem((uint32_t *)(hspi->base+TOFF_SPIM_MWCR));

	if(hspi->op_mode == SPI_MODE_MASTER)
		syslog_2(LOG_NOTICE, "SPI MASTER port(%d) FrameMode[%08x]", hspi->portid, cfg->FrameMode);
	else
		syslog_2(LOG_NOTICE, "SPI SLAVE  port(%d) FrameMode[%08x]", hspi->portid, cfg->FrameMode);

	syslog_2(LOG_NOTICE, " DataSize(%d) (%d)", cfg->DataSize, (cfgr & SPIM_CTLR0_DFS) + 1);
	syslog_2(LOG_NOTICE, " CLKPhase    [%08x][%08x]", cfg->CLKPhase, (cfgr & SPIM_CTLR0_SCPOL));
	syslog_2(LOG_NOTICE, " CLKPolarity [%08x][%08x]", cfg->CLKPolarity, (cfgr & SPIM_CTLR0_SCPH));
	syslog_2(LOG_NOTICE, " XferMode    [%08x][%08x]", cfg->XferMode, (cfgr & SPIM_CTLR0_TMOD));
	if(hspi->op_mode == SPI_MODE_MASTER){
		sdelay = sil_rew_mem((uint32_t *)(hspi->base+TOFF_SPIM_RX_SMPL_DLY)) & SPIM_RX_SMPL_DLY_RSD;
		syslog_2(LOG_NOTICE, " LoopBack    [%08x][%08x]", cfg->LoopBack, (cfgr & SPIM_CTLR0_SRL));
		syslog_2(LOG_NOTICE, " RxSampleDelay (%d) (%d)", cfg->RxSampleDelay, sdelay);
	}
	else{
		syslog_2(LOG_NOTICE, " SlaveOut    [%08x][%08x]", cfg->SlaveOut, (cfgr & SPIS_CTLR0_SLV_OE));
	}
	if(cfg->FrameMode == SPI_FRM_MICROWIRE){
		syslog_2(LOG_NOTICE, " CtlFrameSize  (%d) (%d)", cfg->CtlFrameSize, (((cfgr & SPIM_CTLR0_CFS) >> 12) + 1));
		syslog_2(LOG_NOTICE, " MwMode      [%08x][%08x]", cfg->MwMode, (mwcr & SPIM_MWCR_MWMOD));
		syslog_2(LOG_NOTICE, " MwDir       [%08x][%08x]", cfg->MwDir, (mwcr & SPIM_MWCR_MDD));
		if(hspi->op_mode == SPI_MODE_MASTER)
			syslog_2(LOG_NOTICE, " MwHandshake [%08x][%08x]", cfg->MwHandshake, (mwcr & SPIM_MWCR_MHS));
	}
	dly_tsk(300);
	return E_OK;
}


